/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>
#define __STDC_FORMAT_MACROS
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <algorithm>
#include <limits.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sched.h>
#include <sys/resource.h>

#include <infiniband/ibutils/timespec_utils.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibdm/OutputControl.h>
#include <infiniband/ibdm/SysLogger.h>

#include "ibdiagapp.h"
#include "ibdiagnet_debug_wrap.h"

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h>
#include <infiniband/ibutils/trim.h>

/****************************************************/
/* globals */


/****************************************************/
/* some defines */
#define N_ELEMENTS(array) sizeof(array)/sizeof(array[0])

#define COMPILE_TIME_ASSERT(pred)            \
    switch(0){case 0:case pred:;}

/****************************************************/
/******************private methods*******************/
/****************************************************/
int IBDiagapp::ParseSkipList(const char *skip_list)
{
    IBDIAGNET_ENTER;
    char *pch1 = NULL;
    size_t i, skip_options_list_length = skip_options_list.size();
    u_int32_t skip_mask = 0;
    char *buff = new char[strlen(skip_list) + 1];
    if (!buff) {
        ERR_PRINT("Failed to allocate buff for skip list, exiting...\n");
        IBDIAGNET_RETURN(1);
    }
    strcpy(buff, skip_list);
    pch1 = strtok(buff, ",");
    while (pch1) {
        for (i = 0; i < skip_options_list_length; ++i)
            if (strcmp(pch1, skip_options_list[i].name.c_str()) == 0) {
                skip_mask |= skip_options_list[i].value;
                break;
            }
        if (i == skip_options_list_length) {            //unknown option
            delete [] buff;
            IBDIAGNET_RETURN(1);
        }
        pch1 = strtok(NULL, ",");
    }

    if (skip_mask & SKIP_DUP_GUIDS_VALUE) //skip duplicated guids detection
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->TurnOFFDUPGUIDsDetection();
    if (skip_mask & SKIP_DUP_NODE_DESC_VALUE)               //skip duplicated node description
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->TurnOFFDUPNodeDescDetection();
    // TODO: remove this if entirely on issue #3466238
    if ((skip_mask & SKIP_VS_CAP_SMP_VALUE) || (skip_mask & SKIP_VS_CAP_GMP_VALUE))  // deprecated - skip vs SMP/GMP capability sub stage
        WARN_PRINT("vs SMP/GMP capability stage can no longer be skipped - ignoring skip flag.\n");
    if (skip_mask & SKIP_STAGE_LIDS_VALUE &&
        this->stages[STAGE_IDX_LIDS] != NULL)                  //skip lids stage
        ((LidsStage *)this->stages[STAGE_IDX_LIDS])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_LINKS_VALUE &&
        this->stages[STAGE_IDX_LINKS] != NULL)                 //skip links stage
        ((LinksStage *)this->stages[STAGE_IDX_LINKS])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_SM_VALUE &&
        this->stages[STAGE_IDX_SM] != NULL)                    //skip sm stage
        ((SMStage *)this->stages[STAGE_IDX_SM])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_PM_VALUE &&
        this->stages[STAGE_IDX_PM] != NULL)                    //skip pm stage
        ((PMStage *)this->stages[STAGE_IDX_PM])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_NODES_INFO_VALUE &&
        this->stages[STAGE_IDX_NODES_INFO] != NULL)            //skip nodes_info stage
        ((NodesInfoStage *)this->stages[STAGE_IDX_NODES_INFO])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_SPEED_WIDTH_CHECK &&
        this->stages[STAGE_IDX_SPEED_WIDTH] != NULL)           //skip speed / width check stage
        ((Speed_WidthStage *)this->stages[STAGE_IDX_SPEED_WIDTH])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_PKEY_CHECK &&
        this->stages[STAGE_IDX_PKEY] != NULL)                  //skip pkey stage
        ((PKeyStage *)this->stages[STAGE_IDX_PKEY])->MarkStageAsSkipped();
    if (skip_mask & SKIP_STAGE_TEMPERATURE_SENSING &&
        this->stages[STAGE_IDX_TEMPSENSING] != NULL)           //skip temp sensing stage
        ((TemperatureSensing *)this->stages[STAGE_IDX_TEMPSENSING])->MarkStageAsSkipped();
    if ((skip_mask & SKIP_STAGE_VIRTUALIZATION) &&
        this->stages[STAGE_IDX_VIRTUALIZATION] != NULL)        //skip virtualization stage
        ((VirtualizationStage *)this->stages[STAGE_IDX_VIRTUALIZATION])->MarkStageAsSkipped();
    if ((skip_mask & SKIP_STAGE_ROUTERS) && this->stages[STAGE_IDX_ROUTERS] != NULL)
        ((RoutersStage *)this->stages[STAGE_IDX_ROUTERS])->MarkStageAsSkipped();

    delete [] buff;
    IBDIAGNET_RETURN(0);
}

int IBDiagapp::HandleMKeyParams()
{
    IBDIAGNET_ENTER;

    // parse mkey_val
    if (!mkey_val.empty()) {
        if (is_non_neg_num(mkey_val) || is_hex_num(mkey_val)) {
            ParseMKeyValue();
            if (!mkey_path.empty())
                WARN_PRINT("Ignore mkey_files path, mkey was given as constant value \n");
            goto exit;
        } else {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                           OPTION_MKEY, mkey_val.c_str());
            PRINT("    (Legal value: 64 bit mkey value)\n");
            IBDIAGNET_RETURN(1);
        }
    }

    // parse mkey_path
    if (ParseMKeyFiles())
        IBDIAGNET_RETURN(1);

exit:
    ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetMKeyManager(this->p_mkeymngr);
    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParseKeyValue(const string key_str, u_int64_t &key)
{
    IBDIAGNET_ENTER;

    if ((key_str.size() <= 18) && (key_str.size() >= 3) && (is_hex_num(key_str))) {
        key = strtoull(key_str.c_str(), NULL, 0);
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
}

int IBDiagapp::ParseMKeyValue()
{
    IBDIAGNET_ENTER;

    uint64_t const_mkey = (u_int64_t)strtoull(mkey_val.c_str(), NULL, 0);
    LOG_PRINT("using constant mkey: " U64H_FMT "\n", const_mkey);

    p_mkeymngr = new FilesBasedMKeyManager();
    p_mkeymngr->setConstMKey(const_mkey);

    ibdiag_obj.GetIbisPtr()->SetMKeyManager(p_mkeymngr);
    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParseMKeyFiles()
{
    IBDIAGNET_ENTER;

    p_mkeymngr = new FilesBasedMKeyManager();
    string neighbors_file_path = mkey_path + "/" NEIGHBORS_FILE_NAME;
    string guid2mkey_file_path = mkey_path + "/" GUID2MKEY_FILE_NAME;

    IBDiag::PrintFileTimestamp(neighbors_file_path, "MKEY Neighbors");
    IBDiag::PrintFileTimestamp(guid2mkey_file_path, "MKEY Key");

    int res = p_mkeymngr->parseAndCreateMKeyFabric(neighbors_file_path,
                                                       guid2mkey_file_path);

    char *buffer = ibdmGetAndClearInternalLog();
    if (buffer) {
        PRINT("%s", buffer);
        free(buffer);
    }

    if (res) {
        ERR_PRINT("Can't parse the neighbors and guid2mkey files \n");
        IBDIAGNET_RETURN(1);
    }

    ibdiag_obj.GetIbisPtr()->SetMKeyManager(p_mkeymngr);
    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParseConfigFile()
{
    IBDIAGNET_ENTER;

    struct stat config_file_stat;
    if (stat(this->config_file.c_str(), &config_file_stat) != 0) {
        int errsv = errno;
        if (is_default_config_file) {
            if (errsv == ENOENT) {
                // if defaut config file doesn't exist, runnig will not be failed
                IBDIAGNET_RETURN(0);
            } else {
                ERR_PRINT("Failed to get status for default configuration file: %s, error: %s\n",
                          this->config_file.c_str(), strerror(errsv));
                IBDIAGNET_RETURN(1);
            }
        } else {
            ERR_PRINT("Failed to get status for configuration file: %s, error: %s\n",
                      this->config_file.c_str(), strerror(errsv));
            IBDIAGNET_RETURN(1);
        }
    }

    IBDiag::PrintFileTimestamp(config_file, string("Configuration File"));

    ifstream fs(config_file.c_str());
    if (fs.fail()) {
        if (is_default_config_file)
            ERR_PRINT("Failed to open default configuration file: %s\n", config_file.c_str());
        else
            ERR_PRINT("Failed to open configuration file: %s\n", config_file.c_str());
        IBDIAGNET_RETURN(1);
    }

    char line[1024] = {'\0'};
    unsigned int lineNum = 0;
    list_p_command_line_req p_requesters_list = this->p_main_parser->GetRequestersList();
    list_p_command_line_req::iterator it_req;
    vec_option_t::iterator it_opt;

    while (fs.good()) {
        fs.getline(line, 1024);
        lineNum++;

        // skip empty lines and comments
        if (!line[0] || line[0] == '#')
            continue;

        string line_str(line);
        size_t equal = line_str.find('=');

        // skip lines with no '='
        if (equal == string::npos)
            continue;

        string name = trim(line_str.substr(0, equal));
        string val = trim(line_str.substr(equal + 1, string::npos));

        if (val.empty()) {
            ERR_PRINT("Failed to parse configuration file: %s, invalid value for flag %s, "
                      "value is empty, line: %u\n",
                      config_file.c_str(), name.c_str(), lineNum);
            IBDIAGNET_RETURN(1);
        }

        bool is_bool_flag = false;
        bool is_flag_exist = false;
        for (it_req = p_requesters_list.begin(); it_req != p_requesters_list.end(); ++it_req) {
             vec_option_t options_list = (*it_req)->GetOptions();
             for (it_opt = options_list.begin(); it_opt != options_list.end(); ++it_opt) {
                  if (it_opt->option_name == name) {
                      is_flag_exist = true;
                      if (it_opt->attributes & IBIS_CMD_BOOLEAN_FLAG)
                          is_bool_flag = true;
                      break;
                  }
             }
             if (is_flag_exist)
                 break;
        }

        if (it_req == p_requesters_list.end()) {
            ERR_PRINT("Failed to parse configuration file: %s, flag %s doesn't exist, line: %u\n",
                      config_file.c_str(), name.c_str(), lineNum);
            IBDIAGNET_RETURN(1);
        }

        if (is_bool_flag) {
            if (!(this->IsValidBoolValue(val))) {
                ERR_PRINT("Failed to parse configuration file: %s, "
                          "invalid value - %s - for flag %s, valid values: "
                          OPTION_DEF_VAL_TRUE "/" OPTION_DEF_VAL_FALSE
                          ", line: %u\n", val.c_str(), config_file.c_str(),
                          name.c_str(), lineNum);
                IBDIAGNET_RETURN(1);
            }
        }

        // skip options from command line.
        if (find(p_main_parser->args.begin(), p_main_parser->args.end(), name)
                != p_main_parser->args.end())
            continue;

        for (it_req = p_requesters_list.begin(); it_req != p_requesters_list.end(); ++it_req) {
             int rc = (*it_req)->HandleOption(name, val);
             if (rc == IBDIAG_SUCCESS_CODE)
                 break;
             if (rc != IBDIAG_ERR_CODE_FABRIC_ERROR)
                 IBDIAGNET_RETURN(rc);
        }

        if (it_req == p_requesters_list.end()) {
            ERR_PRINT("Failed to parse configuration file: %s, invalid flag %s, "
                      "line: %u\n", config_file.c_str(), name.c_str(), lineNum);
            IBDIAGNET_RETURN(1);
        }
    }

    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParseGuid2Key(const string &guid_2_key_file_path, map_guid_2_key_t &guid_2_key,
                             string &key_type_str, bool should_print_error,
                             const string &default_file_name)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    struct stat file_path_stat;
    string file_path = guid_2_key_file_path;
    // Check type of file: directory or regular file
    for (int i = 0; i < 2; i++) {
         if (stat(file_path.c_str(),&file_path_stat) == 0) {
             if ((file_path_stat.st_mode & S_IFDIR) && (i < 1)) {
                  file_path += SEPERATOR + default_file_name;
                  continue;
             } else if (!(file_path_stat.st_mode & S_IFREG)) {
                 if (i > 0) {
                     if (should_print_error)
                         ERR_PRINT("%s path with default file name: %s is not regular file\n",
                                   key_type_str.c_str(), file_path.c_str());
                     else
                         WARN_PRINT("%s path with default file name: %s is not regular file\n",
                                    key_type_str.c_str(), file_path.c_str());
                 } else {
                     if (should_print_error)
                     ERR_PRINT("%s path: %s is not regular file or directory\n",
                               key_type_str.c_str(), file_path.c_str());
                     else
                     WARN_PRINT("%s path: %s is not regular file or directory\n",
                                key_type_str.c_str(), file_path.c_str());

                 }
                 IBDIAGNET_RETURN(IBDIAG_ERR_CODE_IO_ERR);
             }
             break;
         } else {
             int ret_v = 0;
             int errsv = errno;
             char buff[1024];

             if (errsv == ENOENT) {
                 snprintf(buff, sizeof(buff), "%s file '%s' wasn't loaded, reason: %s\n",
                          key_type_str.c_str(), file_path.c_str(), strerror(errsv));

                 ret_v = IBDIAG_ERR_CODE_FILE_NOT_EXIST;
             } else {
                 snprintf(buff, sizeof(buff), "Failed to get %s file stat: %s, reason: %s\n",
                          key_type_str.c_str(), file_path.c_str(), strerror(errsv));

                 ret_v = IBDIAG_ERR_CODE_IO_ERR;
             }

             if (should_print_error)
                 ERR_PRINT("%s", buff);
             else
                 WARN_PRINT("%s", buff);

             IBDIAGNET_RETURN(ret_v);
         }
    }

    IBDiag::PrintFileTimestamp(file_path, key_type_str);

    ifstream fs(file_path.c_str());
    if (fs.fail()) {
        ERR_PRINT("Failed to open %s file: %s\n", key_type_str.c_str(),
                  file_path.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FILE_NOT_OPENED);
    }

    // For every line in guid2key file, parse guid and its key.
    // Example :
    //  0x0002c90000000041 0x0000000000000001
    u_int64_t line_num = 0;
    //parse file
    while (fs.good()) {
        char line[1024] = {'\0'};
        fs.getline(line, 1024);
        line_num++;

        char* p_line = line;
        // skip whitespaces
        while(isspace(*p_line))
              ++p_line;

        // skip empty lines and comments
        if (!p_line[0] || p_line[0] == '#')
            continue;

        u_int64_t guid;
        u_int64_t key;
        char tail[100] = {0};
        if (sscanf(p_line, "0x%16" SCNx64 " 0x%16" SCNx64 " %s", &guid, &key, tail) != 2) {
             rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
             WARN_PRINT("Unsupported syntax at line " U64D_FMT ": %s\n", line_num, line);
        } else {
             if (!guid_2_key.insert(pair< u_int64_t, u_int64_t > (guid, key)).second) {
                 WARN_PRINT("Multiple GUID: " U64H_FMT " in %s file: %s\n",
                           guid, key_type_str.c_str(), file_path.c_str());
                 rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
             }
        }
    }

    fs.close();

    if (guid_2_key.empty()) {
        // If file is empty
        if (rc)
             WARN_PRINT("GUID 2 Key database of %s file is empty: %s\n",
                        key_type_str.c_str(), file_path.c_str());
        else
             WARN_PRINT("%s file is empty: %s\n",
                        key_type_str.c_str(), file_path.c_str());
        rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
    }

    IBDIAGNET_RETURN(rc);
}

int IBDiagapp::ParseEnableDisableOutput(const string & lst, bool value)
{
    IBDIAGNET_ENTER;

    string::size_type index = 0;

    do {
    	string::size_type x   = lst.find_first_of(',', index);
    	string            key = lst.substr( index, ( x == string::npos ? lst.size() : x ) - index );

        if(!key.empty()) {
            if (!OutputControl::instance().enabled().set(key, value).is_valid()) {
                ERR_PRINT("Enabled/Disable type cannot be set (type = %s value = %s)\n",
                           key.c_str(), value ? "enabled" : "disabled");
                IBDIAGNET_RETURN(3);
            }
        }
        else
        {
            ERR_PRINT("Type cannot be empty %s\n", lst.c_str() + index);
            IBDIAGNET_RETURN(3);
    	}

        index = (x == string::npos ? string::npos : x + 1);
    } while (index != string::npos);

    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParsePathPerFileOutput(const string & lst)
{
    IBDIAGNET_ENTER;

    if(lst.empty())
    	IBDIAGNET_RETURN(0);

    string::size_type index = 0;

    do {
    	string::size_type x = lst.find_first_of(',', index);
    	string::size_type z = lst.find_first_of('=', index);

    	if(z == string::npos || (x != string::npos && x <= z)) {
            ERR_PRINT("Invalid input format %s\n", lst.c_str() + index);
            IBDIAGNET_RETURN(3);
    	}

    	string key = lst.substr( index, z - index );

    	if(key.empty()) {
            ERR_PRINT("Type cannot be empty %s\n", lst.c_str() + index);
            IBDIAGNET_RETURN(3);
    	}

        string value = lst.substr( z + 1, ( x == string::npos ? lst.size() : x ) - z - 1 );

        if(value.empty()) {
            ERR_PRINT("Value cannot be empty %s\n", lst.c_str() + index);
            IBDIAGNET_RETURN(3);
    	}
        if (!OutputControl::instance().pathes().set(key, value).is_valid()) {
            ERR_PRINT("Path cannot be set (key = %s path = %s)\n", key.c_str(), value.c_str());
            IBDIAGNET_RETURN(3);
        }

        index = (x == string::npos ? string::npos : x + 1);
    } while (index != string::npos);

    IBDIAGNET_RETURN(0);
}

int IBDiagapp::ParsePOption(const char *option_p_list,
        map_str_uint64& counters_to_threshold_map)
{
    IBDIAGNET_ENTER;
    list_string legal_pm_names = this->ibdiag_obj.GetListOFPMNames();
    bool is_legal_counter_name;
    string str = option_p_list;
    string sub_str, sub_str_name, sub_str_thresh;
    size_t i,j;

    do {
        i = str.find(',');
        sub_str = str.substr(0, i);
        str.erase(0, i + 1);
        j = sub_str.find('=');
        sub_str_name = sub_str.substr(0, j);
        sub_str_thresh = sub_str.substr(j + 1, string::npos);

        //check counter for legal name
        is_legal_counter_name = false;
        if (sub_str_name == "all") {
            counters_to_threshold_map[sub_str_name] =
                    (u_int64_t)strtoull(sub_str_thresh.c_str(), NULL, 0);
            is_legal_counter_name = true;
        }
        for (list_string::iterator it = legal_pm_names.begin(); it != legal_pm_names.end(); ++it) {
            if (sub_str_name == (*it)) {
                counters_to_threshold_map[sub_str_name] =
                        (u_int64_t)strtoull(sub_str_thresh.c_str(), NULL, 0);
                is_legal_counter_name = true;
                break;
            }
        }
        if (is_legal_counter_name == false)
            IBDIAGNET_RETURN(1);
    } while (i != string::npos);
    IBDIAGNET_RETURN(0);
}


int IBDiagapp::ParseSkipPluginList(const char *skip_plugin_list)
{
    IBDIAGNET_ENTER;

    char *pch1 = NULL;
    char *buff = new char[strlen(skip_plugin_list) + 1];
    if (!buff) {
        ERR_PRINT("Failed to allocate buff for skip list, exiting...\n");
        IBDIAGNET_RETURN(1);
    }
    strcpy(buff, skip_plugin_list);

    list_string plugins_names;
    string errors;
    if (get_plugins_names(plugins_names, errors)) {
        ERR_PRINT("%s, exiting...\n", errors.c_str());
        delete [] buff;
        IBDIAGNET_RETURN(1);
    }

    pch1 = strtok(buff, ",");
    while (pch1) {
        if (strcmp(pch1, "all") == 0) {
            this->not_to_load_plugins_list.push_front("all");   //we push to front and this
                                                                //will indicate to skip load of all
            delete [] buff;
            IBDIAGNET_RETURN(0);
        }

        list_string::iterator it;
        for (it = plugins_names.begin(); it != plugins_names.end(); ++it)
            if (strcmp(pch1, (*it).c_str()) == 0)
                break;
        if (it == plugins_names.end())  {            //unknown option
            delete [] buff;
            IBDIAGNET_RETURN(1);
        }
        this->not_to_load_plugins_list.push_back(pch1); //push only to back
        pch1 = strtok(NULL, ",");
    }

    delete [] buff;
    IBDIAGNET_RETURN(0);
}


void IBDiagapp::PrintFabricSummaryTable()
{
    IBDIAGNET_ENTER;

    HDR_PRINT("Fabric Summary\n");
    PRINT("\n");

    stringstream summary_str;

    for (int i = 0; i < NUM_OF_STAGES; ++i) {
        if(!this->stages[i])
            continue;
        if (!this->stages[i]->IsSkipped()) {
            if (this->stages[i]->GetFabricSummarySection(summary_str))
                summary_str << "\n";
        }
    }

    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
         it != this->loaded_plugins_list.end(); ++it) {
         Plugin *p_plugin = (*it)->p_plugin_obj;
         if (p_plugin->IsSkipped())
             continue;
         if (p_plugin->GetFabricSummarySection(summary_str))
             summary_str << "\n";
    }

    PRINT("%s", summary_str.str().c_str());

    IBDIAGNET_RETURN_VOID;
}


void IBDiagapp::PrintSummary()
{
    IBDIAGNET_ENTER;

    HDR_PRINT("Summary\n");
    INFO_PRINT("%-35s %-10s %-10s %-10s\n", "Stage", "Warnings", "Errors", "Comment");
    for (int i = 0; i < NUM_OF_STAGES; ++i) {
        if(!this->stages[i])
            continue;
        if (!this->stages[i]->IsSkipped())  //HACK - don't want to see skip comments
            INFO_PRINT("%s\n", this->stages[i]->GetSummaryLine().c_str());
    }
    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
                it != this->loaded_plugins_list.end(); ++it) {
        Plugin *p_plugin = (*it)->p_plugin_obj;
        if (!p_plugin->IsSkipped())         //HACK - don't want to see skip comments
            INFO_PRINT("%s\n", p_plugin->GetSummaryLine().c_str());
    }
    PRINT("\n");
    OutputControl::Properties properties(LOG_FILE);
    if(properties.is_valid()) {
    	if(properties.enabled()) {
            INFO_PRINT("You can find detailed errors/warnings in: %s\n\n\n", properties.path().c_str());
        }
        else {
    		INFO_PRINT("Output log file was disabled.\n\n\n");
    	}
    }
    else {
        ERR_PRINT("PrintSummary::Internal error: output properties for file %s " \
                  "cannot resolved.\n\n\n", LOG_FILE);
    }


    PRINT("%s", this->ibdiag_obj.GetGeneratedFilesListString().c_str());

    for (int i = 0; i < NUM_OF_STAGES; ++i) {
        if(!this->stages[i])
            continue;
        if (this->stages[i]->IsFinished() && this->stages[i]->GetFilesLine() != "")
            PRINT("%s", this->stages[i]->GetFilesLine().c_str());
    }
    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
            it != this->loaded_plugins_list.end(); ++it) {
        Plugin *p_plugin = (*it)->p_plugin_obj;
        if (p_plugin->IsFinished() && p_plugin->GetFilesLine() != "")
            PRINT("%s", p_plugin->GetFilesLine().c_str());
    }

    PRINT("\n");

    IBDIAGNET_RETURN_VOID;
}


//Returns: 0 - OK / 1 - error / 2 - OK but need to exit
int IBDiagapp::GetLocalGUIDFromUser(u_int64_t& guid)
{
    IBDIAGNET_ENTER;

    local_port_t local_ports_array[IBIS_MAX_LOCAL_PORTS];
    u_int32_t local_ports_num;
    u_int32_t choice;
    if (this->ibdiag_obj.GetAllLocalPortGUIDs(local_ports_array, &local_ports_num))
        IBDIAGNET_RETURN(1);

    while (1) {
        PRINT("\nChoose a local port number with which to bind:\n\n");
        for (u_int32_t i = 0; i < local_ports_num; ++i)
            PRINT("\t%u: GUID " U64H_FMT ", lid %u, state %s\n",
                    i + 1, ADB2C_BE64_TO_CPU(local_ports_array[i].guid),
                    local_ports_array[i].lid,
                    portstate2char((IBPortState)local_ports_array[i].logical_state));
        PRINT("\n\t0: Exit\n");
        PRINT("\nEnter choice (0-%u): ", local_ports_num);
        fflush(stdout);
        if (scanf("%u", &choice) <= 0) {
            char junk[128];
            if (scanf("%127s", junk) <= 0)
                PRINT("\nError: Cannot scan!\n");
        } else if (choice == 0) {
            IBDIAGNET_RETURN(2);   //valid exit
        } else if (choice <= local_ports_num) {
            break;
        }
        PRINT("\nLame choice! Please try again.\n");
    }

    --choice;
    //return guid in host order
    guid = ADB2C_BE64_TO_CPU(local_ports_array[choice].guid);
    PRINT("\nChoice GUID=" U64H_FMT "\n", guid);
    IBDIAGNET_RETURN(0);
}

void IBDiagapp::OpenPcapFile()
{
    IBDIAGNET_ENTER;

    OutputControl::Properties properties(PCAP_FILE);

    if(!properties.is_valid() || !properties.enabled())
        return;

    if(!OutputControl::CreateFolder(properties.path()))
        return;

    this->ibdiag_obj.GetIbisPtr()->SetPcapFilePath(properties.path().c_str());

    IBDIAG_RETURN_VOID;
}

/********************************************************/
/*******************protected methods********************/
/********************************************************/

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS
int IBDiagapp::DumpStartInfoCSVTable(int argc, const char **argv)
{
    IBDIAGNET_ENTER;
    list_string plugins_names;
    string errors;
    stringstream sstream;

    if (!csv_out.DumpStart(SECTION_RUN_INFO)) {
        sstream << "IBDIAGNET_Version,IBDIAG_Version,IBDM_Version,IBIS_Version,Date,Args" << endl;
        csv_out.WriteBuf(sstream.str());

        sstream.str("");
        sstream <<  this->get_version();
        sstream  << ",\"" << IBFabric::GetTimestamp() << "\",\"";

        for (int i = 1; i < argc; ++i)
            sstream << " " << argv[i];

        sstream << "\"" << endl;
        csv_out.WriteBuf(sstream.str());

        csv_out.DumpEnd(SECTION_RUN_INFO);
    }

    sstream.str("");
    if (csv_out.DumpStart(SECTION_PLUGINS_INFO))
       IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    sstream << "Plugin,Location" << endl;
    csv_out.WriteBuf(sstream.str());
    if (get_plugins_names(plugins_names, errors)) {
        //no use to print error here
        IBDIAGNET_RETURN(1);
    }

    for (list_string::iterator it = plugins_names.begin(); it != plugins_names.end(); ++it) {
        sstream.str("");
        sstream << *it << "," << get_plugin_path_str(*it) << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_PLUGINS_INFO);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
IBDiagapp::IBDiagapp() : CommandLineRequester("Main"),
        ibdiagnet_status(FAILED_TO_CONSTRUCT),
        p_main_parser(NULL),
        num_of_errors_to_screen(DEFAULT_NUM_ERRORS_TO_SCREEN),
        m_max_smps_on_wire(DEFAULT_SMPS_WINDOW),
        m_max_gmps_on_wire(DEFAULT_GMPS_WINDOW),
        max_iterations(1),
        timeout_iteration_mic_sec(0),
        p_mkeymngr(NULL),
        is_default_config_file(true),
        config_file(DEFAULT_CONFIG_FILE)
{
    IBDIAGNET_ENTER;
    memset(stages, 0, sizeof(stages));
    IBDIAGNET_RETURN_VOID;
}



IBDiagapp::~IBDiagapp()
{
    IBDIAGNET_ENTER;

    //Make sure that there are no pending and on the wire MADs.
    ibdiag_obj.GetIbisPtr()->MadRecAll();

    if (this->ibdiagnet_status != FAILED_TO_CONSTRUCT) {
        this->ibdiagnet_status = FAILED_TO_CONSTRUCT;
        for (int i = 0; i < NUM_OF_STAGES; ++i)
            if (this->stages[i]) {
                delete this->stages[i];
                this->stages[i] = NULL;
            }
        delete this->p_main_parser;
        destroy_log_file();
    }
    if (p_mkeymngr)
        delete p_mkeymngr;

    this->UnloadPlugins();
    IBDIAGNET_RETURN_VOID;
}


//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int IBDiagapp::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;

    /* printf("Parsing (requester=\"%s\") \"%s\" with val=\"%s\"\n",
            this->getName().c_str(), name.c_str(), value.c_str()); */

    bool bool_flag_val = true;

    if (value == OPTION_DEF_VAL_NULL)
        IBDIAGNET_RETURN(0);

    if (name == OPTION_DEVICE) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetDeviceName(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PORT) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'-%c\' option: %s\n",
                    OPTION_PORT_SHORT, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetPortNum((u_int8_t)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_GUID) {
        if (value != "0" && !is_hex_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'-%s\' option: %s\n",
                    OPTION_GUID, value.c_str());
            PRINT("    Either hex numbers or 0 are allowed\n");
            IBDIAGNET_RETURN(3);
        }

        u_int64_t given_guid = strtoull(value.c_str(), NULL, 0);
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetPortGUID(given_guid);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SKIP) {
        if (this->ParseSkipList(value.c_str())) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SKIP, value.c_str());
            PRINT("    (Legal value: ");
            size_t skip_options_list_length = skip_options_list.size();
            for (size_t i = 0; i < skip_options_list_length - 1; ++i) {
                PRINT("%s | ", skip_options_list[i].name.c_str());
            }
            PRINT("%s)\n", skip_options_list[skip_options_list_length - 1].name.c_str());
            IBDIAGNET_RETURN(3);
        }

        bool dupGuidsDetection = ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->GetDUPGUIDsDetection();
        bool switchDupGuidsDetection = ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->GetCheckSwitchDuplicatedGuids();
        if (!dupGuidsDetection && switchDupGuidsDetection){
            ERR_PRINT_ARGS("Inconsistent %s option; You cannot enable it and disable %s\n",
            OPTION_ENABLE_SWITCH_DUP_GUID_CHECK, SKIP_DUP_GUIDS_DESC_PRINT);
            IBDIAGNET_RETURN(3);
        }

        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SKIP_PLUGIN) {
        if (this->ParseSkipPluginList(value.c_str())) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SKIP_PLUGIN, value.c_str());
            list_string plugins_names;
            string errors;
            if (get_plugins_names(plugins_names, errors)) {
                //no use to print error here
                IBDIAGNET_RETURN(3);
            }
            PRINT("    (Legal value: ");
            for (list_string::iterator it = plugins_names.begin(); it != plugins_names.end(); ++it) {
                if (it != plugins_names.begin())
                    PRINT(" | ");
                PRINT("%s", (*it).c_str());
            }
            PRINT(")\n");
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_PM_GET_ALL) {
        this->ParseBoolValue(value, bool_flag_val);

        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPerSLVLCounters(bool_flag_val);
        ((MlnxCntrsStage *)this->stages[STAGE_IDX_MLNX_CNTRS])->MarkStage(bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetCheckExtSpeedsCounters("all");
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPerLaneCounters(bool_flag_val);

        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_PM_CLEAR_ALL) {
        this->ParseBoolValue(value, bool_flag_val);

        ((PMStage*)this->stages[STAGE_IDX_PM])->SetResetCounters(bool_flag_val);
        ((MlnxCntrsStage*)this->stages[STAGE_IDX_MLNX_CNTRS])->SetResetCounters(bool_flag_val);

        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_PER_SLVL_CNTRS) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPerSLVLCounters(bool_flag_val);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_PER_SLVL_SUP_WARN) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPerSLVLSupWarn(bool_flag_val);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_PC) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetResetCounters(bool_flag_val);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_MLNX_CNTRS) {
        this->ParseBoolValue(value, bool_flag_val);
        ((MlnxCntrsStage *)this->stages[STAGE_IDX_MLNX_CNTRS])->MarkStage(bool_flag_val);
        IBDIAGNET_RETURN(0);

     } else if (name == OPTION_MLNX_CNTRS_RESET) {
        this->ParseBoolValue(value, bool_flag_val);
        ((MlnxCntrsStage *)this->stages[STAGE_IDX_MLNX_CNTRS])->SetResetCounters(bool_flag_val);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_P) {
        map_str_uint64& counters_to_threshold_map =
                ((PMStage *)this->stages[STAGE_IDX_PM])->GetRefToCountersToThresholdMap();
        if (this->ParsePOption(value.c_str(), counters_to_threshold_map)) {
            ERR_PRINT_ARGS("Illegal value for \'-%c\' option: %s\n"  \
                    "    (Legal value: one or more \"<PM counter>=<Thresh Limit>\"\n",
                    OPTION_P_SHORT, value.c_str());
            PRINT("    Legal PM Counter names are:\n");
            list_string allowed_names_list = this->ibdiag_obj.GetListOFPMNames();
            for (list_string::iterator it = allowed_names_list.begin();
                    it != allowed_names_list.end(); ++it)
                PRINT("    %s\n", it->c_str());
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PMPAUSE) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_PMPAUSE, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPauseTimeBetweenSamples(
            (u_int32_t)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BER) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetBERTest(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BERDATA) {
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BERTHRESH) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_BERTHRESH, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetBERThreshold(strtoul(value.c_str(), NULL, 0));
        ibdiag_obj.SetBERThreshold(strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_LLR_ACTIVE_CELL) {
        if (strcmp(value.c_str(), "64") && strcmp(value.c_str(), "128") &&
            strcmp(value.c_str(), "0")) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n" \
                    "    (Legal value: %s).\n",
                    OPTION_LLR_ACTIVE_CELL, value.c_str(), OPTION_LLR_ACTIVE_CELL_ARG_DESC);
            IBDIAGNET_RETURN(3);
        }
        ibdiag_obj.SetLLRActiveCell(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_EXT_SPEEDS) {
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetCheckExtSpeedsCounters(value.c_str());
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_ALL_LANES) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPerLaneCounters(bool_flag_val);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_LLR_INTERNAL_COUNTERS) {
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetPrintLLRCounters();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_LS) {
        if ((char2speed(value.c_str()) == IB_UNKNOWN_LINK_SPEED) &&
            (value != IB_DISABLE_LINK_SPEED_STR)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n" \
                           "    (Legal value: %s).\n",
                           OPTION_LS, value.c_str(), OPTION_LS_ARG_DESC);
            IBDIAGNET_RETURN(3);
        }
        ((Speed_WidthStage *)this->stages[STAGE_IDX_SPEED_WIDTH])->SetExpectedLinkSpeed(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_LW) {
        if ((char2width(value.c_str()) == IB_UNKNOWN_LINK_WIDTH) &&
            (value != IB_DISABLE_LINK_WIDTH_STR)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n" \
                    "    (Legal value: %s).\n",
                    OPTION_LW, value.c_str(), OPTION_LW_ARG_DESC);
            IBDIAGNET_RETURN(3);
        }
        ((Speed_WidthStage *)this->stages[STAGE_IDX_SPEED_WIDTH])->SetExpectedLinkWidth(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SCRRENNUMERRS) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SCRRENNUMERRS, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->num_of_errors_to_screen = (unsigned int)strtoul(value.c_str(), NULL, 0);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SMP_WINDOW) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SMP_WINDOW, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->m_max_smps_on_wire = (unsigned int)strtoul(value.c_str(), NULL, 0);
        if (m_max_smps_on_wire > MAX_SMPS_WINDOW) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s > %d\n",
                    OPTION_SMP_WINDOW, value.c_str(), MAX_SMPS_WINDOW);
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_GMP_WINDOW) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_GMP_WINDOW, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->m_max_gmps_on_wire = (unsigned int)strtoul(value.c_str(), NULL, 0);
        if (m_max_gmps_on_wire > MAX_GMPS_WINDOW) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s > %d\n",
                    OPTION_GMP_WINDOW, value.c_str(), MAX_GMPS_WINDOW);
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MAXHOPS) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_MAXHOPS, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetMaxHops((u_int8_t)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_NO_MEPI) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetNoMEPI();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_FEC_MODE) {
        ERR_PRINT("Flag " OPTION_FEC_MODE " is deprecated, please use " OPTION_ENABLE_OUTPUT "/"
                  OPTION_DISABLE_OUTPUT " flags in order to control db_csv sections\n");
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_TOPO_MATCH_LOCAL_SYSTEM) {
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_VERSION) {
        INFO_PRINT("%s\n", IBDIAGNET_VERSION);
        INFO_PRINT("Dependencies: %s\n", get_ibdiag_version().c_str());
        IBDIAGNET_RETURN(2);           //exit with no error
    } else if (name == OPTION_HELP) {
        this->LoadPlugins();
        SCREEN_PRINT("%s\n", this->p_main_parser->GetUsage().c_str());
        IBDIAGNET_RETURN(2);           //exit with no error
    }  else if (name == OPTION_DEEP_HELP) {
        this->LoadPlugins();
        SCREEN_PRINT("%s\n", this->p_main_parser->GetUsage().c_str());
        IBDIAGNET_RETURN(2);           //exit with no error
    }   else if (name == OPTION_MLNX_HELP) {
        this->LoadPlugins();
        SCREEN_PRINT("%s\n", this->p_main_parser->GetUsage(true).c_str());
        IBDIAGNET_RETURN(2);           //exit with no error
    } else if (name == OPTION_TOPO_COMPARE_NAMES) {
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CAPABILITY_WRITE) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetOutputCapabilityFile(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_FULL_CAPABILITY_WRITE) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetOutputFullCapabilityFile(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CAPABILITY_READ) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
            SetInputCapabilityFile(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BACK_COMPAT_DB) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
            SetShowPortsDataFormat(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MADS_TIMEOUT) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_MADS_TIMEOUT, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ibdiag_obj.GetIbisPtr()->SetTimeout((int)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MADS_RETRIES) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_MADS_RETRIES, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        ibdiag_obj.GetIbisPtr()->SetNumOfRetries((int)strtoul(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_DUMP_LST_WITH_LMC) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetDumpLSTWithLMC();
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MKEY) {
        mkey_val = value;
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_MKEY_FILES) {
        mkey_path = value;
        this->SetVSKeyPath(value, false);
        this->SetCCKeyPath(value, false);
        this->SetM2NKeyPath(value, false);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_VS_KEY) {
        u_int64_t vs_key;
        if (!this->ParseKeyValue(value, vs_key))
            this->ibdiag_obj.GetIbisPtr()->GetKeyManager().SetDefaultKey(vs_key, IBIS_VS_KEY);
        else
            WARN_PRINT("Failed to parse VS Key: %s\n", value.c_str());
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_VS_KEY_FILE) {
        this->SetVSKeyPath(value, true);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_CC_KEY) {
        u_int64_t vs_key;
        if (!this->ParseKeyValue(value, vs_key))
            this->ibdiag_obj.GetIbisPtr()->GetKeyManager().SetDefaultKey(vs_key, IBIS_CC_KEY);
        else
            WARN_PRINT("Failed to parse CC Key: %s\n", value.c_str());
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_CC_KEY_FILE) {
        this->SetCCKeyPath(value, true);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_M2N_KEY) {
        u_int64_t key;
        if (!this->ParseKeyValue(value, key))
            this->ibdiag_obj.GetIbisPtr()->GetKeyManager().SetDefaultKey(key, IBIS_CLASS_C_KEY);
        else
            WARN_PRINT("Failed to parse M2N Key: %s\n", value.c_str());
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_M2N_KEY_FILE) {
        this->SetM2NKeyPath(value, true);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_CONFIG) {
        this->SetConfigFilePath(value);
        IBDIAGNET_RETURN(0);

    // deprecated option
    } else if (name == OPTION_LLR_COUNTERS || name == OPTION_LLR_SYMBOL_ERROR) {
        SCREEN_ERR_PRINT("The --%s option is deprecated\n", name.c_str());
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_MAP) {
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->
            SetNodeNameMapFile(value);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_ROUTERS) {
        this->ParseBoolValue(value, bool_flag_val);
        ((RoutersStage *)this->stages[STAGE_IDX_ROUTERS])->
            MarkStage(bool_flag_val);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_QOS) {
        this->ParseBoolValue(value, bool_flag_val);
        ((QoSStage *)this->stages[STAGE_IDX_QOS])->
            MarkStage(bool_flag_val);

        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CONGESTION_CONTROL) {
        if (!this->stages[STAGE_IDX_CC]->IsActive()) {
            this->ParseBoolValue(value, bool_flag_val);
            ((CongestionControlStage *)this->stages[STAGE_IDX_CC])->
                MarkStage(bool_flag_val);
        }

        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_CONGESTION_COUNTERS) {
        this->ParseBoolValue(value, bool_flag_val);
        if (bool_flag_val){
            ((CongestionControlStage *)this->stages[STAGE_IDX_CC])->
                MarkStage(bool_flag_val);
        }

        ((CongestionControlStage *)this->stages[STAGE_IDX_CC])->
            SetCongestionCountersGet(bool_flag_val);

        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CLEAR_CONGESTION_COUNTERS) {
        this->ParseBoolValue(value, bool_flag_val);
        if (bool_flag_val){
            ((CongestionControlStage *)this->stages[STAGE_IDX_CC])->
                MarkStage(bool_flag_val);
        }

	((CongestionControlStage *)this->stages[STAGE_IDX_CC])->
            SetCongestionCountersGet(bool_flag_val);

        ((CongestionControlStage *)this->stages[STAGE_IDX_CC])->
            SetClearCongestionCounters(bool_flag_val);

        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_FAST_RECOVERY) {
         if (!this->stages[STAGE_IDX_FAST_RECOVERY]->IsActive()) {
             this->ParseBoolValue(value, bool_flag_val);
             ((FastRecoveryStage *)this->stages[STAGE_IDX_FAST_RECOVERY])->
                 MarkStage(bool_flag_val);
         }

         IBDIAGNET_RETURN(0);

    } else if (name == OPTION_ENABLE_SWITCH_DUP_GUID_CHECK) {
        this->ParseBoolValue(value, bool_flag_val);
        bool dupGuidDetection = ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->GetDUPGUIDsDetection();
        if (!dupGuidDetection && bool_flag_val){
            ERR_PRINT_ARGS("Inconsistent %s option; You cannot enable it and disable %s\n",
            OPTION_ENABLE_SWITCH_DUP_GUID_CHECK, SKIP_DUP_GUIDS_DESC_PRINT);
            IBDIAGNET_RETURN(3);
        }

        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetCheckSwitchDuplicatedGuids(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_DBG_MODULE) {
    	this->m_debug_wraper.set_modules(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_DBG_LEVEL) {
    	this->m_debug_wraper.set_levels(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_OUTPATH) {
        if(!value.empty())
        {
        	if(value[value.size() - 1] != '/')
        		value += "/";

            if(!OutputControl::instance().pathes().set("default", value).is_valid() ||
               !OutputControl::instance().pathes().set("csv:default", value).is_valid())
            {
                ERR_PRINT_ARGS("'%s' cannot be set to '%s' because default value already set\n",
                               OPTION_OUTPATH, value.c_str());
                IBDIAGNET_RETURN(3);
            }
        }
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_ENABLE_OUTPUT) {
        IBDIAGNET_RETURN(this->ParseEnableDisableOutput(value, true));
    } else if (name == OPTION_DISABLE_OUTPUT) {
        IBDIAGNET_RETURN(this->ParseEnableDisableOutput(value, false));
    } else if (name == OPTION_PATH_PER_FILE) {
        IBDIAGNET_RETURN(this->ParsePathPerFileOutput(value));
    } else if (name == OPTION_DISCOVERY_ONLY) {
        this->ParseBoolValue(value, bool_flag_val);
        if (!bool_flag_val)
            IBDIAGNET_RETURN(0);
        string list="dup_guids,dup_node_desc,lids,sm,pkey,links," \
                    "pm,speed_width_check,temp_sensing,virt,routers";
        this->ParseSkipList(list.c_str());
        list="default";
        this->ParseEnableDisableOutput(list, false);
        list="db_csv,log,debug,ibnetdiscover";
        this->ParseEnableDisableOutput(list, true);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SET_SL) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                    OPTION_SET_SL, value.c_str());
            ERR_PRINT("Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }

        unsigned long sl = (unsigned long)strtoul(value.c_str(), NULL, 0);
        if (sl > MAX_SL) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s > %d\n",
                    OPTION_SET_SL, value.c_str(), MAX_SL);
            IBDIAGNET_RETURN(3);
        }

        //not default configuration
        if (sl != 0) {
            //ibis has max lid value harcoded.
            vector<uint8_t> pslTable(0xffff, (u_int8_t)sl);
            ibdiag_obj.GetIbisPtr()->SetPSLTable(pslTable);
        }

        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_LONG_RUN_ITERATION) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                           OPTION_LONG_RUN_ITERATION, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }
        this->SetMaxIterations((u_int64_t)strtoull(value.c_str(), NULL, 0));
        SysLogger::instance().set_enabled(false);
        IBDIAGNET_RETURN(ParseEnableDisableOutput("all", false));
    } else if (name == OPTION_LONG_RUN_TIMEOUT) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                           OPTION_LONG_RUN_TIMEOUT, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }

        this->SetTimeoutIteration((u_int64_t)strtoull(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SYS_LOG) {
        if (!is_non_neg_num(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n", OPTION_SYS_LOG, value.c_str());
            PRINT("    Only non negative number is allowed\n");
            IBDIAGNET_RETURN(3);
        }

        SysLogger::instance().set_enabled(strtoull(value.c_str(), NULL, 0));
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SKIP_DEF) { // DEBUG HIDDEN OPTION
        this->ParseSkipList("dup_guids,dup_node_desc,lids,sm,nodes_info,pkey,links,pm,speed_width_check,temp_sensing,virt");
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PPCC) {
        ((CongestionControlStage *)this->stages[STAGE_IDX_CC])->SetPPCCAlgoPath(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_VERBS) {
        this->ParseBoolValue(value, bool_flag_val);
        this->ibdiag_obj.GetIbisPtr()->SetVerbsEnabled(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_QDI_HISTOGRAM) {
        this->ParseBoolValue(value, bool_flag_val);
        ((PMStage *)this->stages[STAGE_IDX_PM])->SetQdiHistogram(bool_flag_val);
        IBDIAGNET_RETURN(0);
    }
    //if we reached here than unknown option
    IBDIAGNET_RETURN(1);
}

int IBDiagapp::PostParseFiles()
{
    IBDIAGNET_ENTER;

    int rc;

    if (!this->GetVSKeyPath().empty()) {
        string key_type_str = "VS Key";
        map_guid_2_key_t &vs_key_map =
                     ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->GetVSKeyMap();
        rc = this->ParseGuid2Key(this->GetVSKeyPath(), vs_key_map, key_type_str, false,
                                 string(GUID2VSKEY_FILE_NAME));
        if (rc == IBDIAG_ERR_CODE_FILE_NOT_OPENED || rc == IBDIAG_ERR_CODE_IO_ERR)
            IBDIAGNET_RETURN(3);
        if (rc == IBDIAG_ERR_CODE_PARSE_FILE_FAILED) {
            WARN_PRINT("Parse %s file was finished with warnings\n\n", key_type_str.c_str());
        } else if (rc == IBDIAG_SUCCESS_CODE) {
            INFO_PRINT("Parse %s file was finished successfully\n\n", key_type_str.c_str());
        }
    }

    if (!this->GetCCKeyPath().empty()) {
        string key_type_str = "CC Key";
        map_guid_2_key_t &cc_key_map =
                     ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->GetCCKeyMap();
        rc = this->ParseGuid2Key(this->GetCCKeyPath(), cc_key_map, key_type_str, false,
                                 string(GUID2CCKEY_FILE_NAME));
        if (rc == IBDIAG_ERR_CODE_FILE_NOT_OPENED || rc == IBDIAG_ERR_CODE_IO_ERR)
            IBDIAGNET_RETURN(3);
        if (rc == IBDIAG_ERR_CODE_PARSE_FILE_FAILED) {
            WARN_PRINT("Parse %s file was finished with warnings\n\n", key_type_str.c_str());
        } else if (rc == IBDIAG_SUCCESS_CODE) {
            INFO_PRINT("Parse %s file was finished successfully\n\n", key_type_str.c_str());
        }
    }

    if (!this->GetM2NKeyPath().empty()) {
        string key_type_str = "M2N Key";
        map_guid_2_key_t &m2n_key_map =
                     ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->GetM2NKeyMap();

        rc = this->ParseGuid2Key(this->GetM2NKeyPath(), m2n_key_map, key_type_str, false,
                                 string(GUID2_M2N_KEY_FILE_NAME));

        if (rc == IBDIAG_ERR_CODE_FILE_NOT_OPENED || rc == IBDIAG_ERR_CODE_IO_ERR)
            IBDIAGNET_RETURN(3);

        if (rc == IBDIAG_ERR_CODE_PARSE_FILE_FAILED) {
            WARN_PRINT("Parse %s file was finished with warnings\n\n", key_type_str.c_str());
        } else if (rc == IBDIAG_SUCCESS_CODE) {
            INFO_PRINT("Parse %s file was finished successfully\n\n", key_type_str.c_str());
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiagapp::LoadPlugins()
{
    IBDIAGNET_ENTER;

    /*
     * load all plugins
     */
    list_string plugins_names;
    string errors;
    if (get_plugins_names(plugins_names, errors)) {
        ERR_PRINT("%s\n", errors.c_str());
        IBDIAGNET_RETURN_VOID
    }

    // sort the plugin list in reverse order, to make sure the loading order remains the same,
    // and PHY diag plugin is loaded before cable diag plugin.
    plugins_names.sort(std::greater<string>());

    symbols_plugin_t *p_curr_sym_plugin;
    PRINT("----------\n");
    PRINT("Load Plugins from:\n%s\n", get_plugins_paths_str().c_str());
    PRINT("%-45s %-10s %s\n", "Plugin Name", "Result", "Comment");
    for (list_string:: iterator it = plugins_names.begin(); it != plugins_names.end(); ++it) {
        PRINT("%-45s ", (*it).c_str());
        for (list_string::iterator it2 = this->not_to_load_plugins_list.begin();
                it2 != this->not_to_load_plugins_list.end(); ++it2) {
            if ((*it) == (*it2)) {
                PRINT("%-10s %s\n", "", "Skipped");
                goto next_plugin_load;
            }
        }
        p_curr_sym_plugin = load_plugin((*it),
                &this->ibdiag_obj,
                errors);
        if (!p_curr_sym_plugin) {
            PRINT("%-10s %s\n", "Failed", errors.c_str());
            continue;
        }
        if (this->p_main_parser->AddRequester(p_curr_sym_plugin->p_plugin_obj)) {
            unload_plugin(p_curr_sym_plugin);
            PRINT("%-10s Plugin options issue - %s\n", "Failed", this->p_main_parser->GetErrDesc());
            continue;
        }
        PRINT("%-10s %s\n", "Succeeded", "Plugin loaded");
        this->loaded_plugins_list.push_back(p_curr_sym_plugin);

next_plugin_load:
        continue;
    }
    plugins_names.clear();
    PRINT("\n");

    IBDIAGNET_RETURN_VOID;
}


void IBDiagapp::UnloadPlugins()
{
    IBDIAGNET_ENTER;
    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
            it != this->loaded_plugins_list.end(); ++it) {
        unload_plugin(*it);
    }
    this->loaded_plugins_list.clear();
    this->not_to_load_plugins_list.clear();
    IBDIAGNET_RETURN_VOID;
}


bool IBDiagapp::isValidLoadFromFilePath()
{
    if (((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->IsLoadFromFile()) {
        char *res;
        int   errsv;
        char  in_file[PATH_MAX + 1] = {0};
        char  out_file[PATH_MAX + 1] = {0};

        OutputControl::Properties properties(DB_FILE);
        if(!properties.is_valid()) {
            ERR_PRINT("isValidLoadFromFilePath::Internal error: CSV output path cannot resolved.\n");
            return false;
        }

        string input_file =
        	    ((DiscoveryFabricStage *)this->stages[STAGE_IDX_DISCOVERY])->
                    GetLoadFromFilePath();

        if(!(res = realpath(input_file.c_str(), in_file))) {
            errsv = errno;
            ERR_PRINT("CSV file input path %s: %s\n", input_file.c_str(), strerror(errsv));
            return false;
        }

        if (!(res = realpath(properties.path().c_str(), out_file))) {
        	errsv = errno;

        	if(errsv != ENOENT) {
                ERR_PRINT("CSV output file %s: %s\n", properties.path().c_str(), strerror(errsv));
                return false;
            }
        	else
        		return true;
        }

        if (!strcmp(in_file, out_file)) {
            ERR_PRINT("The CSV file path in the \"load from file\" flag (%s)\n"
                      "    is the same as the output CSV file path (%s)\n",
					  in_file, out_file);
            return false;
        }
    }

    return true;
}


int IBDiagapp::CreateConfigFile(string file_path)
{
    IBDIAGNET_ENTER;

    ofstream fs;
    string error_message;
    int rc = IBFabric::OpenFile(file_path.c_str(),
                                fs,
                                false, // to append
                                error_message,
                                true);
    if (rc) {
        ERR_PRINT("Configuration file: %s, %s\n", file_path.c_str(), error_message.c_str());
        IBDIAGNET_RETURN(1);
    }

    list_p_command_line_req p_requesters_list = this->p_main_parser->GetRequestersList();

    for (list_p_command_line_req::iterator it_req = p_requesters_list.begin();
         it_req != p_requesters_list.end(); ++it_req) {
         vec_option_t options_list = (*it_req)->GetOptions();
         for (vec_option_t::iterator it_opt = options_list.begin();
              it_opt != options_list.end(); ++it_opt) {
              if (it_opt->attributes & IBIS_CMD_NO_CONF_FILE)
                  continue;

              fs << "# ";
              const char *p_description = it_opt->description.c_str();
              size_t line_length = 0;
              for (size_t i = 0; p_description[i]; i++) {
                   line_length++;
                   if ((p_description[i] == ' ') && (line_length > OPT_DESC_CONFIG_LEN)) {
                       fs << "\n# ";
                       line_length = 0;
                       continue;
                   }

                   fs << p_description[i];
                   if (p_description[i] == '\n') {
                       fs << "# ";
                       line_length = 0;
                   }
              }

              fs << endl;

              if (it_opt->attributes & IBIS_CMD_BOOLEAN_FLAG) {
                  fs << "# value type: Boolean" << endl;
              } else {
                  fs << "# value type: " << it_opt->option_value << endl;
              }

              fs << it_opt->option_name << " = " << it_opt->default_value_str << endl;
              fs << endl;
         }
    }

    SCREEN_PRINT("Congiguration file was created: %s\n", file_path.c_str());
    fs.close();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


void IBDiagapp::SetIBNLDir(const string &path) {

	string value = trim(path);

	if(*value.rbegin() != '/')
		value += '/';

	OutputControl::instance().enabled().set("ibnl", true);
	OutputControl::instance().pathes().set("ibnl", value);

}
//Returns: 0 - OK / 1 - error / 2 - OK but need to exit
int IBDiagapp::GetOpt(int argc, const char **argv)
{
    IBDIAGNET_ENTER;

    int rc2, rc = 1;
    list_p_command_line_req ignored_req_list;
    string log_file;
    stringstream sstr;

    if (this->ibdiagnet_status != NEED_TO_GET_ARGS) {
        SCREEN_ERR_PRINT("Can not parse arguments from user\n");
        IBDIAGNET_RETURN(rc);
    }
    /*
     * add the requester to the parser
     */
    if (this->p_main_parser->AddRequester(this)) {
        SCREEN_ERR_PRINT("Failed add options to main parser, err=%s\n", this->p_main_parser->GetErrDesc());
        goto exit;
    }

    IBFabric::running_version = this->get_version();
    // add argv to log and screen
    for (int i = 0; i < argc; ++i)
         IBFabric::running_command += string(argv[i]) + string(" ");

    /*
     * finally parse all options
     */
    rc2 = this->p_main_parser->ParseOptions(argc, argv, true);
    if (rc2 == 1) {         //parse error and we need to dump usage
        goto usage_and_exit;
    } else if (rc2 == 2) {  //OK but need to exit
        rc = 2;
        goto exit;
    } else if (rc2 == 3) {  //error but not need to dump usage
        goto exit;
    }

    // try to construct a log file
    {
        OutputControl::Properties properties(LOG_FILE);
        if(properties.is_valid())
            if(properties.enabled()) {
                if(!OutputControl::CreateFolder(properties.path()) ||
                	contruct_log_file(properties.path().c_str()))
                {
                    SCREEN_ERR_PRINT("Failed to create log file: %s\n", properties.path().c_str());
                    IBDIAGNET_RETURN(1);
                }
            }
            else {
                SCREEN_INFO_PRINT("Log file is disabled.\n");
            }
        else {
        	SCREEN_ERR_PRINT("PrintSummary::Internal error: output properties for file %s " \
                             "cannot resolved.\n", LOG_FILE);
        }
    }

    // Print version to log and screen
    PRINT("Running version:   %s\n", IBFabric::running_version.c_str());
    PRINT("Running command:   %s\n", IBFabric::running_command.c_str());
    PRINT("Running timestamp: %s\n", IBFabric::GetTimestamp().c_str());

    SysLogger::instance().syslog("[CPU: %d] Running command: %s\n", sched_getcpu(),
                                 IBFabric::running_command.c_str());

    // Add Switch Label Port Num Explanation to log and screen
    IBFabric::GetSwitchLabelPortNumExplanation(sstr, "");
    PRINT("\n%s", sstr.str().c_str());

    /*
     * load all plugins
     */
    if ((this->not_to_load_plugins_list.empty() == true) ||
            this->not_to_load_plugins_list.front() != "all") {
        this->LoadPlugins();
    } else {
        PRINT("----------\n");
        INFO_PRINT("Plugins load will be skipped\n");
        PRINT("\n");
    }

    /*
     * parse options again with plugins
     */
    ignored_req_list.push_back(this);
    rc2 = this->p_main_parser->ParseOptions(argc, argv, false, &ignored_req_list);
    if (rc2 == 1) {         //parse error and we need to dump usage
        goto usage_and_exit;
    } else if (rc2 == 2) {  //OK but need to exit
        rc = 2;
        goto exit;
    } else if (rc2 == 3) {  //error but not need to dump usage
        goto exit;
    }

    // get options from configuration file.
    rc2 = this->ParseConfigFile();
    if (rc2 == 2) {
        rc = 2;
        goto exit;
    } else if (rc2) {
        IBDIAGNET_RETURN(3);
    }

    // handle mkey params
    if (!mkey_val.empty() || !mkey_path.empty()) {
        if (HandleMKeyParams()) {
            IBDIAGNET_RETURN(3);
        }
    }

    /*
     *  some stages need to validate params
     */
    ((TopologyMatchStage *)this->stages[STAGE_IDX_TOPOLOGY_MATCH])->ValidateTopoParams();

    /*
     * check if guid 0 was given,
     * if such case than need to present list of guid to the user
     */
    if (((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->IsNeedDisplayAllPortsGuids()) {
        u_int64_t chosen_guid = 0;
        rc2 = this->GetLocalGUIDFromUser(chosen_guid);
        if (rc2 == 1) {
            ERR_PRINT("Failed to display user GUIDs list to choose from, err=%s\n,", this->ibdiag_obj.GetLastError());
            goto exit;
        } else if (rc2 == 2) {
            rc = 2;
            goto exit;
        }
        ((DiscoveryStage *)this->stages[STAGE_IDX_DISCOVERY])->SetPortGUID(chosen_guid);
    }

    if(!this->isValidLoadFromFilePath()){
       rc =1;
       goto exit;
    }

    this->ibdiagnet_status = NOT_INITILIAZED;
    rc = 0;
exit:
    IBDIAGNET_RETURN(rc);
usage_and_exit:
    ERR_PRINT("Failed to parse options, err=%s\n\n", this->p_main_parser->GetErrDesc());
    SCREEN_PRINT("%s\n", this->p_main_parser->GetUsage().c_str());
    IBDIAGNET_RETURN(rc);
}

/* OK: return = 0 , ERROR: return != 0 */
int IBDiagapp::Init()
{
     IBDIAGNET_ENTER;

    /*
     * initialize main parser
     */
    this->p_main_parser = new CommandLineParser(this->GetToolName());
    if (!this->p_main_parser) {
        SCREEN_ERR_PRINT("Failed to allocate main parser\n");
        IBDIAGNET_RETURN(1);
    }

    //init ibdiag lib
    if (this->ibdiag_obj.Init()) {
        SCREEN_ERR_PRINT("Failed to initialize IBDiag lib\n");
        IBDIAGNET_RETURN(1);
    }

    m_lc_plugins_mgr.Init(&this->ibdiag_obj);

    for (uint8_t i = 0; i < NUM_OF_STAGES; ++i) {
        this->stages[i] = CreateStage(i);
    }

    /*
     * add all main options
     */
    char buffer[1024] = {0};

    this->AddOptions(OPTION_DEVICE, OPTION_DEVICE_SHORT, OPTION_DEVICE_ARG_DESC, OPTION_DEVICE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_PORT, OPTION_PORT_SHORT, OPTION_PORT_ARG_DESC, OPTION_PORT_DESC,
                     OPTION_PORT_DEF_VAL);
    this->AddOptions(OPTION_GUID, OPTION_GUID_SHORT, OPTION_GUID_ARG_DESC, OPTION_GUID_DESC,
                     OPTION_GUID_DEF_VAL_STR);

    // TODO: remove the next 4 lines on issue #3466238
    skip_obj_t obj2 = {SKIP_VS_CAP_SMP_VALUE, "vs_cap_smp" };
    skip_options_list.push_back(obj2);
    skip_obj_t obj3 = {SKIP_VS_CAP_GMP_VALUE, "vs_cap_gmp"};
    skip_options_list.push_back(obj3);

    skip_obj_t obj4 = {SKIP_STAGE_LINKS_VALUE, "links"};
    skip_options_list.push_back(obj4);
    skip_obj_t obj5 = {SKIP_STAGE_PM_VALUE, "pm"};
    skip_options_list.push_back(obj5);
    skip_obj_t obj6 = {SKIP_STAGE_SPEED_WIDTH_CHECK, "speed_width_check"};
    skip_options_list.push_back(obj6);
    skip_obj_t obj7 = {SKIP_STAGE_TEMPERATURE_SENSING, "temp_sensing"};
    skip_options_list.push_back(obj7);
    skip_obj_t obj8 = {SKIP_STAGE_VIRTUALIZATION, "virt"};
    skip_options_list.push_back(obj8);
    skip_obj_t obj9 = {SKIP_STAGE_ROUTERS, "routers"};
    skip_options_list.push_back(obj9);
    skip_obj_t obj = {SKIP_ALL_VALUE, "all"};
    skip_options_list.push_back(obj);

    string skip_list_str = OPTION_SKIP_DESC;
    skip_list_str += " Applicable skip stages:\n(";
    size_t skip_options_list_length = skip_options_list.size();
    for (size_t i = 0; i < skip_options_list_length - 1; ++i) {
        // TODO: remove this if on issue #3466238
        if (skip_options_list[i].value == SKIP_VS_CAP_SMP_VALUE
                || skip_options_list[i].value == SKIP_VS_CAP_GMP_VALUE)
            // do not print hidden options
            continue;
        skip_list_str += skip_options_list[i].name.c_str();
        skip_list_str += " | ";
    }
    skip_list_str += skip_options_list[skip_options_list_length - 1].name.c_str();
    skip_list_str += ").";
    this->AddOptions(OPTION_SKIP, OPTION_SKIP_SHORT, OPTION_SKIP_ARG_DESC, skip_list_str,
                     OPTION_DEF_VAL_NULL);

    list_string plugins_names;
    string errors;
    if (get_plugins_names(plugins_names, errors)) {
        SCREEN_ERR_PRINT("%s\n", errors.c_str());
        IBDIAGNET_RETURN(1);
    }

    if (!errors.empty())
        PRINT("%s\n", errors.c_str());

    string skip_plugin_list_str = OPTION_SKIP_PLUGIN_DESC;
    skip_plugin_list_str += " Applicable skip plugins:\n(";
    for (list_string::iterator it = plugins_names.begin(); it != plugins_names.end(); ++it) {
        if (it != plugins_names.begin())
            skip_plugin_list_str += " | ";
        skip_plugin_list_str += (*it).c_str();
    }
    skip_plugin_list_str += ").";
    this->AddOptions(OPTION_SKIP_PLUGIN, OPTION_SKIP_PLUGIN_SHORT, OPTION_SKIP_PLUGIN_ARG_DESC,
                     skip_plugin_list_str, OPTION_DEF_VAL_NULL);

    this->AddOptions(OPTION_PM_GET_ALL, OPTION_PM_GET_ALL_SHORT,
                     OPTION_PM_GET_ALL_ARG_DESC, OPTION_PM_GET_ALL_DESC,
                     OPTION_DEF_VAL_FALSE,IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_PM_CLEAR_ALL, OPTION_PM_CLEAR_ALL_SHORT,
                     OPTION_PM_CLEAR_ALL_ARG_DESC, OPTION_PM_CLEAR_ALL_DESC,
                     OPTION_DEF_VAL_FALSE,IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_PER_SLVL_CNTRS, OPTION_PER_SLVL_CNTRS_SHORT,
                     OPTION_PER_SLVL_CNTRS_ARG_DESC, OPTION_PER_SLVL_CNTRS_DESC,
                     OPTION_DEF_VAL_FALSE,IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_PER_SLVL_SUP_WARN, OPTION_PER_SLVL_SUP_WARN_SHORT,
                     OPTION_PER_SLVL_SUP_WARN_ARG_DESC, OPTION_PER_SLVL_SUP_WARN_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_MLNX_CNTRS, OPTION_MLNX_CNTRS_SHORT, OPTION_MLNX_CNTRS_ARG_DESC,
                     OPTION_MLNX_CNTRS_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_MLNX_CNTRS_RESET, OPTION_MLNX_CNTRS_RESET_SHORT,
                     OPTION_MLNX_CNTRS_RESET_ARG_DESC, OPTION_MLNX_CNTRS_RESET_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PC, OPTION_PC_SHORT, OPTION_PC_ARG_DESC, OPTION_PC_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_P, OPTION_P_SHORT, OPTION_P_ARG_DESC, OPTION_P_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_PMPAUSE, OPTION_PMPAUSE_SHORT, OPTION_PMPAUSE_ARG_DESC,
                     OPTION_PMPAUSE_DESC, OPTION_PMPAUSE_DEF_VAL);
    this->AddOptions(OPTION_BER, OPTION_BER_SHORT, OPTION_BER_ARG_DESC, OPTION_BER_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    // deprecated ber_use_data option
    this->AddOptions(OPTION_BERDATA, OPTION_BERDATA_SHORT, OPTION_BERDATA_ARG_DESC,
                     OPTION_BERDATA_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    snprintf(buffer, sizeof(buffer), "%llu", DEFAULT_BER_THRESHOLD);
    this->AddOptions(OPTION_BERTHRESH, OPTION_BERTHRESH_SHORT, OPTION_BERTHRESH_ARG_DESC,
                     OPTION_BERTHRESH_DESC, string(buffer));
    this->AddOptions(OPTION_LLR_ACTIVE_CELL, OPTION_LLR_ACTIVE_CELL_SHORT,
                     OPTION_LLR_ACTIVE_CELL_ARG_DESC, OPTION_LLR_ACTIVE_CELL_DESC,
                     OPTION_LLR_ACTIVE_CELL_DEF_VAL);
    this->AddOptions(OPTION_EXT_SPEEDS, OPTION_EXT_SPEEDS_SHORT, OPTION_EXT_SPEEDS_ARG_DESC,
                     OPTION_EXT_SPEEDS_DESC, OPTION_EXT_SPEEDS_DEF_VAL);
    this->AddOptions(OPTION_ALL_LANES, OPTION_ALL_LANES_SHORT, OPTION_ALL_LANES_ARG_DESC,
                     OPTION_ALL_LANES_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_LLR_INTERNAL_COUNTERS, OPTION_LLR_INTERNAL_COUNTERS_SHORT,
                     OPTION_LLR_INTERNAL_COUNTERS_ARG_DESC, OPTION_LLR_INTERNAL_COUNTERS_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_LS, OPTION_LS_SHORT, OPTION_LS_ARG_DESC, OPTION_LS_DESC,
                     IB_DISABLE_LINK_SPEED_STR);
    this->AddOptions(OPTION_LW, OPTION_LW_SHORT, OPTION_LW_ARG_DESC, OPTION_LW_DESC,
                     IB_DISABLE_LINK_WIDTH_STR);
    this->AddOptions(OPTION_TOPO_MATCH_LOCAL_SYSTEM, OPTION_TOPO_MATCH_LOCAL_SYSTEM_SHORT,
                     OPTION_TOPO_MATCH_LOCAL_SYSTEM_ARG_DESC, OPTION_TOPO_MATCH_LOCAL_SYSTEM_DESC,
                     OPTION_DEF_VAL_NULL ,IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_TOPO_COMPARE_NAMES, OPTION_TOPO_COMPARE_NAMES_SHORT,
                     OPTION_TOPO_COMPARE_NAMES_ARG_DESC, OPTION_TOPO_COMPARE_NAMES_DESC,
                     OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    // deprecated llr option
    this->AddOptions(OPTION_LLR_COUNTERS, OPTION_LLR_COUNTERS_SHORT, OPTION_LLR_COUNTERS_ARG_DESC,
                     OPTION_LLR_COUNTERS_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_LLR_SYMBOL_ERROR, OPTION_LLR_SYMBOL_ERROR_SHORT,
                     OPTION_LLR_SYMBOL_ERROR_ARG_DESC, OPTION_LLR_SYMBOL_ERROR_DESC,
                     OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SCRRENNUMERRS, OPTION_SCRRENNUMERRS_SHORT,
                     OPTION_SCRRENNUMERRS_ARG_DESC, OPTION_SCRRENNUMERRS_DESC,
                     OPTION_SCRRENNUMERRS_DEF_VAL);
    this->AddOptions(OPTION_SMP_WINDOW, OPTION_SMP_WINDOW_SHORT, OPTION_SMP_WINDOW_ARG_DESC,
                     OPTION_SMP_WINDOW_DESC, OPTION_SMP_WINDOW_DEF_VAL);
    this->AddOptions(OPTION_GMP_WINDOW, OPTION_GMP_WINDOW_SHORT, OPTION_GMP_WINDOW_ARG_DESC,
                     OPTION_GMP_WINDOW_DESC, OPTION_GMP_WINDOW_DEF_VAL);
    this->AddOptions(OPTION_MAXHOPS, OPTION_MAXHOPS_SHORT, OPTION_MAXHOPS_ARG_DESC,
                     OPTION_MAXHOPS_DESC, OPTION_MAXHOPS_DEF_VAL);
    this->AddOptions(OPTION_NO_MEPI, OPTION_NO_MEPI_SHORT, OPTION_NO_MEPI_ARG_DESC,
                     OPTION_NO_MEPI_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_DUMP_LST_WITH_LMC, OPTION_DUMP_LST_WITH_LMC_SHORT,
                     OPTION_DUMP_LST_WITH_LMC_ARG_DESC, OPTION_DUMP_LST_WITH_LMC_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_CAPABILITY_READ,
                     OPTION_CAPABILITY_READ_SHORT,
                     OPTION_CAPABILITY_READ_ARG_DESC,
                     OPTION_CAPABILITY_READ_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_CAPABILITY_WRITE, OPTION_CAPABILITY_WRITE_SHORT,
                     OPTION_CAPABILITY_WRITE_ARG_DESC, OPTION_CAPABILITY_WRITE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_FULL_CAPABILITY_WRITE,
                     OPTION_FULL_CAPABILITY_WRITE_SHORT,
                     OPTION_FULL_CAPABILITY_WRITE_ARG_DESC,
                     OPTION_FULL_CAPABILITY_WRITE_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_BACK_COMPAT_DB, OPTION_BACK_COMPAT_DB_SHORT,
                     OPTION_BACK_COMPAT_DB_ARG_DESC,
                     OPTION_BACK_COMPAT_DB_DESC, OPTION_BACK_COMPAT_DB_DEF_VAL);
    this->AddOptions(OPTION_VERSION, OPTION_VERSION_SHORT, OPTION_VERSION_ARG_DESC,
                     OPTION_VERSION_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_INFO_FLAG | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_HELP, OPTION_HELP_SHORT, OPTION_HELP_ARG_DESC, OPTION_HELP_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_INFO_FLAG | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_DEEP_HELP, OPTION_DEEP_HELP_SHORT, OPTION_DEEP_HELP_ARG_DESC,
                     OPTION_DEEP_HELP_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_INFO_FLAG | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_MLNX_HELP, OPTION_MLNX_HELP_SHORT, OPTION_MLNX_HELP_ARG_DESC,
                     OPTION_MLNX_HELP_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_INFO_FLAG | IBIS_CMD_HIDDEN | IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_MADS_TIMEOUT, OPTION_MADS_TIMEOUT_SHORT, OPTION_MADS_TIMEOUT_ARG_DESC,
                     OPTION_MADS_TIMEOUT_DESC, OPTION_MADS_TIMEOUT_DEF_VAL);
    this->AddOptions(OPTION_MADS_RETRIES, OPTION_MADS_RETRIES_SHORT, OPTION_MADS_RETRIES_ARG_DESC,
                     OPTION_MADS_RETRIES_DESC, OPTION_MADS_RETRIES_DEF_VAL);
    this->AddOptions(OPTION_MAP, OPTION_MAP_SHORT, OPTION_MAP_ARG_DESC, OPTION_MAP_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_MKEY, OPTION_MKEY_SHORT, OPTION_MKEY_ARG_DESC,
                     OPTION_MKEY_DESC, OPTION_MKEY_DEF_VAL);
    this->AddOptions(OPTION_MKEY_FILES, OPTION_MKEY_FILES_SHORT, OPTION_MKEY_FILES_ARG_DESC,
                     OPTION_MKEY_FILES_DESC, OPTION_DEF_VAL_NULL);

    this->AddOptions(OPTION_VS_KEY, OPTION_VS_KEY_SHORT, OPTION_VS_KEY_ARG_DESC,
                     OPTION_VS_KEY_DESC, DEFAULT_KEY_STR);
    this->AddOptions(OPTION_VS_KEY_FILE, OPTION_VS_KEY_FILE_SHORT, OPTION_VS_KEY_FILE_ARG_DESC,
                     OPTION_VS_KEY_FILE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_CC_KEY, OPTION_CC_KEY_SHORT, OPTION_CC_KEY_ARG_DESC,
                     OPTION_CC_KEY_DESC, DEFAULT_KEY_STR);
    this->AddOptions(OPTION_CC_KEY_FILE, OPTION_CC_KEY_FILE_SHORT, OPTION_CC_KEY_FILE_ARG_DESC,
                     OPTION_CC_KEY_FILE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_M2N_KEY, OPTION_M2N_KEY_SHORT, OPTION_M2N_KEY_ARG_DESC,
                     OPTION_M2N_KEY_DESC, DEFAULT_KEY_STR);
    //the option routers is depricated
    this->AddOptions(OPTION_ROUTERS, OPTION_ROUTERS_SHORT, OPTION_ROUTERS_ARG_DESC,
                     OPTION_ROUTERS_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_M2N_KEY_FILE, OPTION_M2N_KEY_FILE_SHORT, OPTION_M2N_KEY_FILE_ARG_DESC,
                     OPTION_M2N_KEY_FILE_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_QOS, OPTION_QOS_SHORT, OPTION_QOS_ARG_DESC, OPTION_QOS_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_FEC_MODE, OPTION_FEC_MODE_SHORT, OPTION_FEC_MODE_ARG_DESC,
                     OPTION_FEC_MODE_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_DEPRECATED);
    this->AddOptions(OPTION_CONFIG, OPTION_CONFIG_SHORT, OPTION_CONFIG_ARG_DESC,
                     OPTION_CONFIG_DESC, OPTION_DEF_VAL_NULL, IBIS_CMD_NO_CONF_FILE);

    this->AddOptions(OPTION_CONGESTION_CONTROL, OPTION_CONGESTION_CONTROL_SHORT,
                     OPTION_CONGESTION_CONTROL_ARG_DESC, OPTION_CONGESTION_CONTROL_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_CONGESTION_COUNTERS, OPTION_CONGESTION_COUNTERS_SHORT,
                     OPTION_CONGESTION_COUNTERS_ARG_DESC, OPTION_CONGESTION_COUNTERS_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_CLEAR_CONGESTION_COUNTERS, OPTION_CLEAR_CONGESTION_COUNTERS_SHORT,
                     OPTION_CLEAR_CONGESTION_COUNTERS_ARG_DESC,
                     OPTION_CLEAR_CONGESTION_COUNTERS_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_FAST_RECOVERY, OPTION_FAST_RECOVERY_SHORT,
                     OPTION_FAST_RECOVERY_ARG_DESC, OPTION_FAST_RECOVERY_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);

    this->AddOptions(OPTION_ENABLE_SWITCH_DUP_GUID_CHECK,
                     OPTION_ENABLE_SWITCH_DUP_GUID_CHECK_SHORT,
                     OPTION_ENABLE_SWITCH_DUP_GUID_CHECK_ARG_DESC,
                     OPTION_ENABLE_SWITCH_DUP_GUID_CHECK_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_DBG_MODULE, OPTION_DBG_MODULE_SHORT,
                     OPTION_DBG_MODULE_ARG_DESC, OPTION_DBG_MODULE_DESC,
                     OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_DBG_LEVEL, OPTION_DBG_LEVEL_SHORT,
                     OPTION_DBG_LEVEL_ARG_DESC, OPTION_DBG_LEVEL_DESC,
                     OPTION_DEF_VAL_NULL);
	this->AddOptions(OPTION_OUTPATH, OPTION_OUTPATH_SHORT, OPTION_OUTPATH_ARG_DESC,
                     OPTION_OUTPATH_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_ENABLE_OUTPUT, OPTION_ENABLE_OUTPUT_SHORT,
                     OPTION_ENABLE_OUTPUT_ARG_DESC, OPTION_ENABLE_OUTPUT_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_DEFAULT_OPT_ATTRIBUTES_FLAGS);
    this->AddOptions(OPTION_DISABLE_OUTPUT, OPTION_DISABLE_OUTPUT_SHORT,
                     OPTION_DISABLE_OUTPUT_ARG_DESC, OPTION_DISABLE_OUTPUT_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_DEFAULT_OPT_ATTRIBUTES_FLAGS);
    this->AddOptions(OPTION_PATH_PER_FILE, OPTION_PATH_PER_FILE_SHORT,
                     OPTION_PATH_PER_FILE_ARG_DESC, OPTION_PATH_PER_FILE_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_DEFAULT_OPT_ATTRIBUTES_FLAGS);
    this->AddOptions(OPTION_DISCOVERY_ONLY, OPTION_DISCOVERY_ONLY_SHORT,
                     OPTION_DISCOVERY_ONLY_ARG_DESC, OPTION_DISCOVERY_ONLY_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SET_SL, OPTION_SET_SL_SHORT,
                     OPTION_SET_SL_ARG_DESC, OPTION_SET_SL_DESC,
                     OPTION_SET_SL_DEF_VAL);
    this->AddOptions(OPTION_LONG_RUN_ITERATION, OPTION_LONG_RUN_ITERATION_SHORT,
                     OPTION_LONG_RUN_ITERATION_ARG_DESC, OPTION_LONG_RUN_ITERATION_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_LONG_RUN_TIMEOUT, OPTION_LONG_RUN_TIMEOUT_SHORT,
                     OPTION_LONG_RUN_TIMEOUT_ARG_DESC, OPTION_LONG_RUN_TIMEOUT_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_SYS_LOG, OPTION_SYS_LOG_SHORT,
                     OPTION_SYS_LOG_ARG_DESC, OPTION_SYS_LOG_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_SKIP_DEF, OPTION_SKIP_DEF_SHORT,
                     OPTION_SKIP_DEF_ARG_DESC, OPTION_SKIP_DEF_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_PPCC, OPTION_PPCC_SHORT, OPTION_PPCC_ARGS_DESC,
                     OPTION_PPCC_DESC, OPTION_DEF_VAL_NULL);
    this->AddOptions(OPTION_VERBS, OPTION_VERBS_SHORT, OPTION_VERBS_ARGS_DESC,
                     OPTION_VERBS_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);

    this->ibdiagnet_status = NEED_TO_GET_ARGS;

    IBDIAGNET_RETURN(0);
}

int IBDiagapp::PostInit()
{
    IBDIAGNET_ENTER;

    if (this->ibdiagnet_status == NEED_TO_GET_ARGS) {
        ERR_PRINT("Need to get args from user before initialize\n");
        IBDIAGNET_RETURN(1);
    }
    if (this->ibdiagnet_status == READY) {
        ERR_PRINT("Already initialized\n");
        IBDIAGNET_RETURN(1);
    }

    //open db_csv file
    string err_message;
    int rc = csv_out.Open(DB_FILE, err_message);
    if (rc) {
        IBDIAGNET_LOG(TT_LOG_LEVEL_ERROR, "%s", err_message.c_str());
        ERR_PRINT("Failed to open database file %s for writing (%s)\n",
        	      csv_out.GetFileName().empty() ? "'Invalid Database filename'" : csv_out.GetFileName().c_str(),
        	      err_message.c_str());
        ibdiag_obj.AddGeneratedFile("Database", "Invalid Database filename'" );
        IBDIAGNET_RETURN(1);
    }
    else
        if (!csv_out.GetFileName().empty())
            ibdiag_obj.AddGeneratedFile("Database", csv_out.GetFileName());

    //set some stuff in all stages
    for (int i = 0; i < NUM_OF_STAGES; ++i) {
        if(!this->stages[i])
            continue;
        this->stages[i]->SetNumErrsToScreen(&this->num_of_errors_to_screen);
        this->stages[i]->SetBasePath(&this->base_path);
        this->stages[i]->SetCSVOut(&this->csv_out);
    }
    for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
            it != this->loaded_plugins_list.end(); ++it) {
        (*it)->p_plugin_obj->SetNumErrsToScreen(&this->num_of_errors_to_screen);
        (*it)->p_plugin_obj->SetBasePath(&this->base_path);
        (*it)->p_plugin_obj->SetCSVOut(&this->csv_out);
    }

    ibdiag_obj.GetIbisPtr()->SetMaxMadsOnWire((u_int16_t)m_max_gmps_on_wire,
                                              (u_int8_t)m_max_smps_on_wire);

    this->ibdiagnet_status = READY;
    IBDIAGNET_RETURN(0);
}

void IBDiagapp::TimeoutActivation(struct timeval *iteration_time)
{
    IBDIAGNET_ENTER;

    u_int64_t iteration_time_mic_sec = (u_int64_t)((iteration_time->tv_sec * 1000000) +
                                        iteration_time->tv_usec);

    if (iteration_time_mic_sec >= this->GetTimeoutIteration())
        IBDIAGNET_RETURN_VOID;

    u_int64_t time_left_mic_sec = this->GetTimeoutIteration() - iteration_time_mic_sec;
    u_int64_t time_left_sec     = time_left_mic_sec / 1000000;
              time_left_mic_sec %= 1000000;
    u_int64_t time_left_mil_sec = time_left_mic_sec / 1000;
    INFO_PRINT("Going to sleep for %" PRIu64 " seconds %"  PRIu64 " mili seconds until next iteration\n",
            time_left_sec, time_left_mil_sec);

    sleep((unsigned int)time_left_sec);
    usleep((unsigned int)time_left_mic_sec);

    IBDIAGNET_RETURN_VOID;
}

int IBDiagapp::RunStages(int from, int to)
{
    IBDIAGNET_ENTER;

    for (int i = from; i < to; ++i) {
        if(!this->stages[i])
            continue;

        if (this->RunSingleStage(this->stages[i]))
            IBDIAGNET_RETURN(1);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagapp::RunSingleStage(Stage* stage)
{
    IBDIAGNET_ENTER;

    int status = IBDIAG_SUCCESS_CODE;

    bool need_to_run = !stage->IsSkipped();

    if (this->ibdiag_obj.curr_control_session) {
        if (this->ibdiag_obj.ControlIsStageEnabled(this->ibdiag_obj.curr_control_session,
                stage->GetName().c_str())) {
            need_to_run = true;
        } else {
            need_to_run = false;
            stage->CleanResources();
        }
    }

    if (need_to_run) {

        struct timespec began_tm;
        struct timespec end_tm;

        struct rusage   began_usage;
        struct rusage   end_usage;

        struct timespec & real  = stage->m_perf_time_info.real;
        struct timeval  & user  = stage->m_perf_time_info.user;
        struct timeval  & sys   = stage->m_perf_time_info.sys;

        clock_gettime(CLOCK_REALTIME, &began_tm);
        getrusage(RUSAGE_SELF, &began_usage);

        this->ibdiag_obj.GetIbisPtr()->m_mads_stat.start(stage->GetName());

        if (!(status = stage->CleanResources()))
            if (!(status = stage->Prepare()))
                if (!(status = stage->RetrieveInfo()))
                    if (!(status = stage->RunCheck()))
                        stage->MarkStageAsFinished();

        this->ibdiag_obj.GetIbisPtr()->m_mads_stat.stop();

        getrusage(RUSAGE_SELF, &end_usage);
        clock_gettime(CLOCK_REALTIME, &end_tm);

        diff(end_tm, began_tm, real);
        diff(end_usage.ru_utime, began_usage.ru_utime, user);
        diff(end_usage.ru_stime, began_usage.ru_stime, sys);

        SysLogger::instance().syslog("Stage \"%s\" real: %lld.%.6ld user: %ld.%.6ld, sys: %ld.%.6ld\n",
                                     stage->GetName().c_str(),
                                     (long long)real.tv_sec, (long)(real.tv_nsec / 1000),
                                     user.tv_sec, user.tv_usec,
                                     sys.tv_sec, sys.tv_usec);

        m_stage_info_csv << stage->GetName() << std::setfill('0') << ','
                         << real.tv_sec << '.' << std::setw(6) << (real.tv_nsec / 1000) << ','
                         << user.tv_sec << '.' << std::setw(6) << user.tv_usec << ','
                         << sys.tv_sec << '.' << std::setw(6) << sys.tv_usec << '\n';
    }

    IBDIAGNET_RETURN(status);
}

int IBDiagapp::RunAllStages()
{
    IBDIAGNET_ENTER;

    int rc = 1;
    const u_int64_t max_iterations = this->GetMaxIterations();

    if (this->ibdiagnet_status != READY) {
        ERR_PRINT("Not initialized ibdiagnet class\n");
        IBDIAGNET_RETURN(rc);
    }

    m_stage_info_csv << "Name,Real,User,Sys\n";

    if (this->RunStages(0, STAGE_IDX_POST_DISCOVERY))
        goto exit;

    for (u_int64_t iteration = 0; (max_iterations == 0) || (iteration < max_iterations);
         iteration++) {
         struct timeval start_iteration;
         gettimeofday(&start_iteration, NULL);

         this->ibdiag_obj.curr_iteration = iteration;

         this->ibdiag_obj.curr_control_session = this->ibdiag_obj.ControlOpenSession(iteration);
         if (this->ibdiag_obj.HaveControlAPI() && !this->ibdiag_obj.curr_control_session)
             break;

         control_scope_t scope = {0};
         if (this->ibdiag_obj.ControlGetScope(this->ibdiag_obj.curr_control_session, &scope)) {
             if (this->ibdiag_obj.GetDiscoverFabricPtr()->markInScopeNodes(&scope))
                 goto exit;
         }

         if (this->RunStages(STAGE_IDX_POST_DISCOVERY, NUM_OF_PRE_PLUGIN_STAGES))
             goto exit;

         for (list_p_symbols_plugin::iterator it = this->loaded_plugins_list.begin();
                 it != this->loaded_plugins_list.end(); ++it) {
             try {
                 Stage *stage = static_cast<Stage*>((*it)->p_plugin_obj);
                 /*
                  * IsActive() method of PHY plugin also activate it self
                  * if rail validation is active.
                  * Casting to void is to avoid "CHECKED_RETURN" coverity issue.
                  * This should be removed when plugins will be merged to core.
                  */
                 (void)stage->IsActive();

                 if (this->RunSingleStage(stage))
                     goto exit;
             }
             catch(...) {            // Handle all exceptions
                 ERR_PRINT("Plugin %s raised an exception\n",
                           (*it)->library_file_name.c_str());
                 goto exit;
             }
         }
         if (this->RunStages(NUM_OF_PRE_PLUGIN_STAGES, NUM_OF_STAGES))
             goto exit;

         // Activate timeout only if we are not in last iteration
         if ((max_iterations == 0) || (iteration != max_iterations - 1)) {
             struct timeval end_iteration;
             gettimeofday(&end_iteration, NULL);
             struct timeval sub_time;
             timersub(&end_iteration, &start_iteration, &sub_time);
             this->TimeoutActivation(&sub_time);
         }

         bool need_to_continue = this->ibdiag_obj.ControlCloseSession(
                 this->ibdiag_obj.curr_control_session, 0);
         if (!need_to_continue)
             break;
    }

    rc = 0;
exit:
    this->ibdiag_obj.ControlCloseSession(this->ibdiag_obj.curr_control_session, rc);

    if (!csv_out.DumpStart(SECTION_STAGE_PERF_INFO)) {
        csv_out.WriteBuf(m_stage_info_csv.str());
        csv_out.DumpEnd(SECTION_STAGE_PERF_INFO);
    }

    if (!csv_out.DumpStart(SECTION_IBIS_MAD_INFO)) {
        std::stringstream ss;
        this->ibdiag_obj.GetIbisPtr()->m_mads_stat.output_all_records_csv(ss);
        csv_out.WriteBuf(ss.str());
        csv_out.DumpEnd(SECTION_IBIS_MAD_INFO);
    }

    if (!csv_out.DumpStart(SECTION_IBIS_PERF_INFO)) {
        std::stringstream ss;
        this->ibdiag_obj.GetIbisPtr()->m_mads_stat.output_all_perf_records_csv(ss);
        csv_out.WriteBuf(ss.str());
        csv_out.DumpEnd(SECTION_IBIS_PERF_INFO);
    }

    if (this->ibdiag_obj.WriteIbisStatFile(IBIS_STAT_FILE))
        rc = 1;

    this->PrintFabricSummaryTable();
    this->PrintSummary();
    PRINT("\n");
    if (rc)
        ERR_PRINT("A fatal error occurred, exiting...\n");
    IBDIAGNET_RETURN(rc);
}

int IBDiagapp::Run(int argc, const char **argv)
{

    IBDIAGNET_ENTER;
    //assert to make sure number of elements in bad_direct_route_reasons equal
    //to IBDIAG_BAD_DR_MAX
    COMPILE_TIME_ASSERT(N_ELEMENTS(bad_direct_route_reasons) ==
        IBDIAG_BAD_DR_MAX);

    //init
    if (this->Init()) {
        SCREEN_ERR_PRINT("%s failed to initialize, exiting...\n", argv[0]);
        IBDIAGNET_RETURN(1);
    }

    // set IBNL default directory
    SetIBNLDir(IBNL_OUTPUT_DIR);

    //get args
    int rc = this->GetOpt(argc, argv);
    if (rc) {
        if (rc == 2)            //2 - OK but need to exit
            IBDIAGNET_RETURN(0);
        IBDIAGNET_RETURN(1);
    }

    // get base path for old open file mechanism
    if(!OutputControl::instance().pathes().get("default", this->base_path).is_valid()) {
        SCREEN_ERR_PRINT("Internal error: Cannot retrieve output properties for base_path");
        IBDIAGNET_RETURN(1);
    }

    // trim
    string::size_type index = this->base_path.find_last_of('/');
    if(index != string::npos)
       this->base_path.erase(index + 1);

    m_debug_wraper.OpenFile();
    this->OpenPcapFile();

    //init
    if (this->PostInit()) {
        ERR_PRINT("%s failed to initialize, exiting...\n", argv[0]);
        IBDIAGNET_RETURN(1);
    }

    //Write START_INFO section into db_csv
    rc = DumpStartInfoCSVTable(argc, argv);
    if (rc)
        IBDIAGNET_RETURN(1);

    // Parse files after all flags are parsed
    rc = this->PostParseFiles();
    if (rc) {
        if (rc == 2)            //2 - OK but need to exit
            IBDIAGNET_RETURN(0);
        IBDIAGNET_RETURN(1);
    }

    //run all
    if (this->RunAllStages())
        IBDIAGNET_RETURN(1);

    csv_out.Close();
    IBDIAGNET_RETURN(0);

}

