/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#ifndef _CABLE_DIAG_H_
#define _CABLE_DIAG_H_

#include <stdio.h>
#include <stdlib.h>

#include <vector>
using namespace std;

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include "../internal_packets/internal_packets_layouts.h"
#include "cable_diag_types.h"
#include "infiniband/ibdm/cable/CableRecord.h"

#define FLT_FMT     "%.3f"

/******************************************************/
/******************************************************/
/******************************************************/
class CableInfo : public CableRecord {
private:

    CableInfo_Payload_Page_E9_Addr_128_175 cableinfo_E9_128_175;
    CableInfo_Payload_Page_E9_Addr_176_211 cableinfo_E9_176_211;

    IBPort *p_port;

    bool IsValidDateCode(char *p_date_code);
    string ConvertPageE9Addr128ToStr();
    string ConvertPageE9Addr176ToStr();
public:
    /*
    CableInfo(IBPort *p_port,
            u_int8_t *eeprom_payload_128_167,
            u_int8_t *eeprom_payload_168_211);
            */

    CableInfo(IBPort *port): CableRecord(),
        cableinfo_E9_128_175(), cableinfo_E9_176_211(),
        p_port(port){}

    ~CableInfo() { return; }

    int SetCableInfo(u_int8_t not_supported_status,
                     u_int8_t *eeprom_payload,
                     u_int8_t eeprom_index,
                     u_int8_t page_num,
                     IBPort *p_port,
                     list_p_fabric_general_err *p_cable_errors);
    int ExportToIBPort();

    string csv_str();
    static string hdr_str();

    string csv_str_db_1();

    int GetCableType() {
        return cable_type;
    }

    inline u_int8_t GetConnector() {return connector;}
    inline string GetOUI() {return oui;}
    inline u_int8_t GetMlnxVendorByte() {return mlnx_vendor_byte;}
    bool IsCMISCable();
};


typedef struct port_cable_data {
    IBPort *p_port;
    SMP_EyeOpen *eye_open[MAX_GROUP_LANES];
    CableInfo *cable_info;
} port_cable_data_t;


typedef struct cable_data {
    port_cable_data_t data_per_port[2];
    int app_data;
} cable_data_t;

/**
 * Function pointer to check the cables indication before sending the page.
 * All indication are listed in section 5.1 in ModuelInfo spec.
 */
typedef bool (*func_get_condition_t) (CableInfo *cable_info);
typedef struct cable_info_address {

    u_int8_t page_num;
    u_int8_t address;
    u_int8_t addr_size;
    u_int32_t password;
    func_get_condition_t func_get_condition;
    cable_info_address(u_int8_t page, u_int8_t addr, u_int8_t size, u_int32_t pw,
                       func_get_condition_t func_get_cond):
                       page_num(page), address(addr), addr_size(size), password(pw),
                       func_get_condition(func_get_cond) {}
} cable_info_address_t;

typedef vector < cable_data_t * > vec_p_cable_data;
typedef list < cable_info_address_t * > list_p_cable_info_addr_t;

/******************************************************/
/******************************************************/
/******************************************************/
class CableDiag : public Plugin {
private:
    ////////////////////
    //members
    ////////////////////
    IBFabric *p_discovered_fabric;
    Ibis *p_ibis_obj;
    CapabilityModule *p_capability_module;
    vec_p_cable_data cables_vector;

    u_int16_t eye_open_bound_min_thresh;
    u_int16_t eye_open_bound_max_thresh;
    bool to_check_eye_bounds;
    bool to_get_eye_open_info;
    bool to_dump_eye_expert_file;
    bool to_get_cable_info;
    bool to_get_disconnected_port_info;

    int m_ClbckErrorState;
    list_p_fabric_general_err *p_cable_errors;

    list_p_cable_info_addr_t  cable_info_addr_list_by_phase[MAX_PHASES];
    ////////////////////
    //methods
    ////////////////////
    /*
     * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int EyeOpenGetByLid(u_int16_t lid,
            u_int8_t port_num, u_int8_t group_lane,
            struct SMP_EyeOpen *p_smp_eye_open,
            const clbck_data_t *p_clbck_data = NULL);
    int EyeOpenGetByDirect(direct_route_t *p_direct_route,
            u_int8_t port_num, u_int8_t group_lane,
            struct SMP_EyeOpen *p_smp_eye_open,
            const clbck_data_t *p_clbck_data = NULL);
    int CableInfoGetByLid(IN u_int16_t lid,
            IN u_int8_t port_num,
            IN u_int8_t addr, IN u_int8_t size, IN u_int8_t page_num,
            OUT struct SMP_CableInfo *p_smp_cable_info,
            OUT u_int8_t& vs_status,
            const clbck_data_t *p_clbck_data = NULL);
    int CableInfoGetByDirect(IN direct_route_t *p_direct_route,
            IN u_int8_t port_num,
            IN u_int8_t addr, IN u_int8_t size, IN u_int8_t page_num,
            IN u_int32_t password,
            OUT struct SMP_CableInfo *p_smp_cable_info,
            OUT u_int8_t& vs_status,
            const clbck_data_t *p_clbck_data = NULL);

    //Returns: IBDIAG_SUCCESS_CODE / IBDIAG_ERR_CODE_DB_ERR
    int MarkAllPortsNotVisited(OUT u_int32_t& max_ports_per_node);

    //Returns: IBDIAG_SUCCESS_CODE / IBDIAG_ERR_CODE_NO_MEM / IBDIAG_ERR_CODE_DB_ERR
    int AddSmpEyeOpen(IBPort *p_port, IBPort *p_rem_port,
            SMP_EyeOpen& smp_eye_open, u_int8_t group_lane);
    //Returns: IBDIAGNET_SUCCESS_CODE / IBDIAGNET_ERR_CODE_NO_MEM / IBDIAGNET_ERR_CODE_DB_ERR
    int GetSMPCableInfo(IBPort *p_port, IBPort *p_rem_port,
                        CableInfo *& p_cable_info);

    //int AddSMPCableInfo(IBPort *p_port, IBPort *p_rem_port,
    //        SMP_CableInfo& smp_cable_info_128_167, SMP_CableInfo& smp_cable_info_168_211);

    void GetDirectRoute(IBNode *p_node, IBPort *p_port, direct_route_t **p_direct_route);
    //Returns: IBDIAG_SUCCESS_CODE / IBDIAG_ERR_CODE_DB_ERR / IBDIAG_ERR_CODE_NO_MEM / IBDIAG_ERR_CODE_FABRIC_ERROR
    int BuildEyeOpenDB(list_p_fabric_general_err &cable_errors);
    //TODO - reports eye errors more user frirndly with cable info
    int BuildCableInfoDB(list_p_fabric_general_err &cable_errors,
                         u_int8_t phase, u_int32_t max_ports_per_node);

    //Returns: IBDIAG_SUCCESS_CODE / IBDIAG_ERR_CODE_CHECK_FAILED / IBDIAG_ERR_CODE_NO_MEM
    int CheckEyeBoundSum(list_p_fabric_general_err &cable_errors);
    //TODO - add method to check cable info

    void DumpCSVCablesInfo(CSVOut &csv_out);
    void DumpCSVEyeOpenInfo(CSVOut &csv_out);
    void DumpCablesInfo(ofstream &sout);
    void DumpEyeOpenInfo(ofstream &sout);

    //Returns: SUCCESS_CODE / ERR_CODE_IO_ERR
    int WriteCableFile(const string &file_name);
    int WriteEyeExpertFile(const string &file_name);

    void CreatePagesList();

    static bool IsActiveCableActiveModule(CableInfo *p_cable_info);
    static bool IsMlnxMMFMlnxPSM(CableInfo *p_cable_info);
public:
    ////////////////////
    //methods
    ////////////////////
    CableDiag(IBDiag *p_ibdiag);
    ~CableDiag();

    inline void SetEyeOpenBoundMinThresh(u_int16_t threshold) {
        this->eye_open_bound_min_thresh = threshold;
    }
    inline void SetEyeOpenBoundMaxThresh(u_int16_t threshold) {
        this->eye_open_bound_max_thresh = threshold;
    }
    inline void SetEyeOpenBoundCheck() {
        this->to_check_eye_bounds = true;
    }
    inline void SetEyeOpenInfoGet() {
        this->to_get_eye_open_info = true;
        this->MarkStageAsActive();
    }
    inline void SetDumpEyeExpertFile() {
        this->to_dump_eye_expert_file = true;
    }
    inline void SetCableInfoGet(bool set_val) {
        this->to_get_cable_info = set_val;
        this->MarkStage(set_val);
    }
    inline void SetQueryDisconnectedPorts(bool set_val) {
        this->to_get_disconnected_port_info = set_val;
    }

    void Init() {}
    int Prepare();
    int RetrieveInfo();
    int RunCheck();

    //Returns: 0 - OK / 1 - error / 2 - OK but need to exit
    int HandleOption(string name, string value);

    // CableDiag Clbck

    void EyeOpenGetClbck(const clbck_data_t &clbck_data,
                         int rec_status,
                         void *p_attribute_data);

    void CableInfoGetClbck(const clbck_data_t &clbck_data,
                           int rec_status,
                           void *p_attribute_data);

    virtual int CleanResources();
};


/******************************************************/
/******************************************************/
/******************************************************/
// the class factories
extern "C" {

Plugin * create(IBDiag *p_ibdiag)
{
    return(new CableDiag(p_ibdiag));
}

void destroy(Plugin *p)
{
    delete (CableDiag *)p;
    return;
}

}


#endif          /* not defined _CABLE_DIAG_H_ */
