/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef _ACC_REGISTER_H_
#define _ACC_REGISTER_H_

#include "acc_reg_key.h"
#include "phy_diag_types.h"
#include "phy_export_api.h"

class PhyDiag;

union AccessRegister {
    struct SMP_AccessRegister smp_reg;
    struct GMP_AccessRegister gmp_reg;
};


typedef union
{
    struct pcam_reg pcam;

    struct slrg_reg slrg;
    struct sllm_reg sllm;
    struct slrp_reg slrp;
    struct sltp_reg sltp;
    struct slsir_reg slsir;
    struct ptas_reg ptas;
    struct pmdr_reg pmdr;
    struct pphcr_reg pphcr;
    struct ppll_reg ppll;
    struct mpein_reg mpein;
    struct mpir_reg mpir;
    struct pmcr_reg pmcr;
    struct ptys_reg ptys;

    // fans registers
    struct mfcr_reg mfcr;
    struct mfsm_reg mfsm;
    struct mfsl_reg mfsl;
    struct fore_reg fore;
    struct mfnr_reg mfnr;
    // temperature registers
    struct mtcap_reg mtcap;
    struct mtmp_reg mtmp;
    struct mtwe_reg mtwe;
    // current/voltage register
    struct mvcap_reg mvcap;
    struct mvcr_reg mvcr;
    // power supplies register
    struct msps_reg msps;
    // general information register
    struct msgi_reg msgi;

    struct slrip_reg slrip;
    struct slreg_reg slreg;
    struct slcct_reg slcct;
    struct ppamp_reg ppamp;
    struct peucg_reg peucg;
    struct pddr_reg pddr;
    struct pemi_reg pemi;
    struct mpcnt_reg mpcnt;

    struct prtl_reg prtl;
    struct ppbmp_reg ppbmp;
    struct mrcs_reg mrcs;

    // PDDR Layouts
    DDModuleInfo        pddr_module;
    DDLatchedFlagInfo   pddr_latched;
} reg_list;

typedef struct
{
    uint8_t raw[sizeof(reg_list) + 160];
    union {
        struct {
            uint8_t module_info_ext;
        }; //here you may add structures to keep data related to different
        // registers
        // struct {
        //         uint32_t xxxxxxx;
        //         uint32_t zzzzzzz;
        //}
    };
} reg_raw_data;

struct acc_reg_data {
    //This union should help access the data without pointer casting
    //and this fixes the strict aliasing rule error/warning.
    union {
        reg_raw_data data;
        reg_list regs;
    };
};

typedef map < AccRegKey *, acc_reg_data, bool (*)(AccRegKey *, AccRegKey *)> map_akey_areg;
typedef map < IBNode *, u_int8_t > map_pnode_to_max_group_num_t;
class AccRegHandler;

bool keycomp(AccRegKey *k1, AccRegKey *k2);


//Register class defines common members to all registers.
//Each register needs to have its dedicated class, that inherits Register
//and defines its own PackData and DumpRegisterData.
class Register {

protected:
    PhyDiag   * m_phy_diag;
    uint32_t    m_register_id;
    uint32_t    m_fields_num;
    //That is a bit to set in case register isn't supported
    //and to avoid sending the register more than once to the same node.
    uint64_t    m_not_supported_bit;
    string      m_section_name;
    string      m_header;
    string      m_name;
    PhyPluginSupportedNodesType m_support_nodes;
    bool        m_retrive_disconnected;
    bool        m_dump_enabled;
    AccRegVia_t m_acc_reg_type;
    AccRegVia_t m_acc_type_priority;

    unpack_data_func_t unpack_data_func;

public:
    Register(PhyDiag * phy_diag,
             uint32_t reg_id,
             const unpack_data_func_t p_unpack_func,
             string sn,
             string name,
             uint32_t fn,
             uint64_t nsb,
             string header = "",
             PhyPluginSupportedNodesType support_nodes = SUPPORT_SW_CA,
             bool dump_enabled = true,
             bool retrive_disconnected = false,
             AccRegVia_t acc_reg_type = VIA_SMP,
             AccRegVia_t acc_type_priority = VIA_GMP);

    virtual ~Register(){}

    int PackDataSMP(AccRegKey *p_key, SMP_AccessRegister *acc_reg, const IBNode *p_node);
    int PackDataGMP(AccRegKey *p_key, GMP_AccessRegister *acc_reg, const IBNode *p_node);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) = 0;

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) {
        IBDIAGNET_ENTER;
        unpack_data_func(data_to_unpack, unpacked_buffer);
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    bool IsEnabledByPCAM(const IBNode* p_node) const;

    inline PhyDiag *        GetPhyDiag()                    { return m_phy_diag;                }

    inline uint64_t         GetNotSupportedBit() const      { return m_not_supported_bit;       }
    inline const string&    GetName() const                 { return m_name;                    }
    inline const string&    GetHeader() const               { return m_header;                  }
    inline const string&    GetSectionName() const          { return m_section_name;            }
    inline uint32_t         GetRegisterID() const           { return m_register_id;             }
    inline uint32_t         GetFieldsNum() const            { return m_fields_num;              }
    inline bool             GetDumpEnabled() const          { return m_dump_enabled;            }
    inline AccRegVia_t      GetRegisterType() const         { return m_acc_reg_type;            }
    inline AccRegVia_t      GetRegisterPriority() const     { return m_acc_type_priority;       }

    inline bool             GetRetriveDisconnected() const  { return m_retrive_disconnected;    }
    inline void             SetRetriveDisconnected(bool retrive_disconnected) {
        this->m_retrive_disconnected = retrive_disconnected;
    }

    inline bool IsRegSupportNodeType(const IBNode* type);

    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header);

    virtual void DumpRegisterData(const struct acc_reg_data &, stringstream &sstream, const AccRegKey &key) const = 0;
    virtual void GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices);
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors);
    static int SensorsBitsToList(u_int64_t sensors, u_int8_t max_sensors, list_uint8& sensors_list);
    static int SensorsCountToList(u_int8_t sensor_count, list_uint8& sensors_list);

    virtual int BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar);

    virtual bool IsValidPakcet(const struct acc_reg_data & reg_data) const { return true; }

    int HandleNodeNotSupportAccReg(PhyDiag *p_phy_diag,
                                   IBNode* p_node,
                                   uint64_t not_supported_bit);

    virtual AccRegVia_t Validation(IBNode *p_node, int &rc); // todo - change to abstract

    virtual void HandleData(IBNode* p_node, AccRegKey* p_key, acc_reg_data& reg_data) const { }

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);

    virtual bool CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                  export_data_phy_node_t *p_export_data_phy_node,
                                  AccRegKey *p_key);
};


class PCAMRegister : public Register {
public:
    uint8_t m_access_reg_group;
    uint8_t m_feature_group;

public:
    PCAMRegister(PhyDiag * phy_diag);
    virtual ~PCAMRegister() {}

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const;
    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);

    virtual AccRegVia_t Validation(IBNode *p_node, int &rc);
};


class SLRegister : public Register {
private:
    map_akey_areg *p_mpein_map; // for PCI mode only (pnat = 3)
    map_akey_areg *p_mpir_map;  // for PCI mode only (pnat = 3)

public:
    SLRegister(PhyDiag * phy_diag,
               uint32_t reg_id,
               const unpack_data_func_t p_unpack_func,
               string sn,
               string name,
               uint32_t fn,
               uint64_t nsb,
               map_akey_areg *mpein_map,
               map_akey_areg *mpir_map);
    virtual ~SLRegister(){}

    virtual void DumpRegisterData(const struct acc_reg_data &, stringstream &sstream, const AccRegKey &key) const = 0;

    virtual int BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar);

    virtual bool CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                  export_data_phy_node_t *p_export_data_phy_node,
                                  AccRegKey *p_key);
};

class SLTPRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLTPRegister(PhyDiag * phy_diag,
                     u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLTP_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLTPRegister(){}


        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        void Header_Dump_40nm_28nm(stringstream &sstream) const;
        void Header_Dump_16nm(stringstream &sstream) const;
        void Header_Dump_7nm(stringstream &sstream) const;
        void Header_Dump_5nm(stringstream &sstream) const;

        void Dump_40nm_28nm(const struct sltp_reg &sltp, stringstream &sstream) const;
        void Dump_16nm(const struct sltp_reg &sltp, stringstream &sstream) const;
        void Dump_7nm(const struct sltp_reg &sltp, stringstream &sstream) const;
        void Dump_5nm(const struct sltp_reg &sltp, stringstream &sstream) const;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLRGRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLRGRegister(PhyDiag * phy_diag,
                     u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLRG_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLRGRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;


        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        void Dump_40nm_28nm(const struct slrg_reg &slrg, stringstream &sstream) const;
        void Dump_16nm(const struct slrg_reg &slrg, stringstream &sstream) const;
        void Dump_7nm(const struct slrg_reg &slrg, stringstream &sstream) const;

        void Header_Dump_40nm_28nm(stringstream &sstream) const;
        void Header_Dump_16nm(stringstream &sstream) const;
        void Header_Dump_7nm_5nm(stringstream &sstream) const;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLLMRegister : public SLRegister {
private:
        u_int8_t    m_pnat;
public:
    SLLMRegister(PhyDiag * phy_diag,
                 u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                 string sn = ACC_REG_SLLM_INTERNAL_SECTION_NAME,
                 map_akey_areg *mpein_map = NULL,
                 map_akey_areg *mpir_map = NULL);

    virtual ~SLLMRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

    void Dump_16nm(const struct sllm_reg &sllm, stringstream &sstream) const;
    void Dump_7nm(const struct sllm_reg &sllm, stringstream &sstream) const;

    void Header_Dump_16nm(stringstream &sstream);
    void Header_Dump_7nm(stringstream &sstream);

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLRPRegister : public SLRegister {
private:
        u_int8_t    m_pnat;

public:
        SLRPRegister(PhyDiag * phy_diag,
                     u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                     string sn = ACC_REG_SLRP_INTERNAL_SECTION_NAME,
                     map_akey_areg *mpein_map = NULL,
                     map_akey_areg *mpir_map = NULL);
        virtual ~SLRPRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        void Header_Dump_40nm_28nm(stringstream &sstream) const;
        void Header_Dump_16nm(stringstream &sstream) const;
        void Header_Dump_7nm(stringstream &sstream) const;
        void Header_Dump_5nm(stringstream &sstream) const;

        void Dump_40nm_28nm(const struct slrp_reg &slrp, stringstream &sstream) const;
        void Dump_16nm(const struct slrp_reg &slrp, stringstream &sstream) const;
        void Dump_7nm(const struct slrp_reg &slrp, stringstream &sstream) const;
        void Dump_5nm_scc(const struct slrp_reg &slrp, stringstream &sstream) const;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLSIRRegister : public SLRegister {
private:
    u_int8_t    m_pnat;

public:
    SLSIRRegister(PhyDiag * phy_diag,
                  u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
                  string sn = ACC_REG_SLSIR_INTERNAL_SECTION_NAME,
                  map_akey_areg *mpein_map = NULL,
                  map_akey_areg *mpir_map = NULL);

    virtual ~SLSIRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

    virtual AccRegVia_t Validation(IBNode *p_node, int &rc) override;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class PTASRegister : public Register {
public:

        PTASRegister(PhyDiag * phy_diag);
        virtual ~PTASRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};

class PMDRRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
        PMDRRegister(PhyDiag * phy_diag);
        virtual ~PMDRRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class PMCRRegister : public Register {
public:
        PMCRRegister(PhyDiag * phy_diag);
        virtual ~PMCRRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class PTYSRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
        PTYSRegister(PhyDiag * phy_diag);
        virtual ~PTYSRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class PPHCRRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
        PPHCRRegister(PhyDiag * phy_diag);
        virtual ~PPHCRRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

// PEUCG - UPHY
class PEUCGRegister : public Register {
protected:
        u_int8_t           m_pnat;
        UPHY::MadBuilder*  m_cur_mad_builder;

public:
        PEUCGRegister(PhyDiag * phy_diag, string section_name);
        virtual ~PEUCGRegister(){}

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
};

class PEUCG_Ver_Register : public PEUCGRegister {
public:
        PEUCG_Ver_Register(PhyDiag * phy_diag);
        virtual ~PEUCG_Ver_Register(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
        virtual void HandleData(IBNode* p_node, AccRegKey* p_key, acc_reg_data& reg_data) const;
};

class PEUCG_DLN_Register : public PEUCGRegister {

public:
        PEUCG_DLN_Register(PhyDiag * phy_diag);
        virtual ~PEUCG_DLN_Register(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);

        virtual int BuildDB(AccRegHandler* p_handler,
                            list_p_fabric_general_err &phy_errors,
                            ProgressBar *progress_bar);
};

class PEUCG_CLN_Register : public PEUCGRegister {

    map_akey_areg *p_pmdr_map;

public:
    PEUCG_CLN_Register(PhyDiag * phy_diag, map_akey_areg *pmdr_map);
    virtual ~PEUCG_CLN_Register(){}

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);

    virtual int BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar);
};

class PPLLRegister : public Register {
public:

        PPLLRegister(PhyDiag * phy_diag);
        virtual ~PPLLRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

        void Header_Dump_28nm(stringstream &sstream) const;
        void Header_Dump_16nm(stringstream &sstream) const;
        void Header_Dump_7nm(stringstream &sstream) const;
        void Header_Dump_5nm(stringstream &sstream) const;

        void Dump_28nm(const struct ppll_reg &ppll, stringstream &sstream) const;
        void Dump_16nm(const struct ppll_reg &ppll, stringstream &sstream) const;
        void Dump_7nm(const struct ppll_reg &ppll, stringstream &sstream) const;
        void Dump_5nm(const struct ppll_reg &ppll, stringstream &sstream) const;

        void Dump_pll_28nm(const struct pll_28nm &pll_28, stringstream &sstream) const;
        void Dump_pll_16nm(const struct pll_status_data &pll_16, stringstream &sstream) const;

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class MPEINRegister : public Register {
public:
        MPEINRegister(PhyDiag * phy_diag);
        virtual ~MPEINRegister(){}

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

        virtual int BuildDB(AccRegHandler* p_handler,
                            list_p_fabric_general_err &phy_errors,
                            ProgressBar *progress_bar) override;

        virtual bool IsValidPakcet(const struct acc_reg_data & reg_data) const override;

        int LoopPCIDepth(const clbck_data_t &clbck_data,
                         int rec_status,
                         void *p_attribute_data);

        int LoopPCINode(const clbck_data_t &clbck_data,
                        int rec_status,
                        void *p_attribute_data);

        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;

        virtual bool CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      AccRegKey *p_key) override;
};

class MPIRRegister : public Register {

    map_akey_areg &mpein_map;

public:
    MPIRRegister(PhyDiag * phy_diag, map_akey_areg &mpein_map);
    virtual ~MPIRRegister(){}

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;

    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;

    virtual int BuildDB(AccRegHandler* p_handler, // handler containing the source map of mpein
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar) override;
};


// fans registers
class MFCRRegister : public Register {
public:
    MFCRRegister(PhyDiag * phy_diag);
    virtual ~MFCRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MFSMRegister : public Register {
public:
    MFSMRegister(PhyDiag * phy_diag);
    virtual ~MFSMRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

class MFSLRegister : public Register {
public:
    MFSLRegister(PhyDiag * phy_diag);
    virtual ~MFSLRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

class FORERegister : public Register {
public:
    FORERegister(PhyDiag * phy_diag);
    virtual ~FORERegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

class MFNRRegister : public Register {
public:
    MFNRRegister(PhyDiag * phy_diag);
    virtual ~MFNRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

// temperature registers
class MTCAPRegister : public Register {
public:
    MTCAPRegister(PhyDiag * phy_diag);
    virtual ~MTCAPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MTMPRegister : public Register {
public:
    MTMPRegister(PhyDiag * phy_diag);
    virtual ~MTMPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

class MTWERegister : public Register {
public:
    MTWERegister(PhyDiag * phy_diag);
    virtual ~MTWERegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

// current/voltage registers
class MVCAPRegister : public Register {
public:
    MVCAPRegister(PhyDiag * phy_diag);
    virtual ~MVCAPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const;
    virtual int AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list);
};

class MVCRRegister : public Register {
public:
    MVCRRegister(PhyDiag * phy_diag);
    virtual ~MVCRRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

// power supplies register
class MSPSRegister : public Register {
public:
    MSPSRegister(PhyDiag * phy_diag);
    virtual ~MSPSRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
};

// general information register
class MSGIRegister : public Register {
public:
    MSGIRegister(PhyDiag * phy_diag);
    virtual ~MSGIRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key);
};

/*
 * Registers for HDR bringup
 * */

class SLRIPRegister : public SLRegister {
private:
    u_int8_t    m_pnat;

public:
     SLRIPRegister(PhyDiag * phy_diag,
        u_int8_t pnat = ACC_REG_PNAT_IB_PORT,
        string sn = ACC_REG_SLRIP_INTERNAL_SECTION_NAME,
        map_akey_areg *mpein_map = NULL,
        map_akey_areg *mpir_map = NULL);

    virtual ~SLRIPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

    void Header_Dump_16nm(stringstream &sstream) const;
    void Header_Dump_7nm(stringstream &sstream) const;
    void Header_Dump_5nm(stringstream &sstream) const;

    void Dump_16nm(const struct slrip_reg &slrip, stringstream &sstream) const;
    void Dump_7nm(const struct slrip_reg &slrip, stringstream &sstream) const;
    void Dump_5nm(const struct slrip_reg &slrip, stringstream &sstream) const;

    virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
};

class SLREGRegister : public Register {
public:
    SLREGRegister(PhyDiag * phy_diag);
    virtual ~SLREGRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;

    void Header_Dump_16nm(stringstream &sstream) const;
    void Header_Dump_7nm(stringstream &sstream) const;

    void Dump_16nm(const struct slreg_reg &slreg, stringstream &sstream) const;
    void Dump_7nm(const struct slreg_reg &slreg, stringstream &sstream) const;
};

class SLCCTRegister : public Register {
public:
    SLCCTRegister(PhyDiag * phy_diag);
    virtual ~SLCCTRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};

class PPAMPRegister : public Register {
public:
    PPAMPRegister(PhyDiag * phy_diag);
    virtual ~PPAMPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    virtual void GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices) override;
};


class PDDRRegister : public Register {
protected:
     u_int8_t m_port_type;
     u_int8_t m_lp_msb;
     u_int8_t m_pnat;
     //u_int8_t m_local_port; // will be set by key
     u_int8_t m_page_select;
     u_int8_t m_module_info_ext;

public:
    PDDRRegister(PhyDiag * phy_diag,
                 uint32_t page_id,
                 const unpack_data_func_t p_unpack_func,
                 string sn,
                 string name,
                 uint32_t fn,
                 string header,
                 PhyPluginSupportedNodesType support_nodes,
                 bool enabled_dump = true,
                 bool retrieve_disconnected = false)

        : Register(phy_diag,
                    ACCESS_REGISTER_ID_PDDR,
                    p_unpack_func,
                    sn,
                    name,
                    fn,
                    NOT_SUPPORT_PDDR,
                    header,
                    support_nodes,
                    enabled_dump,
                    retrieve_disconnected,
                    VIA_GMP,
                    VIA_GMP),

         m_port_type(0),
         m_lp_msb(0),
         m_pnat(ACC_REG_PNAT_IB_PORT),
         m_page_select(page_id),
         m_module_info_ext(0)
    {
    }

    virtual ~PDDRRegister(){}

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
};

class PDDRModuleInfoRegister : public PDDRRegister {
    public:
        PDDRModuleInfoRegister(PhyDiag * phy_diag);
        virtual ~PDDRModuleInfoRegister() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    };

class PDDRLatchedFlagInfoRegister : public PDDRRegister {
    public:
        PDDRLatchedFlagInfoRegister(PhyDiag * phy_diag);
        virtual ~PDDRLatchedFlagInfoRegister() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    };


class PEMIRegister : public Register {
private:
     u_int8_t m_lp_msb;
     u_int8_t m_pnat;
     //u_int8_t m_local_port; // will be set by key
     u_int8_t m_page_select;
     u_int8_t m_module_info_ext;

public:
     PEMIRegister(PhyDiag * phy_diag,
                 uint32_t page_id,
                 const unpack_data_func_t p_unpack_func,
                 string sn,
                 string name,
                 uint32_t fn,
                 string header = "",
                 PhyPluginSupportedNodesType support_nodes = SUPPORT_SW_CA,
                 bool enabled_dump = true,
                 bool retrieve_disconnected = false)

        : Register(phy_diag,
                    ACCESS_REGISTER_ID_PEMI,
                    p_unpack_func,
                    sn,
                    name,
                    fn,
                    NOT_SUPPORT_PEMI,
                    header,
                    support_nodes,
                    enabled_dump,
                    retrieve_disconnected,
                    VIA_GMP,
                    VIA_GMP),

         m_lp_msb(0),
         m_pnat(ACC_REG_PNAT_IB_PORT),
         m_page_select(page_id),
         m_module_info_ext(0)
    {
    }

    virtual ~PEMIRegister(){}

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);
};

class PEMI_Module_Samples_Register : public PEMIRegister {
    public:
        PEMI_Module_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Module_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_SNR_Samples_Register : public PEMIRegister {
    public:
        PEMI_SNR_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_SNR_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_Laser_Samples_Register : public PEMIRegister {
    public:
        PEMI_Laser_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Laser_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_PAM4_Samples_Register : public PEMIRegister {
    public:
        PEMI_PAM4_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_PAM4_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_PRE_FEC_BER_Samples_Register : public PEMIRegister {
    public:
        PEMI_PRE_FEC_BER_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_PRE_FEC_BER_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_FERC_Samples_Register : public PEMIRegister {
    public:
        PEMI_FERC_Samples_Register(PhyDiag * phy_diag);
        virtual ~PEMI_FERC_Samples_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_Module_Properties_Register : public PEMIRegister {
    public:
        PEMI_Module_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Module_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };

class PEMI_SNR_Properties_Register : public PEMIRegister {
    public:
        PEMI_SNR_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_SNR_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_Laser_Properties_Register : public PEMIRegister {
    public:
        PEMI_Laser_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_Laser_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_PAM4_Properties_Register : public PEMIRegister {
    public:
        PEMI_PAM4_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_PAM4_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_PRE_FEC_BER_Properties_Register : public PEMIRegister {
    public:
        PEMI_PRE_FEC_BER_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_PRE_FEC_BER_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PEMI_FERC_Properties_Register : public PEMIRegister {
    public:
        PEMI_FERC_Properties_Register(PhyDiag * phy_diag);
        virtual ~PEMI_FERC_Properties_Register() { }

        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class PRTLRegister : public Register {
    public:
        PRTLRegister(PhyDiag * phy_diag);
        virtual ~PRTLRegister() { }

        virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
        virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
        virtual void DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key) override;
    };


class MPCNTRegister : public Register {
private:
     map_akey_areg &m_mpein_map;

     u_int8_t m_depth;
     u_int8_t m_pcie_index;
     u_int8_t m_node;
     u_int8_t m_grp;
     bool m_clr;

public:
     MPCNTRegister(PhyDiag * phy_diag,
                     map_akey_areg &mpein_map,
                     uint32_t grp,
                     const unpack_data_func_t p_unpack_func,
                     string sn,
                     string name,
                     uint32_t fn,
                     string header = "",
                     PhyPluginSupportedNodesType support_nodes = SUPPORT_CA,
                     bool enabled_dump = true,
                     bool retrieve_disconnected = false)

        : Register(phy_diag,
                    ACCESS_REGISTER_ID_MPCNT,
                    p_unpack_func,
                    sn,
                    name,
                    fn,
                    NOT_SUPPORT_PEMI,
                    header,
                    support_nodes,
                    enabled_dump,
                    retrieve_disconnected,
                    VIA_GMP,
                    VIA_GMP),

         m_mpein_map(mpein_map),
         m_depth(0),
         m_pcie_index(0),
         m_node(0),
         m_grp(grp),
         m_clr(false)
    {
    }

    virtual ~MPCNTRegister(){}

    virtual int UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer);
    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node);

    virtual int BuildDB(AccRegHandler* p_handler, // handler containing the source map of mpein
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *progress_bar);
};



class MPCNT_PCIe_Counters_Register : public MPCNTRegister {
    public:
        MPCNT_PCIe_Counters_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map);
        virtual ~MPCNT_PCIe_Counters_Register() { }

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
    };


class MPCNT_PCIe_Lane_Counters_Register : public MPCNTRegister {
    public:
        MPCNT_PCIe_Lane_Counters_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map);
        virtual ~MPCNT_PCIe_Lane_Counters_Register() { }

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
    };

class MPCNT_PCIe_Timers_Register : public MPCNTRegister {
    public:
        MPCNT_PCIe_Timers_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map);
        virtual ~MPCNT_PCIe_Timers_Register() { }

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const;
    };

class MPCNT_PCIe_Phys_Layer_Register : public MPCNTRegister {
    public:
        MPCNT_PCIe_Phys_Layer_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map);
        virtual ~MPCNT_PCIe_Phys_Layer_Register() { }

        virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
        virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
    };


class PPBMPRegister : public Register {
private:
        u_int8_t    m_pnat;

public:
    PPBMPRegister(PhyDiag * phy_diag);
    virtual ~PPBMPRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};


class MRCSRegister : public Register {
public:
    MRCSRegister(PhyDiag * phy_diag);
    virtual ~MRCSRegister() { }

    virtual int PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node) override;
    virtual void DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const override;
    virtual void DumpRegisterHeader(stringstream &sstream, const string& handler_header) override;
};


// AccRegHandler defines a generic way of reading registers' data across the fabric.
// There are 4 types of access to register:
// by port
// by port + lane
// by node
// by pll group
class AccRegHandler {
protected:
    list_p_fabric_general_err   *phy_errors;
    bool                        can_send_mads_by_lid;
    int                         clbck_error_state;
    Register                    *p_reg;
    string                      handler_header;

public:
    map_akey_areg               data_map;

    AccRegHandler(Register *p_register,  string hh);
    virtual ~AccRegHandler();

    virtual int BuildDB(list_p_fabric_general_err &phy_errors) = 0;


    int PCIAddressCollectGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data);

    int SMPAccessRegisterHandlerGetClbck(const clbck_data_t &clbck_data,
                                         int rec_status,
                                         void *p_attribute_data);

    int GMPAccessRegisterHandlerGetClbck(const clbck_data_t &clbck_data,
                                         int rec_status,
                                         void *p_attribute_data);

    void DumpCSV(CSVOut &csv_out);

    Register * GetPReg() {
        return p_reg;
    }

    bool GetDumpEnabled() {
        return p_reg->GetDumpEnabled();
    }

    void SetHeader(string header) {
        handler_header = header;
    }

    int SendSMPReg(IBNode *p_curr_node,
                   phys_port_t port_num,
                   SMP_AccessRegister *smp_reg,
                   AccRegKey *p_key,
                   ProgressBar *p_progress_bar,
                   clbck_data_t *call_back_data = NULL);

    int SendGMPReg(IBNode *p_curr_node,
                   uint16_t lid,
                   GMP_AccessRegister *gmp_reg,
                   AccRegKey *p_key,
                   ProgressBar *p_progress_bar,
                   clbck_data_t *call_back_data = NULL);

    int SendAccReg(AccRegVia_t acc_reg_via,
                   IBNode *p_curr_node,
                   phys_port_t port_num,
                   uint16_t lid,
                   AccessRegister &acc_reg,
                   AccRegKey *p_key,
                   ProgressBar *p_progress_bar,
                   clbck_data_t *call_back_data = NULL);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);

protected:
    void ExportDataImplNode(export_data_phy_node_t *p_export_data_phy_node);

    template<typename KeyType>
    void ExportDataImplPort(export_data_phy_port_t *p_export_data_phy_port);
};

class AccRegNodeHandler : public AccRegHandler {
public:
    AccRegNodeHandler(Register *p_reg):
        AccRegHandler(p_reg, NODE_HEADER){}
    virtual ~AccRegNodeHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};

class AccRegPortHandler : public AccRegHandler {
public:
    AccRegPortHandler(Register *p_reg):
        AccRegHandler(p_reg, PORT_HEADER){}
    virtual ~AccRegPortHandler(){}

    virtual AccRegKey* CreateKey(uint64_t ng, uint64_t pg, uint8_t pn);

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};

class AccRegPortIndexHandler : public AccRegPortHandler {
private:

    vector_uint64  indexes;
    uint64_t       cur_index;

public:
    AccRegPortIndexHandler(Register *p_reg, vector_uint64& idxs) :
        AccRegPortHandler(p_reg), indexes(idxs), cur_index(0)
    {
        handler_header = PORT_INDEX_HEADER;
    }

    virtual ~AccRegPortIndexHandler(){}

    virtual AccRegKey* CreateKey(uint64_t ng, uint64_t pg, uint8_t pn);
    virtual int BuildDB(list_p_fabric_general_err &phy_errors);
};

class AccRegGroupHandler : public AccRegHandler {
private:
    uint32_t group_max_idx;
public:
    AccRegGroupHandler(Register *p_reg):
        AccRegHandler(p_reg, GROUP_HEADER), group_max_idx(0){}
    virtual ~AccRegGroupHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};

class AccRegNodeSensorHandler : public AccRegHandler {
private:
    AccRegHandler *p_acc_reg_handler; // pointer for the capability register

public:
    AccRegNodeSensorHandler(Register *p_reg, AccRegHandler *p_acc_reg):
        AccRegHandler(p_reg, NODE_SENSOR_HEADER),
        p_acc_reg_handler(p_acc_reg){}
    virtual ~AccRegNodeSensorHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};

class AccRegTypeGroupHandler : public AccRegHandler {
private:

    uint8_t max_type;
    uint16_t max_indices;

public:
    AccRegTypeGroupHandler(Register *p_reg, uint8_t types, uint16_t indices)
        : AccRegHandler(p_reg, TYPE_GROUP_HEADER), max_type(types), max_indices(indices){}
    virtual ~AccRegTypeGroupHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);
};


class AccRegSpecificHandler : public AccRegHandler {
public:
    AccRegSpecificHandler(Register *p_reg, string header = PORT_HEADER)
        : AccRegHandler(p_reg, header){}
    virtual ~AccRegSpecificHandler(){}

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};


class AccRegPortLaneHandler : public AccRegHandler {
    u_int8_t max_idx_in_lane;
public:
    AccRegPortLaneHandler(Register *p_reg, u_int8_t max_idx = 1):
        AccRegHandler(p_reg, PORT_LANE_HEADER), max_idx_in_lane(max_idx){}

    virtual ~AccRegPortLaneHandler() { }

    virtual int BuildDB(list_p_fabric_general_err &phy_errors);

    virtual void ExportData(export_data_phy_port_t *p_export_data_phy_port,
                            export_data_phy_node_t *p_export_data_phy_node);
};



#endif //_ACC_REGISTER_H_

