/*
 * Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>

#include "ibis.h"
#include "packets/packets_layouts.h"


/****************************************************/
/******************private methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCMadGetSet(u_int16_t lid,
        u_int8_t sl,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        const data_func_set_t *p_cc_log_attribute_data,
        const data_func_set_t *p_cc_mgt_attribute_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct MAD_CongestionControl cc_mad;

    CLEAR_STRUCT(cc_mad);

    //build cc lid route header
    this->CommonMadHeaderBuild(&cc_mad.MAD_Header_Common,
            IBIS_IB_CLASS_CC,
            method,
            attribute_id,
            attribute_modifier);

    cc_mad.CC_Key = this->key_manager.GetKey(lid, IBIS_CC_KEY);

    /*
    if (p_cc_mgt_attribute_data) {
        IBIS_LOG(TT_LOG_LEVEL_MAD, "p_cc_mgt_attribute_data = %u\n", *(int*)p_cc_mgt_attribute_data);
    } else if (p_cc_log_attribute_data) {
        IBIS_LOG(TT_LOG_LEVEL_MAD, "p_cc_log_attribute_data = %u\n", *(int*)p_cc_log_attribute_data);
    }
    */
    //send the mad
    IBIS_RETURN(this->MadGetSet(lid, 1, sl, IBIS_IB_DEFAULT_QP1_QKEY,
            IBIS_IB_CLASS_CC,
            method, attribute_id, attribute_modifier,
            (p_cc_mgt_attribute_data) ? IBIS_IB_MGT_DATA_OFFSET_CC : IBIS_IB_LOG_DATA_OFFSET_CC,
            data_func_set_t(&cc_mad, IBIS_FUNC_LST(MAD_CongestionControl)),
            (p_cc_mgt_attribute_data) ? *p_cc_mgt_attribute_data : *p_cc_log_attribute_data,
            p_clbck_data));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCClassPortInfoGet(u_int16_t lid,
        u_int8_t sl,
        struct IB_ClassPortInfo *p_ib_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_ib_class_port_info);

    data_func_set_t data(p_ib_class_port_info, IBIS_FUNC_LST(IB_ClassPortInfo));

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo Get MAD lid = %u\n", lid);
    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,                        /* attribute modifier */
            NULL,                     /* p_cc_log_attribute_data */
            &data,                    /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCClassPortInfoSet(u_int16_t lid,
        u_int8_t sl,
        struct IB_ClassPortInfo *p_ib_class_port_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending IB_ClassPortInfo Get MAD lid = %u\n", lid);

    data_func_set_t data(p_ib_class_port_info, IBIS_FUNC_LST(IB_ClassPortInfo));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CLASS_PORT_INFO,
            0,                        /* attribute modifier */
            NULL,                     /* p_cc_log_attribute_data */
            &data,     /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionInfoGet(u_int16_t lid,
        u_int8_t sl,
        struct CC_CongestionInfo *p_cc_congestion_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionInfo Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_congestion_info, IBIS_FUNC_LST(CC_CongestionInfo));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_INFO,
            0,                        /* attribute modifier */
            NULL,                     /* p_cc_log_attribute_data */
            &data,                    /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionKeyInfoGet(u_int16_t lid,
        u_int8_t sl,
        struct CC_CongestionKeyInfo *p_cc_congestion_key_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_key_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionKeyInfo Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_congestion_key_info, IBIS_FUNC_LST(CC_CongestionKeyInfo));


    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_KEY_INFO,
            0,                        /* attribute modifier */
            NULL,                     /* p_cc_log_attribute_data */
            &data,                    /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionKeyInfoSet(u_int16_t lid,
        u_int8_t sl,
        struct CC_CongestionKeyInfo *p_cc_congestion_key_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionKeyInfo Set MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_congestion_key_info, IBIS_FUNC_LST(CC_CongestionKeyInfo));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_KEY_INFO,
            0,                        /* attribute modifier */
            NULL,                     /* p_cc_log_attribute_data */
            &data,                    /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionLogSwitchGet(u_int16_t lid,
        u_int8_t sl,
        struct CC_CongestionLogSwitch *p_cc_congestion_log_sw,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_log_sw);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionLogSwitch Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_congestion_log_sw, IBIS_FUNC_LST(CC_CongestionLogSwitch));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_LOG,
            0,                       /* attribute modifier */
            &data,                   /* p_cc_log_attribute_data */
            NULL,                    /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionLogCAGet(u_int16_t lid,
        u_int8_t sl,
        struct CC_CongestionLogCA *p_cc_congestion_log_ca,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_log_ca);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionLogCA Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_congestion_log_ca, IBIS_FUNC_LST(CC_CongestionLogCA));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CONG_LOG,
            0,                       /* attribute modifier */
            &data,                   /* p_cc_log_attribute_data */
            NULL,                    /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCSwitchCongestionSettingGet(u_int16_t lid,
        u_int8_t sl,
        struct CC_SwitchCongestionSetting *p_cc_sw_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_sw_congestion_setting);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchCongestionSetting Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_sw_congestion_setting, IBIS_FUNC_LST(CC_SwitchCongestionSetting));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_CONG_SETTING,
            0,                          /* attribute modifier */
            NULL,                       /* p_cc_log_attribute_data */
            &data,                      /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCSwitchCongestionSettingSet(u_int16_t lid,
        u_int8_t sl,
        struct CC_SwitchCongestionSetting *p_cc_sw_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchCongestionSetting Set MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_sw_congestion_setting, IBIS_FUNC_LST(CC_SwitchCongestionSetting));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_CONG_SETTING,
            0,                          /* attribute modifier */
            NULL,                       /* p_cc_log_attribute_data */
            &data,                      /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCSwitchPortCongestionSettingGet(u_int16_t lid,
        u_int8_t sl,
        u_int8_t block_idx,
        struct CC_SwitchPortCongestionSetting *p_cc_sw_port_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_sw_port_congestion_setting);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchPortCongestionSetting Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_sw_port_congestion_setting, IBIS_FUNC_LST(CC_SwitchPortCongestionSetting));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_PORT_CONG_SETTING,
            block_idx,                       /* attribute modifier */
            NULL,                            /* p_cc_log_attribute_data */
            &data,                           /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCSwitchPortCongestionSettingSet(u_int16_t lid,
        u_int8_t sl,
        u_int8_t block_idx,
        struct CC_SwitchPortCongestionSetting *p_cc_sw_port_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchPortCongestionSetting Set MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_sw_port_congestion_setting, IBIS_FUNC_LST(CC_SwitchPortCongestionSetting));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_PORT_CONG_SETTING,
            block_idx,                       /* attribute modifier */
            NULL,                            /* p_cc_log_attribute_data */
            &data,                           /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCACongestionSettingGet(u_int16_t lid,
        u_int8_t sl,
        struct CC_CACongestionSetting *p_cc_ca_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_ca_congestion_setting);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCACongestionSetting Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_ca_congestion_setting, IBIS_FUNC_LST(CC_CACongestionSetting));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CA_CONG_SETTING,
            0,                          /* attribute modifier */
            NULL,                       /* p_cc_log_attribute_data */
            &data,                      /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCACongestionSettingSet(u_int16_t lid,
        u_int8_t sl,
        struct CC_CACongestionSetting *p_cc_ca_congestion_setting,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCACongestionSetting Set MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_ca_congestion_setting, IBIS_FUNC_LST(CC_CACongestionSetting));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_CA_CONG_SETTING,
            0,                          /* attribute modifier */
            NULL,                       /* p_cc_log_attribute_data */
            &data,                      /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionControlTableGet(u_int16_t lid,
        u_int8_t sl,
        u_int8_t block_idx,
        struct CC_CongestionControlTable *p_cc_congestion_control_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_congestion_control_table);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionControlTable Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_congestion_control_table, IBIS_FUNC_LST(CC_CongestionControlTable));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_CC_TBL,
            block_idx,                     /* attribute modifier */
            NULL,                          /* p_cc_log_attribute_data */
            &data,                         /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCCongestionControlTableSet(u_int16_t lid,
        u_int8_t sl,
        u_int8_t block_idx,
        struct CC_CongestionControlTable *p_cc_congestion_control_table,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCCongestionControlTable Set MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_congestion_control_table, IBIS_FUNC_LST(CC_CongestionControlTable));


    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_CONGESTION_CONT_CC_TBL,
            block_idx,                     /* attribute modifier */
            NULL,                          /* p_cc_log_attribute_data */
            &data,                         /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::CCTimeStampGet(u_int16_t lid,
        u_int8_t sl,
        struct CC_TimeStamp *p_cc_time_stamp,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_time_stamp);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCTimeStamp Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_time_stamp, IBIS_FUNC_LST(CC_TimeStamp));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_TIME_STAMP,
            0,              /* attribute modifier */
            NULL,           /* p_cc_log_attribute_data */
            &data,          /* p_cc_mgt_attribute_data */
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

// CC configuration
int Ibis::CCEnhancedInfoGet(u_int16_t lid,
        u_int8_t sl,
        struct CC_EnhancedCongestionInfo *p_cc_enhanced_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_enhanced_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCEnhancedInfo Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_enhanced_info, IBIS_FUNC_LST(CC_EnhancedCongestionInfo));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_ENHANCED_INFO,
            0,                  // attribute modifier
            NULL,               // p_cc_log_attribute_data
            &data,              // p_cc_mgt_attribute_data
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

// switch MADs
int Ibis::CCSwitchGeneralSettingsGet(u_int16_t lid,
           u_int8_t sl,
           struct CC_CongestionSwitchGeneralSettings *p_cc_switch_general_settings,
           const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_switch_general_settings);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSwitchGeneralSettings Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_switch_general_settings, IBIS_FUNC_LST(CC_CongestionSwitchGeneralSettings));


    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_SW_GENERAL_SETTINGS,
            0,                  // attribute modifier
            NULL,               // p_cc_log_attribute_data
            &data,              // p_cc_mgt_attribute_data
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCPortProfileSettingsGet(u_int16_t lid,
           u_int8_t port,
           u_int8_t sl,
           struct CC_CongestionPortProfileSettings *p_cc_port_profile_settings,
           const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCPortProfileSettings Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_port_profile_settings, IBIS_FUNC_LST(CC_CongestionPortProfileSettings));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_PORT_PROFILE_SETTINGS,
            port,               // attribute modifier
            NULL,               // p_cc_log_attribute_data
            &data,              // p_cc_mgt_attribute_data
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCSLMappingSettingsGet(u_int16_t lid,
           u_int8_t port,
           u_int8_t sl,
           struct CC_CongestionSLMappingSettings *p_cc_sl_mapping_settings,
           const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_sl_mapping_settings);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCSLMappingSettings Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_sl_mapping_settings, IBIS_FUNC_LST(CC_CongestionSLMappingSettings));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_SL_MAPPING_SETTINGS,
            port,               // attribute modifier
            NULL,               // p_cc_log_attribute_data
            &data,              // p_cc_mgt_attribute_data
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

// HCA MADs
int Ibis::CCHCAGeneralSettingsGet(u_int16_t lid,
           u_int8_t sl,
           struct CC_CongestionHCAGeneralSettings *p_cc_hca_general_settings,
           const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_hca_general_settings);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCHCAGeneralSettings Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_hca_general_settings, IBIS_FUNC_LST(CC_CongestionHCAGeneralSettings));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_HCA_GENERAL_SETTINGS,
            0,                  // attribute modifier
            NULL,               // p_cc_log_attribute_data
            &data,              // p_cc_mgt_attribute_data
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCHCARPParametersGet(u_int16_t lid,
           u_int8_t sl,
           struct CC_CongestionHCARPParameters *p_cc_hca_rp_parameters,
           const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_hca_rp_parameters);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCHCARPParameters Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_hca_rp_parameters, IBIS_FUNC_LST(CC_CongestionHCARPParameters));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_HCA_RP_PARAMETERS,
            0,                  // attribute modifier
            NULL,               // p_cc_log_attribute_data
            &data,              // p_cc_mgt_attribute_data
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCHCANPParametersGet(u_int16_t lid,
           u_int8_t sl,
           struct CC_CongestionHCANPParameters *p_cc_hca_np_parameters,
           const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_hca_np_parameters);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCHCANPParameters Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_hca_np_parameters, IBIS_FUNC_LST(CC_CongestionHCANPParameters));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_HCA_NP_PARAMETERS,
            0,                  // attribute modifier
            NULL,               // p_cc_log_attribute_data
            &data,              // p_cc_mgt_attribute_data
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCHCAStatisticsQueryGet(u_int16_t lid,
           u_int8_t sl,
           struct CC_CongestionHCAStatisticsQuery *p_cc_hca_statistics_query,
           const clbck_data_t *p_clbck_data,
           bool to_clear_congestion_counters)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_hca_statistics_query);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CCHCAStatisticsQuery Get MAD lid = %u\n", lid);

    p_cc_hca_statistics_query->clear = (u_int8_t)to_clear_congestion_counters;

    data_func_set_t data(p_cc_hca_statistics_query, IBIS_FUNC_LST(CC_CongestionHCAStatisticsQuery));

    int rc = this->CCMadGetSet(lid,
            sl,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_CONGESTION_CONT_HCA_STATISTICS_QUERY,
            0,                  // attribute modifier
            NULL,               // p_cc_log_attribute_data
            &data,              // p_cc_mgt_attribute_data
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCHCAAlgoConfigGet(u_int16_t lid,
           u_int8_t algo_slot,
           u_int8_t encap_type,
           struct CC_CongestionHCAAlgoConfig *p_cc_hca_algo_config,
           const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_hca_algo_config);
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending CCHCAAlgoConfig Get MAD lid = %u, slot %u\n",
             lid, algo_slot);

    data_func_set_t data(p_cc_hca_algo_config, IBIS_FUNC_LST(CC_CongestionHCAAlgoConfig));

    u_int32_t attribute_modifier = (algo_slot & 0x0F) << 24;
    p_cc_hca_algo_config->encap_type = encap_type;

    int rc = this->CCMadGetSet(lid,
                               0,
                               IBIS_IB_MAD_METHOD_GET,
                               IBIS_IB_ATTR_CONGESTION_CONT_HCA_ALGO_CONFIG,
                               attribute_modifier, // attribute modifier
                               NULL,               // p_cc_log_attribute_data
                               &data,              // p_cc_mgt_attribute_data
                               p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCHCAAlgoConfigParamGet(u_int16_t lid,
           u_int8_t algo_slot,
           u_int8_t encap_type,
           struct CC_CongestionHCAAlgoConfigParams *p_cc_hca_algo_config_param,
           const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_hca_algo_config_param);
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending CCHCAAlgoConfigParam Get MAD lid = %u, slot %u\n",
             lid, algo_slot);

    data_func_set_t data(p_cc_hca_algo_config_param, IBIS_FUNC_LST(CC_CongestionHCAAlgoConfigParams));

    u_int32_t attribute_modifier = (algo_slot & 0x0F) << 24;
    p_cc_hca_algo_config_param->encap_type = encap_type;

    int rc = this->CCMadGetSet(lid,
                               0,
                               IBIS_IB_MAD_METHOD_GET,
                               IBIS_IB_ATTR_CONGESTION_CONT_HCA_ALGO_CONFIG_PARAMS,
                               attribute_modifier, // attribute modifier
                               NULL,               // p_cc_log_attribute_data
                               &data,              // p_cc_mgt_attribute_data
                               p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCHCAAlgoCountersGet(u_int16_t lid,
           u_int8_t algo_slot,
           u_int8_t encap_type,
           struct CC_CongestionHCAAlgoCounters *p_cc_hca_algo_counters,
           const clbck_data_t *p_clbck_data,
           bool to_clear_congestion_counters)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_hca_algo_counters);
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending CCHCAAlgoCounters Get MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_hca_algo_counters, IBIS_FUNC_LST(CC_CongestionHCAAlgoCounters));

    u_int32_t attribute_modifier = (algo_slot & 0x0F) << 24;
    p_cc_hca_algo_counters->clear = to_clear_congestion_counters;
    p_cc_hca_algo_counters->encap_type = encap_type;

    int rc = this->CCMadGetSet(lid,
                               0,
                               IBIS_IB_MAD_METHOD_GET,
                               IBIS_IB_ATTR_CONGESTION_CONT_HCA_ALGO_COUNTERS,
                               attribute_modifier, // attribute modifier
                               NULL,               // p_cc_log_attribute_data
                               &data,              // p_cc_mgt_attribute_data
                               p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::CCHCAAlgoCountersSet(u_int16_t lid,
           u_int8_t algo_slot,
           u_int8_t encap_type,
           struct CC_CongestionHCAAlgoCounters *p_cc_hca_algo_counters,
           const clbck_data_t *p_clbck_data,
           bool to_clear_congestion_counters)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_cc_hca_algo_counters);
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending CCHCAAlgoCounters Set MAD lid = %u\n", lid);

    data_func_set_t data(p_cc_hca_algo_counters, IBIS_FUNC_LST(CC_CongestionHCAAlgoCounters));

    u_int32_t attribute_modifier = (algo_slot & 0x0F) << 24;
    p_cc_hca_algo_counters->clear = to_clear_congestion_counters;
    p_cc_hca_algo_counters->encap_type = encap_type;

    int rc = this->CCMadGetSet(lid,
                               0,
                               IBIS_IB_MAD_METHOD_SET,
                               IBIS_IB_ATTR_CONGESTION_CONT_HCA_ALGO_COUNTERS,
                               attribute_modifier, // attribute modifier
                               NULL,               // p_cc_log_attribute_data
                               &data,              // p_cc_mgt_attribute_data
                               p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}
