/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <iomanip>
#include <sstream>
#include <typeinfo>
#include <math.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unordered_set>
#include "uphy/uphy.h"
#include "phy_diag.h"
#include "phy_diag_errs.h"
#include "infiniband/ibutils/output_stream_primitives.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_version.h"
#include "infiniband/ibdm/cable/CombinedCable.h"

#define DACS_MEDIA_TYPE_STR "DACs"
#define ACC_MEDIA_TYPE_STR "ACC"
#define ACTIVE_MEDIA_TYPE_STR "Active"
#define ACTIVE_DID_MEDIA_TYPE_STR "Active_DiD"
#define UNKNOWN_MEDIA_TYPE_STR "Unknown"

#define UNKNOWN_PAGE_INDX -100

static const char *media_type2char(const MediaType media_type)
{
    switch (media_type) {
    case MT_DACS:       return(DACS_MEDIA_TYPE_STR);
    case MT_ACC:        return(ACC_MEDIA_TYPE_STR);
    case MT_ACTIVE:     return(ACTIVE_MEDIA_TYPE_STR);
    case MT_ACTIVE_DID: return(ACTIVE_DID_MEDIA_TYPE_STR);
    default:            return(UNKNOWN_MEDIA_TYPE_STR);
    }
}

static MediaType char2media_type(const char *p_str)
{
    if (!strcmp(p_str, DACS_MEDIA_TYPE_STR))        return MT_DACS;
    if (!strcmp(p_str, ACC_MEDIA_TYPE_STR))         return MT_ACC;
    if (!strcmp(p_str, ACTIVE_MEDIA_TYPE_STR))      return MT_ACTIVE;
    if (!strcmp(p_str, ACTIVE_DID_MEDIA_TYPE_STR))  return MT_ACTIVE_DID;

    return MT_UNKNOWN;
}

static inline bool IsSupportAccRegFailed(u_int64_t app_data,
                                         u_int64_t not_support_bit)
{
    return (app_data & (NOT_SUPPORT_SMP_ACCESS_REGISTER | \
                        not_support_bit));
}

bool IsPhyPluginSupportNodeType(PhyPluginSupportedNodesType supported_nodes, const IBNode * node)
{
    IBDIAGNET_ENTER;

    if(node == NULL)
        IBDIAGNET_RETURN(false);

    if (supported_nodes == SUPPORT_ALL)
        IBDIAGNET_RETURN(true);

    switch(node->type)
    {
        case IB_SW_NODE:
            IBDIAGNET_RETURN(supported_nodes & SUPPORT_SW);

        case IB_CA_NODE:
            IBDIAGNET_RETURN(
                    node->isSpecialNode() ? supported_nodes & SUPPORT_SPECIAL_NODES
                                          : supported_nodes & SUPPORT_CA);

        case IB_RTR_NODE:
            IBDIAGNET_RETURN(supported_nodes & SUPPORT_RTR);

        default:
            IBDIAGNET_RETURN(false);
    }
}

/****************************************************/
/*
 * public methods
 */
PhyDiag::PhyDiag(IBDiag *p_ibdiag) : Plugin(PLUGIN_NAME, p_ibdiag),
        can_send_mads_by_lid(false),
        to_get_phy_info(false), to_get_ppamp(false), to_get_slrg(false), to_get_mpcnt(false), to_dump_cap_reg(false),
        to_reset_counters(false), to_get_pci_info(false), to_reset_pci_counters(false),
        to_show_ber_warnings(false), to_load_ber_thresholds_table(false),
        to_get_cable_disconnected_ports(false), to_get_cable_full_data(false),
        clbck_error_state(IBDIAG_SUCCESS_CODE), total_ber_errors(0),
        module_info_dd_index(UNKNOWN_PAGE_INDX), module_info_acc_reg_index(UNKNOWN_PAGE_INDX),
        pf_export_data_phy_node(NULL), pf_export_data_phy_port(NULL)
{
    IBDIAGNET_ENTER;
    /*
     * this stage is skipped by default
     */
    this->MarkStageAsSkipped();

    /*
     * add all options
     */
    stringstream def_val_sstream;
    string def_val_string;

    this->AddOptions(OPTION_PHY_INFO_GET, OPTION_PHY_INFO_GET_SHORT,
                     OPTION_PHY_INFO_GET_ARG_DESC, OPTION_PHY_INFO_GET_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PHY_CABLE_DISCONNECTED, OPTION_PHY_CABLE_DISCONNECTED_SHORT,
                     OPTION_PHY_CABLE_DISCONNECTED_ARG_DESC,
                     OPTION_PHY_CABLE_DISCONNECTED_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_PPAMP_GET, OPTION_PPAMP_GET_SHORT,
                     OPTION_PPAMP_GET_ARG_DESC, OPTION_PPAMP_GET_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_SHOW_CAP_REG, OPTION_SHOW_CAP_REG_SHORT,
                     OPTION_SHOW_CAP_REG_ARG_DESC, OPTION_SHOW_CAP_REG_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PHY_INFO_RESET, OPTION_PHY_INFO_RESET_SHORT,
                     OPTION_PHY_INFO_RESET_ARG_DESC,
                     OPTION_PHY_INFO_RESET_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PHY_CABLE_FULL_DATA, OPTION_PHY_CABLE_FULL_DATA_SHORT,
                     OPTION_PHY_CABLE_FULL_DATA_ARG_DESC,
                     OPTION_PHY_CABLE_FULL_DATA_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_PCI_INFO_GET, OPTION_PCI_INFO_GET_SHORT,
                     OPTION_PCI_INFO_GET_ARG_DESC, OPTION_PCI_INFO_GET_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG);
    this->AddOptions(OPTION_PCI_INFO_RESET, OPTION_PCI_INFO_RESET_SHORT,
                     OPTION_PCI_INFO_RESET_ARG_DESC,
                     OPTION_PCI_INFO_RESET_DESC, OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG);

    def_val_sstream.str("");
    def_val_sstream << BER_THRESHOLD_ERROR;
    def_val_string = def_val_sstream.str();

    this->AddOptions(OPTION_BER_THRESHOLD_ERROR,
                     OPTION_BER_THRESHOLD_ERROR_SHORT,
                     OPTION_BER_THRESHOLD_ERROR_ARG_DESC,
                     OPTION_BER_THRESHOLD_ERROR_DESC,
                     def_val_string,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN);

    def_val_sstream.str("");
    def_val_sstream << BER_THRESHOLD_WARNING;
    def_val_string = def_val_sstream.str();
    this->AddOptions(OPTION_BER_THRESHOLD_WARNING,
                     OPTION_BER_THRESHOLD_WARNING_SHORT,
                     OPTION_BER_THRESHOLD_WARNING_ARG_DESC,
                     OPTION_BER_THRESHOLD_WARNING_DESC,
                     def_val_string,
                     IBIS_CMD_DEPRECATED | IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_SHOW_BER_WARNINGS, OPTION_SHOW_BER_WARNINGS_SHORT,
                     OPTION_SHOW_BER_WARNINGS_ARG_DESC, OPTION_SHOW_BER_WARNINGS_DESC,
                     OPTION_DEF_VAL_FALSE, IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_BER_THRESHOLD_TABLE, OPTION_BER_THRESHOLD_TABLE_SHORT,
                     OPTION_BER_THRESHOLD_TABLE_ARG_DESC, OPTION_BER_THRESHOLD_TABLE_DESC,
                     OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_CREATE_BER_THRESHOLD_TABLE, OPTION_CREATE_BER_THRESHOLD_TABLE_SHORT,
                     OPTION_CREATE_BER_THRESHOLD_TABLE_ARG_DESC,
                     OPTION_CREATE_BER_THRESHOLD_TABLE_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_NO_CONF_FILE | IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_ENABLED_REGS, OPTION_ENABLED_REGS_SHORT,
                     OPTION_ENABLED_REGS_ARG_DESC, OPTION_ENABLED_REGS_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_DISABLED_REGS, OPTION_DISABLED_REGS_SHORT,
                     OPTION_DISABLED_REGS_ARG_DESC, OPTION_DISABLED_REGS_DESC, OPTION_DEF_VAL_NULL,
                     IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_SLRG_REG, OPTION_SLRG_REG_SHORT,
                     OPTION_SLRG_REG_ARG_DESC, OPTION_SLRG_REG_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_MPCNT_REG, OPTION_MPCNT_REG_SHORT,
                     OPTION_MPCNT_REG_ARG_DESC, OPTION_MPCNT_REG_DESC, OPTION_DEF_VAL_FALSE,
                     IBIS_CMD_BOOLEAN_FLAG | IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_ACC_REG_PRIORITY,
                     OPTION_ACC_REG_PRIORITY_SHORT,
                     OPTION_ACC_REG_PRIORITY_ARG_DESC,
                     OPTION_ACC_REG_PRIORITY_DESC,
                     OPTION_DEF_VAL_NULL,
                     IBIS_CMD_HIDDEN);

    this->AddOptions(OPTION_UPHY, OPTION_UPHY_SHORT, OPTION_UPHY_ARG_DESC,
                     OPTION_UPHY_DESC, OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);
    this->AddOptions(OPTION_UPHY_DUMPS, OPTION_UPHY_DUMPS_SHORT, OPTION_UPHY_DUMPS_ARG_DESC,
                     OPTION_UPHY_DUMPS_DESC, OPTION_DEF_VAL_NULL, IBIS_CMD_HIDDEN);

    this->AddDescription(PLUGIN_DESC);

    this->p_discovered_fabric = this->p_ibdiag->GetDiscoverFabricPtr();
    this->p_ibis_obj = this->p_ibdiag->GetIbisPtr();
    this->p_ibdm_extended_info = this->p_ibdiag->GetIBDMExtendedInfoPtr();
    this->p_capability_module = this->p_ibdiag->GetCapabilityModulePtr();

    this->InitRegisterDependencies();
    this->InitBERTable();

    phy_errors.clear();

    IBDIAGNET_RETURN_VOID;
}

int PhyDiag::CleanResources()
{
    IBDIAGNET_ENTER;

    // Clear PHY objects
    map_guid_pnode::iterator it = this->p_discovered_fabric->NodeByGuid.begin();
    for (; it != this->p_discovered_fabric->NodeByGuid.end(); it++) {

        if (!it->second)
            continue;

        delete it->second->p_phy_data;
        it->second->p_phy_data = NULL;

        for (phys_port_t i = 0; i <= it->second->numPorts; ++i) {
            IBPort *p = it->second->getPort(i);

            if (!p)
                continue;

            delete p->p_phy_data;
            p->p_phy_data = NULL;
        }
    }

    release_container_data(diagnostic_data_vec);
    release_container_data(reg_handlers_vec);
    release_container_data(pci_reg_handlers_vec);

    this->module_info_dd_index = UNKNOWN_PAGE_INDX;
    this->module_info_acc_reg_index = UNKNOWN_PAGE_INDX;

    // cannot be cleaned by release_container_data, map::key have outside owner
    for (vec_map_akey_diagnostic_data_t::iterator it_vec = dd_idx_and_akey_to_dd_page.begin();
         it_vec != dd_idx_and_akey_to_dd_page.end(); ++it_vec) {
         if (!(*it_vec))
             continue;
         map_akey_diagnostic_data_t *p_map_akey_dd = *it_vec;
         for (map_akey_diagnostic_data_t::iterator it_map = p_map_akey_dd->begin();
              it_map != p_map_akey_dd->end(); ++it_map) {
              delete (it_map->second);
         }
         p_map_akey_dd->clear();
         delete (*it_vec);
    }

    release_container_data(ber_vec);
    release_container_data(eff_ber_vec);

    release_container_data(port_cidx_and_dd_idx_to_dd_page);
    release_container_data(node_cidx_and_dd_idx_to_dd_page);

    dd_idx_and_akey_to_dd_page.clear();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

PhyDiag::~PhyDiag()
{
    IBDIAGNET_ENTER;

    PhyDiag::CleanResources();

    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::InitRegisterDependencies()
{
    IBDIAGNET_ENTER;

    // PHY Registers
    this->register_dependencies[ACC_REG_MFSM_NAME].insert(ACC_REG_MFCR_NAME);
    this->register_dependencies[ACC_REG_MFSL_NAME].insert(ACC_REG_MFCR_NAME);
    this->register_dependencies[ACC_REG_MFNR_NAME].insert(ACC_REG_MFCR_NAME);
    this->register_dependencies[ACC_REG_MTMP_NAME].insert(ACC_REG_MTCAP_NAME);
    this->register_dependencies[ACC_REG_MVCR_NAME].insert(ACC_REG_MVCAP_NAME);
    this->register_dependencies[ACC_REG_PEUCG_NAME].insert(ACC_REG_PMDR_NAME);

    // PCI Registers
    this->register_dependencies[ACC_REG_MPIR_NAME].insert(ACC_REG_MPEIN_NAME);
    this->register_dependencies[ACC_REG_SLRG_NAME].insert(ACC_REG_MPIR_NAME);
    this->register_dependencies[ACC_REG_SLRP_NAME].insert(ACC_REG_MPIR_NAME);
    this->register_dependencies[ACC_REG_SLTP_NAME].insert(ACC_REG_MPIR_NAME);
    this->register_dependencies[ACC_REG_SLRIP_NAME].insert(ACC_REG_MPIR_NAME);

    this->register_dependencies[ACC_REG_MPCNT_PCI_CNT_NAME].insert(ACC_REG_MPEIN_NAME);
    this->register_dependencies[ACC_REG_MPCNT_PCI_LCNT_NAME].insert(ACC_REG_MPEIN_NAME);
    this->register_dependencies[ACC_REG_MPCNT_PCI_TIMERS_NAME].insert(ACC_REG_MPEIN_NAME);
    this->register_dependencies[ACC_REG_MPCNT_PCI_PHYS_NAME].insert(ACC_REG_MPEIN_NAME);

    // Diagnostic Data PHY
    this->register_dependencies[DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_OPERATION_INFO_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_PHY_INFO_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_MODULE_INFO_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_MODULE_INFO_NAME].insert(ACC_REG_PRTL_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_LINK_DOWN_INFO_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_LINK_UP_INFO_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_NAME].insert(ACC_REG_PRTL_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_PLR_COUNTERS_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    this->register_dependencies[DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_NAME].insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);

    // Diagnostic Data PCI
    set_string_t &pcie_cntrs = this->register_dependencies[DIAGNOSTIC_DATA_PCIE_CNTRS_NAME];
    pcie_cntrs.insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    pcie_cntrs.insert(ACC_REG_MPEIN_NAME);
    set_string_t &pcie_lanes = this->register_dependencies[DIAGNOSTIC_DATA_PCIE_LANES_NAME];
    pcie_lanes.insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    pcie_lanes.insert(ACC_REG_MPEIN_NAME);
    set_string_t &pcie_timers = this->register_dependencies[DIAGNOSTIC_DATA_PCIE_TIMERS_NAME];
    pcie_timers.insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
    pcie_timers.insert(ACC_REG_MPEIN_NAME);

    // MPCNT
    {
        set_string_t &alias = this->register_dependencies[ACC_REG_ALL_MPCNT_NAME];
        alias.insert(ACC_REG_MPCNT_PCI_CNT_NAME);
        alias.insert(ACC_REG_MPCNT_PCI_LCNT_NAME);
        alias.insert(ACC_REG_MPCNT_PCI_TIMERS_NAME);
        alias.insert(ACC_REG_MPCNT_PCI_PHYS_NAME);
    }

    // PCI
    {
        set_string_t &alias = this->register_dependencies[ACC_REG_ALL_PCI_NAME];
        alias.insert(ACC_REG_MPEIN_NAME);
        alias.insert(ACC_REG_MPIR_NAME);
        alias.insert(ACC_REG_SLRG_NAME);
        alias.insert(ACC_REG_SLRP_NAME);
        alias.insert(ACC_REG_SLTP_NAME);
        alias.insert(ACC_REG_SLRIP_NAME);
        alias.insert(ACC_REG_ALL_MPCNT_NAME);
    }


    // PCAM dependencies
    this->register_dependencies[ACC_REG_SLRG_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_SLRP_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_SLTP_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_SLSIR_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_PTAS_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_ALL_PDDR_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_PMDR_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_PPHCR_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_PEUCG_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_PPLL_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_PMCR_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_PTYS_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_SLRIP_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_SLREG_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_SLCCT_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_SLLM_NAME].insert(ACC_REG_PCAM_NAME);
    this->register_dependencies[ACC_REG_PPAMP_NAME].insert(ACC_REG_PCAM_NAME);
    
    // Alias - PDDR
    {
        set_string_t &alias = this->register_dependencies[ACC_REG_ALL_PDDR_NAME];
        alias.insert(ACC_REG_PDDR_MODULE_NAME);
        alias.insert(ACC_REG_PDDR_LATCHED_NAME);
    }

    // PRTL dependencies
    this->register_dependencies[ACC_REG_PDDR_MODULE_NAME].insert(ACC_REG_PRTL_NAME);
    this->register_dependencies[ACC_REG_PDDR_LATCHED_NAME].insert(ACC_REG_PRTL_NAME);

    // Alias - PEMI
    {
        set_string_t &alias = this->register_dependencies[ACC_REG_ALL_PEMI_NAME];
        alias.insert(ACC_REG_PEMI_MODULE_S_NAME);
        alias.insert(ACC_REG_PEMI_SNR_S_NAME);
        alias.insert(ACC_REG_PEMI_LASER_S_NAME);
        alias.insert(ACC_REG_PEMI_PAM4_S_NAME);
        alias.insert(ACC_REG_PEMI_BER_S_NAME);
        alias.insert(ACC_REG_PEMI_FERC_S_NAME);

        alias.insert(ACC_REG_PEMI_MODULE_P_NAME);
        alias.insert(ACC_REG_PEMI_SNR_P_NAME);
        alias.insert(ACC_REG_PEMI_LASER_P_NAME);
        alias.insert(ACC_REG_PEMI_PAM4_P_NAME);
        alias.insert(ACC_REG_PEMI_BER_P_NAME);
        alias.insert(ACC_REG_PEMI_FERC_P_NAME);
    }

    // Alias - All registers
    {
        set_string_t &alias = this->register_dependencies[ACC_REG_ALL_NAME];
        alias.insert(ACC_REG_PCAM_NAME);
        alias.insert(ACC_REG_ALL_PEMI_NAME);
        alias.insert(ACC_REG_ALL_PCI_NAME);
        alias.insert(ACC_REG_ALL_PDDR_NAME);
        alias.insert(ACC_REG_ALL_MPCNT_NAME);
        alias.insert(ACC_REG_PTAS_NAME);
        alias.insert(ACC_REG_PPLL_NAME);
        alias.insert(ACC_REG_SLRIP_NAME);
        alias.insert(ACC_REG_SLREG_NAME);
        alias.insert(ACC_REG_SLCCT_NAME);
        alias.insert(ACC_REG_PPAMP_NAME);
        alias.insert(ACC_REG_SLLM_NAME);
        alias.insert(ACC_REG_PMDR_NAME);
        alias.insert(ACC_REG_PPHCR_NAME);
        alias.insert(ACC_REG_SLSIR_NAME);
        alias.insert(ACC_REG_PMCR_NAME);
        alias.insert(ACC_REG_PTYS_NAME);
        alias.insert(ACC_REG_PEUCG_NAME);
        alias.insert(ACC_REG_MFCR_NAME);
        alias.insert(ACC_REG_MFSM_NAME);
        alias.insert(ACC_REG_MFSL_NAME);
        alias.insert(ACC_REG_FORE_NAME);
        alias.insert(ACC_REG_MFNR_NAME);
        alias.insert(ACC_REG_MTCAP_NAME);
        alias.insert(ACC_REG_MTMP_NAME);
        alias.insert(ACC_REG_MTWE_NAME);
        alias.insert(ACC_REG_MVCAP_NAME);
        alias.insert(ACC_REG_MVCR_NAME);
        alias.insert(ACC_REG_MSPS_NAME);
        alias.insert(ACC_REG_MSGI_NAME);
        alias.insert(ACC_REG_PPBMP_NAME);
        alias.insert(ACC_REG_MRCS_NAME);
    }

    // Alias - All Diagnostic Data pages
    {
        set_string_t &alias = this->register_dependencies[DIAGNOSTIC_DATA_ALL_NAME];
        alias.insert(DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_NAME);
        alias.insert(DIAGNOSTIC_DATA_PHYS_LAYER_CNTRS_NAME);
        alias.insert(DIAGNOSTIC_DATA_PCIE_CNTRS_NAME);
        alias.insert(DIAGNOSTIC_DATA_PCIE_LANES_NAME);
        alias.insert(DIAGNOSTIC_DATA_PCIE_TIMERS_NAME);
        alias.insert(DIAGNOSTIC_DATA_TROUBLESHOOTING_INFO_NAME);
        alias.insert(DIAGNOSTIC_DATA_OPERATION_INFO_NAME);
        alias.insert(DIAGNOSTIC_DATA_PHY_INFO_NAME);
        alias.insert(DIAGNOSTIC_DATA_MODULE_INFO_NAME);
        alias.insert(DIAGNOSTIC_DATA_LINK_DOWN_INFO_NAME);
        alias.insert(DIAGNOSTIC_DATA_LINK_UP_INFO_NAME);
        alias.insert(DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_NAME);
        alias.insert(DIAGNOSTIC_DATA_RS_FEC_HISTOGRAM_NAME);
        alias.insert(DIAGNOSTIC_DATA_PLR_COUNTERS_NAME);
        alias.insert(DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_NAME);
        alias.insert(DIAGNOSTIC_DATA_INFINIBAND_GENERAL_COUNTERS_NAME);
    }

    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::InitBERTable()
{
    IBDIAGNET_ENTER;

    this->Init7NMBERTable();
    this->Init16NMBERTable();

    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::Init7NMBERTable()
{
    IBDIAGNET_ENTER;

    //7nm_NDR_14
    vec_ber_thresholds_val_t &ndr_14_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_DACS)];
    ndr_14_dacs.resize(NUM_OF_BER_TYPES);
    ndr_14_dacs[IB_RAW_BER].error = 1e-6;
    ndr_14_dacs[IB_RAW_BER].warning = 5e-7;
    ndr_14_dacs[IB_RAW_BER].normal = 1e-7;
    ndr_14_dacs[IB_EFFECTIVE_BER].error = 1e-11;
    ndr_14_dacs[IB_EFFECTIVE_BER].warning = 5e-12;
    ndr_14_dacs[IB_EFFECTIVE_BER].normal = 1e-12;
    ndr_14_dacs[IB_SYMBOL_BER].error = 1e-14;
    ndr_14_dacs[IB_SYMBOL_BER].warning = 5e-15;
    ndr_14_dacs[IB_SYMBOL_BER].normal = 1e-15;

    vec_ber_thresholds_val_t &ndr_14_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACC)];
    ndr_14_acc = ndr_14_dacs;

    vec_ber_thresholds_val_t &ndr_14_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACTIVE)];
    ndr_14_active.resize(NUM_OF_BER_TYPES);
    ndr_14_active[IB_RAW_BER].error = 1e-5;
    ndr_14_active[IB_RAW_BER].warning = 5e-6;
    ndr_14_active[IB_RAW_BER].normal = 1e-6;
    ndr_14_active[IB_EFFECTIVE_BER].error = 1e-11;
    ndr_14_active[IB_EFFECTIVE_BER].warning = 5e-12;
    ndr_14_active[IB_EFFECTIVE_BER].normal = 1e-12;
    ndr_14_active[IB_SYMBOL_BER].error = 1e-14;
    ndr_14_active[IB_SYMBOL_BER].warning = 5e-15;
    ndr_14_active[IB_SYMBOL_BER].normal = 1e-15;

    vec_ber_thresholds_val_t &ndr_14_active_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACTIVE_DID)];
    ndr_14_active_did = ndr_14_active;

    //7nm_NDR_13
    vec_ber_thresholds_val_t &ndr_13_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_DACS)];
    ndr_13_dacs.resize(NUM_OF_BER_TYPES);
    ndr_13_dacs[IB_RAW_BER].error = 1e-6;
    ndr_13_dacs[IB_RAW_BER].warning = 5e-7;
    ndr_13_dacs[IB_RAW_BER].normal = 1e-7;
    ndr_13_dacs[IB_EFFECTIVE_BER].error = 1e-11;
    ndr_13_dacs[IB_EFFECTIVE_BER].warning = 5e-12;
    ndr_13_dacs[IB_EFFECTIVE_BER].normal = 1e-12;
    ndr_13_dacs[IB_SYMBOL_BER].error = 1e-14;
    ndr_13_dacs[IB_SYMBOL_BER].warning = 5e-15;
    ndr_13_dacs[IB_SYMBOL_BER].normal = 1e-15;

    vec_ber_thresholds_val_t &ndr_13_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACC)];
    ndr_13_acc = ndr_13_dacs;

    vec_ber_thresholds_val_t &ndr_13_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACTIVE)];
    ndr_13_active.resize(NUM_OF_BER_TYPES);
    ndr_13_active[IB_RAW_BER].error = 1e-5;
    ndr_13_active[IB_RAW_BER].warning = 5e-6;
    ndr_13_active[IB_RAW_BER].normal = 1e-6;
    ndr_13_active[IB_EFFECTIVE_BER].error = 1e-11;
    ndr_13_active[IB_EFFECTIVE_BER].warning = 5e-12;
    ndr_13_active[IB_EFFECTIVE_BER].normal = 1e-12;
    ndr_13_active[IB_SYMBOL_BER].error = 1e-14;
    ndr_13_active[IB_SYMBOL_BER].warning = 5e-15;
    ndr_13_active[IB_SYMBOL_BER].normal = 1e-15;

    vec_ber_thresholds_val_t &ndr_13_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACTIVE_DID)];
    ndr_13_did = ndr_13_active;

    //7nm_NDR_4
    vec_ber_thresholds_val_t &ndr_4_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_DACS)];
    ndr_4_dacs.resize(NUM_OF_BER_TYPES);
    ndr_4_dacs[IB_RAW_BER].error = 1e-7;
    ndr_4_dacs[IB_RAW_BER].warning = 5e-8;
    ndr_4_dacs[IB_RAW_BER].normal = 1e-8;
    ndr_4_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    ndr_4_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    ndr_4_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    ndr_4_dacs[IB_SYMBOL_BER].error = 1e-13;
    ndr_4_dacs[IB_SYMBOL_BER].warning = 5e-14;
    ndr_4_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &ndr_4_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACC)];
    ndr_4_acc = ndr_4_dacs;

    vec_ber_thresholds_val_t &ndr_4_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACTIVE)];
    ndr_4_active.resize(NUM_OF_BER_TYPES);
    ndr_4_active[IB_RAW_BER].error = 1e-5;
    ndr_4_active[IB_RAW_BER].warning = 5e-6;
    ndr_4_active[IB_RAW_BER].normal = 1e-6;
    ndr_4_active[IB_EFFECTIVE_BER].error = 1e-13;
    ndr_4_active[IB_EFFECTIVE_BER].warning = 5e-14;
    ndr_4_active[IB_EFFECTIVE_BER].normal = 1e-14;
    ndr_4_active[IB_SYMBOL_BER].error = 1e-13;
    ndr_4_active[IB_SYMBOL_BER].warning = 5e-14;
    ndr_4_active[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &ndr_4_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_100,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACTIVE_DID)];
    ndr_4_did = ndr_4_active;

    //7nm_HDR_13
    vec_ber_thresholds_val_t &hdr_13_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_DACS)];
    hdr_13_dacs.resize(NUM_OF_BER_TYPES);
    hdr_13_dacs[IB_RAW_BER].error = 1e-07;
    hdr_13_dacs[IB_RAW_BER].warning = 5e-08;
    hdr_13_dacs[IB_RAW_BER].normal = 1e-08;
    hdr_13_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    hdr_13_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    hdr_13_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    hdr_13_dacs[IB_SYMBOL_BER].error = 1e-14;
    hdr_13_dacs[IB_SYMBOL_BER].warning = 5e-15;
    hdr_13_dacs[IB_SYMBOL_BER].normal = 1e-15;

    vec_ber_thresholds_val_t &hdr_13_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACC)];
    hdr_13_acc = hdr_13_dacs;

    vec_ber_thresholds_val_t &hdr_13_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACTIVE)];
    hdr_13_active = hdr_13_dacs;

    vec_ber_thresholds_val_t &hdr_13_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACTIVE_DID)];
    hdr_13_did = hdr_13_dacs;


    //7nm_HDR_4
    vec_ber_thresholds_val_t &hdr_4_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_DACS)];
    hdr_4_dacs.resize(NUM_OF_BER_TYPES);
    hdr_4_dacs[IB_RAW_BER].error = 1e-07;
    hdr_4_dacs[IB_RAW_BER].warning = 5e-08;
    hdr_4_dacs[IB_RAW_BER].normal = 1e-08;
    hdr_4_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    hdr_4_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    hdr_4_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    hdr_4_dacs[IB_SYMBOL_BER].error = 1e-13;
    hdr_4_dacs[IB_SYMBOL_BER].warning = 5e-14;
    hdr_4_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &hdr_4_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACC)];
    hdr_4_acc = hdr_4_dacs;

    vec_ber_thresholds_val_t &hdr_4_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACTIVE)];
    hdr_4_active = hdr_4_dacs;

    vec_ber_thresholds_val_t &hdr_4_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACTIVE_DID)];
    hdr_4_did = hdr_4_dacs;

    //7nm_HDR_14
    vec_ber_thresholds_val_t &hdr_14_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_DACS)];
    hdr_14_dacs.resize(NUM_OF_BER_TYPES);
    hdr_14_dacs[IB_RAW_BER].error = 1e-07;
    hdr_14_dacs[IB_RAW_BER].warning = 5e-08;
    hdr_14_dacs[IB_RAW_BER].normal = 1e-08;
    hdr_14_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    hdr_14_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    hdr_14_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    hdr_14_dacs[IB_SYMBOL_BER].error = 1e-14;
    hdr_14_dacs[IB_SYMBOL_BER].warning = 5e-15;
    hdr_14_dacs[IB_SYMBOL_BER].normal = 1e-15;

    vec_ber_thresholds_val_t &hdr_14_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACC)];
    hdr_14_acc = hdr_14_dacs;

    vec_ber_thresholds_val_t &hdr_14_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACTIVE)];
    hdr_14_active = hdr_14_dacs;

    vec_ber_thresholds_val_t &hdr_14_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACTIVE_DID)];
    hdr_14_did = hdr_14_dacs;

    //7nm_EDR_3
    vec_ber_thresholds_val_t &edr_3_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_LL_RS_FEC,
                                                MT_DACS)];
    edr_3_dacs.resize(NUM_OF_BER_TYPES);
    edr_3_dacs[IB_RAW_BER].error = 1e-12;
    edr_3_dacs[IB_RAW_BER].warning = 5e-13;
    edr_3_dacs[IB_RAW_BER].normal = 1e-13;
    edr_3_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    edr_3_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    edr_3_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    edr_3_dacs[IB_SYMBOL_BER].error = 1e-13;
    edr_3_dacs[IB_SYMBOL_BER].warning = 5e-14;
    edr_3_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &edr_3_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACC)];
    edr_3_acc = edr_3_dacs;

    vec_ber_thresholds_val_t &edr_3_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACTIVE)];
    edr_3_active = edr_3_dacs;

    vec_ber_thresholds_val_t &edr_3_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACTIVE_DID)];
    edr_3_did = edr_3_dacs;

    //7nm_EDR_2
    vec_ber_thresholds_val_t &edr_2_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_RS_FEC,
                                                MT_DACS)];
    edr_2_dacs.resize(NUM_OF_BER_TYPES);
    edr_2_dacs[IB_RAW_BER].error = 1e-12;
    edr_2_dacs[IB_RAW_BER].warning = 5e-13;
    edr_2_dacs[IB_RAW_BER].normal = 1e-13;
    edr_2_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    edr_2_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    edr_2_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    edr_2_dacs[IB_SYMBOL_BER].error = 1e-13;
    edr_2_dacs[IB_SYMBOL_BER].warning = 5e-14;
    edr_2_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &edr_2_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_RS_FEC,
                                                MT_ACC)];
    edr_2_acc = edr_2_dacs;

    vec_ber_thresholds_val_t &edr_2_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_RS_FEC,
                                                MT_ACTIVE)];
    edr_2_active = edr_2_dacs;

    vec_ber_thresholds_val_t &edr_2_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_RS_FEC,
                                                MT_ACTIVE_DID)];
    edr_2_did = edr_2_dacs;

    //7nm_EDR_0
    vec_ber_thresholds_val_t &edr_0_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_NO_FEC,
                                                MT_DACS)];
    edr_0_dacs.resize(NUM_OF_BER_TYPES);
    edr_0_dacs[IB_RAW_BER].error = 1e-13;
    edr_0_dacs[IB_RAW_BER].warning = 5e-14;
    edr_0_dacs[IB_RAW_BER].normal = 1e-14;
    edr_0_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    edr_0_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    edr_0_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    edr_0_dacs[IB_SYMBOL_BER].error = 1e-13;
    edr_0_dacs[IB_SYMBOL_BER].warning = 5e-14;
    edr_0_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &edr_0_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_NO_FEC,
                                                MT_ACC)];
    edr_0_acc = edr_0_dacs;

    vec_ber_thresholds_val_t &edr_0_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_NO_FEC,
                                                MT_ACTIVE)];
    edr_0_active = edr_0_dacs;

    vec_ber_thresholds_val_t &edr_0_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_7nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_NO_FEC,
                                                MT_ACTIVE_DID)];
    edr_0_did = edr_0_dacs;

    IBDIAGNET_RETURN_VOID;

}

void PhyDiag::Init16NMBERTable()
{
    IBDIAGNET_ENTER;

    //16nm_HDR_13
    vec_ber_thresholds_val_t &hdr_13_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_DACS)];
    hdr_13_dacs.resize(NUM_OF_BER_TYPES);
    hdr_13_dacs[IB_RAW_BER].error = 1e-5;
    hdr_13_dacs[IB_RAW_BER].warning = 5e-6;
    hdr_13_dacs[IB_RAW_BER].normal = 1e-6;
    hdr_13_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    hdr_13_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    hdr_13_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    hdr_13_dacs[IB_SYMBOL_BER].error = 1e-13;
    hdr_13_dacs[IB_SYMBOL_BER].warning = 5e-14;
    hdr_13_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &hdr_13_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACC)];
    hdr_13_acc = hdr_13_dacs;

    vec_ber_thresholds_val_t &hdr_13_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACTIVE)];
    hdr_13_active = hdr_13_dacs;

    vec_ber_thresholds_val_t &hdr_13_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_544_514_PLR,
                                                MT_ACTIVE_DID)];
    hdr_13_did = hdr_13_dacs;

    //16nm_HDR_4
    vec_ber_thresholds_val_t &hdr_4_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_DACS)];
    hdr_4_dacs.resize(NUM_OF_BER_TYPES);
    hdr_4_dacs[IB_RAW_BER].error = 1e-05;
    hdr_4_dacs[IB_RAW_BER].warning = 5e-06;
    hdr_4_dacs[IB_RAW_BER].normal = 1e-06;
    hdr_4_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    hdr_4_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    hdr_4_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    hdr_4_dacs[IB_SYMBOL_BER].error = 1e-13;
    hdr_4_dacs[IB_SYMBOL_BER].warning = 5e-14;
    hdr_4_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &hdr_4_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACC)];
    hdr_4_acc = hdr_4_dacs;

    vec_ber_thresholds_val_t &hdr_4_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACTIVE)];
    hdr_4_active = hdr_4_dacs;

    vec_ber_thresholds_val_t &hdr_4_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_RS_FEC_544_514,
                                                MT_ACTIVE_DID)];
    hdr_4_did = hdr_4_dacs;

    //16nm_HDR_14
    vec_ber_thresholds_val_t &hdr_14_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_DACS)];
    hdr_14_dacs.resize(NUM_OF_BER_TYPES);
    hdr_14_dacs[IB_RAW_BER].error = 1e-05;
    hdr_14_dacs[IB_RAW_BER].warning = 5e-06;
    hdr_14_dacs[IB_RAW_BER].normal = 1e-06;
    hdr_14_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    hdr_14_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    hdr_14_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    hdr_14_dacs[IB_SYMBOL_BER].error = 1e-13;
    hdr_14_dacs[IB_SYMBOL_BER].warning = 5e-14;
    hdr_14_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &hdr_14_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACC)];
    hdr_14_acc = hdr_14_dacs;

    vec_ber_thresholds_val_t &hdr_14_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACTIVE)];
    hdr_14_active = hdr_14_dacs;

    vec_ber_thresholds_val_t &hdr_14_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_MLNX_RS_271_257_PLR,
                                                MT_ACTIVE_DID)];
    hdr_14_did = hdr_14_dacs;

    //16nm_HDR_3
    vec_ber_thresholds_val_t &hdr_3_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_LL_RS_FEC,
                                                MT_DACS)];
    hdr_3_dacs.resize(NUM_OF_BER_TYPES);
    hdr_3_dacs[IB_RAW_BER].error = 1e-05;
    hdr_3_dacs[IB_RAW_BER].warning = 5e-06;
    hdr_3_dacs[IB_RAW_BER].normal = 1e-06;
    hdr_3_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    hdr_3_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    hdr_3_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    hdr_3_dacs[IB_SYMBOL_BER].error = 1e-13;
    hdr_3_dacs[IB_SYMBOL_BER].warning = 5e-14;
    hdr_3_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &hdr_3_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACC)];
    hdr_3_acc = hdr_3_dacs;

    vec_ber_thresholds_val_t &hdr_3_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACTIVE)];
    hdr_3_active = hdr_3_dacs;

    vec_ber_thresholds_val_t &hdr_3_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_50,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACTIVE_DID)];
    hdr_3_did = hdr_3_dacs;

    //16nm_EDR_3
    vec_ber_thresholds_val_t &edr_3_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_LL_RS_FEC,
                                                MT_DACS)];
    edr_3_dacs.resize(NUM_OF_BER_TYPES);
    edr_3_dacs[IB_RAW_BER].error = 1e-12;
    edr_3_dacs[IB_RAW_BER].warning = 5e-13;
    edr_3_dacs[IB_RAW_BER].normal = 1e-13;
    edr_3_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    edr_3_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    edr_3_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    edr_3_dacs[IB_SYMBOL_BER].error = 1e-13;
    edr_3_dacs[IB_SYMBOL_BER].warning = 5e-14;
    edr_3_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &edr_3_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACC)];
    edr_3_acc = edr_3_dacs;

    vec_ber_thresholds_val_t &edr_3_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACTIVE)];
    edr_3_active = edr_3_dacs;

    vec_ber_thresholds_val_t &edr_3_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_LL_RS_FEC,
                                                MT_ACTIVE_DID)];
    edr_3_did = edr_3_dacs;

    //16nm_EDR_2
    vec_ber_thresholds_val_t &edr_2_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_RS_FEC,
                                                MT_DACS)];
    edr_2_dacs.resize(NUM_OF_BER_TYPES);
    edr_2_dacs[IB_RAW_BER].error = 1e-09;
    edr_2_dacs[IB_RAW_BER].warning = 5e-10;
    edr_2_dacs[IB_RAW_BER].normal = 1e-10;
    edr_2_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    edr_2_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    edr_2_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    edr_2_dacs[IB_SYMBOL_BER].error = 1e-13;
    edr_2_dacs[IB_SYMBOL_BER].warning = 5e-14;
    edr_2_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &edr_2_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_RS_FEC,
                                                MT_ACC)];
    edr_2_acc = edr_2_dacs;

    vec_ber_thresholds_val_t &edr_2_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_RS_FEC,
                                                MT_ACTIVE)];
    edr_2_active = edr_2_dacs;

    vec_ber_thresholds_val_t &edr_2_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_RS_FEC,
                                                MT_ACTIVE_DID)];
    edr_2_did = edr_2_dacs;

    //16nm_EDR_0
    vec_ber_thresholds_val_t &edr_0_dacs =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_NO_FEC,
                                                MT_DACS)];
    edr_0_dacs.resize(NUM_OF_BER_TYPES);
    edr_0_dacs[IB_RAW_BER].error = 1e-13;
    edr_0_dacs[IB_RAW_BER].warning = 5e-14;
    edr_0_dacs[IB_RAW_BER].normal = 1e-14;
    edr_0_dacs[IB_EFFECTIVE_BER].error = 1e-13;
    edr_0_dacs[IB_EFFECTIVE_BER].warning = 5e-14;
    edr_0_dacs[IB_EFFECTIVE_BER].normal = 1e-14;
    edr_0_dacs[IB_SYMBOL_BER].error = 1e-13;
    edr_0_dacs[IB_SYMBOL_BER].warning = 5e-14;
    edr_0_dacs[IB_SYMBOL_BER].normal = 1e-14;

    vec_ber_thresholds_val_t &edr_0_acc =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_NO_FEC,
                                                MT_ACC)];
    edr_0_acc = edr_0_dacs;

    vec_ber_thresholds_val_t &edr_0_active =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_NO_FEC,
                                                MT_ACTIVE)];
    edr_0_active = edr_0_dacs;

    vec_ber_thresholds_val_t &edr_0_did =
        this->ber_thresholds_table[BER_thresholds_key(SerDes_16nm,
                                                IB_LINK_SPEED_25,
                                                IB_FEC_NO_FEC,
                                                MT_ACTIVE_DID)];
    edr_0_did = edr_0_dacs;

    IBDIAGNET_RETURN_VOID;
}

bool PhyDiag::GetModuleInfo(const IBPort &port, DDModuleInfo &container,
                            const DDModuleInfo *&p_module_info)
{
    IBDIAGNET_ENTER;

    bool ret1 = this->GetModuleInfoByAccRegister(port, p_module_info);
    if (p_module_info)
        IBDIAGNET_RETURN(true);

    bool ret2 = this->GetModuleInfoByDiagnosticData(port, container, p_module_info);
    if (p_module_info)
       IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(ret1 || ret2);
}

bool PhyDiag::GetModuleInfoByAccRegister(const IBPort &port,
                                        const DDModuleInfo *&p_module_info)
{
    IBDIAGNET_ENTER;

    p_module_info = NULL;

    if (this->module_info_acc_reg_index == UNKNOWN_PAGE_INDX) {
        this->module_info_acc_reg_index = -1;

        for (size_t i = 0; i < this->reg_handlers_vec.size(); ++i) {
            if (this->reg_handlers_vec[i] && //TODO::consider use register index instead of name
                    this->reg_handlers_vec[i]->GetPReg()->GetName() == ACC_REG_PDDR_MODULE_NAME) {
                this->module_info_acc_reg_index = static_cast<int>(i);
                break;
            }
        }
    }

    if (this->module_info_acc_reg_index < 0)
        IBDIAGNET_RETURN(false);

    AccRegPortHandler *p_module =
                dynamic_cast<AccRegPortHandler*>(this->reg_handlers_vec[this->module_info_acc_reg_index]);
    if (!p_module)
        IBDIAGNET_RETURN(false);

    if (!port.p_node)
        IBDIAGNET_RETURN(false);

    AccRegKey *p_key = p_module->CreateKey(port.p_node->guid_get(), port.guid_get(), port.num);
    map_akey_areg::const_iterator itr = p_module->data_map.find(p_key);
    if (itr != p_module->data_map.end())
        p_module_info = &itr->second.regs.pddr_module;

    delete p_key;
    IBDIAGNET_RETURN(true);
}

bool PhyDiag::GetModuleInfoByDiagnosticData(const IBPort &port,
                                            DDModuleInfo &container,
                                            const DDModuleInfo *&p_module_info)
{
    IBDIAGNET_ENTER;

    p_module_info = NULL;

    //this->module_info_dd_index is set to UNKNOWN_PAGE_INDX anytime the plugin runs.
    if (this->module_info_dd_index == UNKNOWN_PAGE_INDX) {
        this->module_info_dd_index = -1;

        for (size_t i = 0; i < this->diagnostic_data_vec.size(); ++i) {
            if (this->diagnostic_data_vec[i] &&
                    this->diagnostic_data_vec[i]->GetPageId() == DIAGNOSTIC_DATA_MODULE_INFO_PAGE) {
                this->module_info_dd_index = static_cast<int>(i);
                break;
            }
        }
    }

    //BER thresholds caluclation depends on the the "Module Info" diagnostic data page
    //we will not proceed if it does not exist
    if (this->module_info_dd_index < 0)
        IBDIAGNET_RETURN(false);

    //if a node does not support the "Module Info" diagnostic page
    //or by any reason a MAD is not found for the node return the worst "Media Type" scenario
    const VS_DiagnosticData *p_curr_data_module_info =
                this->getPhysLayerPortCounters(port.createIndex, this->module_info_dd_index);
    if (!p_curr_data_module_info)
        IBDIAGNET_RETURN(true);

    p_module_info = &container;
    DDModuleInfo_unpack(&container, (uint8_t *)&(p_curr_data_module_info->data_set));
    IBDIAGNET_RETURN(true);
}

MediaType PhyDiag::GetMediaType(const IBPort &port)
{
    IBDIAGNET_ENTER;

    DDModuleInfo module_info;
    const DDModuleInfo *p_module_info = NULL;
    if (!this->GetModuleInfo(port, module_info, p_module_info))
        IBDIAGNET_RETURN(MT_UNKNOWN);

    //if a node does not support the "Module Info" diagnostic page or "PDDR Register"
    //or by any reason a MAD is not found for the node return the worst "Media Type" scenario
    if (!p_module_info)
        IBDIAGNET_RETURN(MT_DACS);

    if (p_module_info->cable_technology >= 0xC && p_module_info->cable_technology <= 0xF)
        IBDIAGNET_RETURN(MT_ACC);

    // Passive_copper_cable
    if (p_module_info->cable_type == 3)
        IBDIAGNET_RETURN(MT_DACS);

    // Direct Drive module
    if (p_module_info->did_cap)
        IBDIAGNET_RETURN(MT_ACTIVE_DID);

    // Either Active_cable - (active copper / optics) or Optical_Module - (separated)
    if (p_module_info->cable_type == 1 || p_module_info->cable_type == 2)
        IBDIAGNET_RETURN(MT_ACTIVE);

    IBDIAGNET_RETURN(MT_UNKNOWN);
}

const vec_ber_thresholds_val_t *PhyDiag::GetBerThresholdEntry(const IBPort &port)
{
    IBDIAGNET_ENTER;

    SerDesVersion technology = SerDes_NA;
    if (port.p_node) {
        if (port.p_node->serdes_version != SerDes_NA &&
                port.p_node->serdes_version <= SerDes_16nm)
            technology = SerDes_16nm;
        else
            technology = port.p_node->serdes_version;
    }

    map_ber_thresholds_t::const_iterator itr =
            this->ber_thresholds_table.find(BER_thresholds_key(technology,
                                                    port.get_internal_speed(),
                                                    port.get_fec_mode(),
                                                    this->GetMediaType(port)));
    IBDIAGNET_RETURN(itr != this->ber_thresholds_table.end() ? &itr->second : NULL);
}

int PhyDiag::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
                                      this->can_send_mads_by_lid) ||
        !this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n",
                    CHECK_NAME_PHY_CNTRS_RETRIEVE);
        PRINT("\n");
    }

    //reset the app of the nodes, to track supported capabilities
    //related to this plugin
    this->p_ibdiag->ResetAppData(true);

    if (this->to_load_ber_thresholds_table) {
        if (this->ParseBERThresholdTable())
            ERR_PRINT("Parse BER Threshold Table was finished with errors, "
                      "continuing with default values\n\n");
        else
            INFO_PRINT("Parse BER Threshold Table was finished successfully\n\n");
    }

    // initialize the p_phy_data on the nodes
    int rc = this->InitPhyDataOnNodes();
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(0);
}

int PhyDiag::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    int rc = 0;
    int rc2;

    if (this->to_get_phy_info) {
        if (this->can_send_mads_by_lid) {
            // Page Identification
            // This page must be the first one who is being sent!
            this->diagnostic_data_vec.insert(
                    this->diagnostic_data_vec.end(),
                    {
                        new DiagnosticDataPageIdentification(),
                        new DiagnosticDataPhysLayerCntrs(),           // PPCNT Phy counters
                        new DiagnosticDataTroubleshootingInfo(),      // PDDR Troubleshooting Info
                        new DiagnosticDataOperationInfo(),            // PDDR Operation Info
                        new DiagnosticDataPhyInfo(),                  // PDDR Phy Info
                        new DiagnosticDataLinkDownInfo(),             // PDDR Link Down Info
                        new DiagnosticDataLinkUpInfo(),               // PDDR Link Up Info
                        new DiagnosticDataRSHistograms(),             // PPCNT RS Histograms
                        new DiagnosticDataPLRCounters(),              // PPCNT PLR Counters
                        new DiagnosticDataPhyStatistics(),            // PPCNT Phy Statistics
                        new DiagnosticDataInfiniBandGeneralCounters() // PPCNT Phy Statistics
                        // PDDR Module Info
                        //new DiagnosticDataModuleInfo(this->to_get_cable_disconnected_ports),
                        // PDDR Latched Flag Info
                        //new DiagnosticDataLatchedFlagInfo(this->to_get_cable_disconnected_ports)
                    });


            if (this->IsEnabledByFilter(DIAGNOSTIC_DATA_MODULE_INFO_NAME) ||
                    this->IsEnabledByFilter(DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_NAME))
                p_ibdiag->SetCableExported(true);

            string method_name = this->to_reset_counters ? "Reset" : "Build";
            INFO_PRINT("%s PHY Counters DB\n", method_name.c_str());
            for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size(); ++dd_idx) {

                DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];
                p_dd->SetIBDiag(this->p_ibdiag);

                if (!this->IsEnabledByFilter(p_dd->GetName()))
                    continue;

                if (p_dd->GetDDType() != DD_PHY_TYPE)
                    continue;

                INFO_PRINT("%s %s DB\n", method_name.c_str(), this->to_reset_counters ?
                                                                  SECTION_PHY_CNTRS :
                                                                  p_dd->GetSectionHeader().c_str());
                rc2 = this->BuildPhyCountersDB(this->phy_errors,
                                               dd_idx);
                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(this->phy_errors,
                                               this->to_reset_counters ?
                                                   CHECK_NAME_PHY_CNTRS_RESET((string)SECTION_PHY_CNTRS) :
                                                   CHECK_NAME_RETRIEVE(p_dd->GetSectionHeader()),
                                               rc2,
                                               IBDIAG_ERR_CODE_FABRIC_ERROR,
                                               this->num_errors,
                                               this->num_warnings,
                                               true);
                if (rc)
                    IBDIAGNET_RETURN(rc);
            }
        }

        AccRegHandler   *p_areg_mfcr    = nullptr;
        AccRegHandler   *p_areg_pmdr    = nullptr;
        AccRegHandler   *p_areg_mtcap   = nullptr;
        AccRegHandler   *p_areg_mvcap   = nullptr;

        // make indexes list for PPBMP
        vector_uint64 indexes;
        indexes.push_back(PPBMP_RAW_BER_RS);
        indexes.push_back(PPBMP_EFFECTIVE_BER);
        indexes.push_back(PPBMP_SYMBOL_BER);

        this->reg_handlers_vec.insert(
                this->reg_handlers_vec.end(),
                {
                    // Register removed, because we have problems with part of registers that was fetch without PCAM.
                    // In PCAM part of registers disabled, but can be retrived without errors.
                    // new AccRegNodeHandler(new PCAMRegister(this)),

                    new AccRegPortLaneHandler(new SLRGRegister(this)),
                    new AccRegPortLaneHandler(new SLLMRegister(this)),
                    new AccRegPortLaneHandler(new SLRPRegister(this)),
                    new AccRegPortLaneHandler(new SLTPRegister(this)),

                    new AccRegNodeHandler(new PTASRegister(this)),

                    p_areg_pmdr = new AccRegPortHandler(new PMDRRegister(this)),
                    new AccRegPortHandler(new PPHCRRegister(this)),
                    new AccRegPortHandler(new PMCRRegister(this)),
                    new AccRegPortHandler(new PTYSRegister(this)),

                    new AccRegGroupHandler(new PPLLRegister(this)),

                    // PDDR
                    new AccRegPortHandler(new PDDRModuleInfoRegister(this)),
                    new AccRegPortHandler(new PDDRLatchedFlagInfoRegister(this)),

                    // fans registers
                    p_areg_mfcr = new AccRegNodeHandler(new MFCRRegister(this)),
                    new AccRegNodeSensorHandler(new MFSMRegister(this), p_areg_mfcr),
                    new AccRegNodeSensorHandler(new MFSLRegister(this), p_areg_mfcr),
                    new AccRegNodeSensorHandler(new MFNRRegister(this), p_areg_mfcr),
                    new AccRegNodeHandler(new FORERegister(this)),

                    // temperature registers
                    p_areg_mtcap = new AccRegNodeHandler(new MTCAPRegister(this)),
                    new AccRegNodeSensorHandler(new MTMPRegister(this), p_areg_mtcap),
                    new AccRegNodeHandler(new MTWERegister(this)),

                    // current/voltage register
                    p_areg_mvcap = new AccRegNodeHandler(new MVCAPRegister(this)),
                    new AccRegNodeSensorHandler(new MVCRRegister(this), p_areg_mvcap),

                    // power supplies register
                    new AccRegNodeHandler(new MSPSRegister(this)),

                    // System General Information register
                    new AccRegNodeHandler(new MSGIRegister(this)),

                    // NEW HDR registers
                    new AccRegPortLaneHandler(new SLRIPRegister(this), SLRIP_IB_SEL_NUM),
                    new AccRegPortLaneHandler(new SLREGRegister(this), SLREG_EIDX_NUM),
                    new AccRegPortLaneHandler(new SLCCTRegister(this), SLCCT_CYCLES_NUM),
                    new AccRegPortLaneHandler(new SLSIRRegister(this)),

                    // PEMI
                    new AccRegPortHandler(new PEMI_Module_Samples_Register(this)),
                    new AccRegPortHandler(new PEMI_SNR_Samples_Register(this)),
                    new AccRegPortHandler(new PEMI_Laser_Samples_Register(this)),
                    new AccRegPortHandler(new PEMI_PAM4_Samples_Register(this)),
                    new AccRegPortHandler(new PEMI_PRE_FEC_BER_Samples_Register(this)),
                    new AccRegPortHandler(new PEMI_FERC_Samples_Register(this)),
                    new AccRegPortHandler(new PEMI_Module_Properties_Register(this)),
                    new AccRegPortHandler(new PEMI_SNR_Properties_Register(this)),
                    new AccRegPortHandler(new PEMI_Laser_Properties_Register(this)),
                    new AccRegPortHandler(new PEMI_PAM4_Properties_Register(this)),
                    new AccRegPortHandler(new PEMI_PRE_FEC_BER_Properties_Register(this)),
                    new AccRegPortHandler(new PEMI_FERC_Properties_Register(this)),

                    // PRTL register
                    new AccRegPortHandler(new PRTLRegister(this)),

                    // PPBMP Register
                    new AccRegPortIndexHandler(new PPBMPRegister(this), indexes),

                    // MRCS register
                    new AccRegNodeHandler(new MRCSRegister(this))
                });

        // Enable combined cable stage
        if (this->IsEnabledByFilter(ACC_REG_PDDR_MODULE_NAME) ||
                this->IsEnabledByFilter(ACC_REG_PDDR_LATCHED_NAME))
            p_ibdiag->SetCableExported(true);

        // UPHY-PEUCG - run only if we get JSON files.
        if (!UPHY::DB::instance().empty())
            this->reg_handlers_vec.insert(
                    this->reg_handlers_vec.end(),
                    {
                        new AccRegNodeHandler(new PEUCG_Ver_Register(this)),
                        new AccRegSpecificHandler(new PEUCG_DLN_Register(this), PORT_LANE_HEADER),
                        new AccRegSpecificHandler(new PEUCG_CLN_Register(this, &p_areg_pmdr->data_map), GROUP_HEADER)
                    });

        if (this->to_get_ppamp)
            this->reg_handlers_vec.push_back(
                    new AccRegTypeGroupHandler(new PPAMPRegister(this),
                            MAX_OPAMP_GROUP_TYPE, PPAMP_IDX_DATA_SIZE));

        INFO_PRINT("Build PHY Registers DB\n");
        for (uint32_t i = 0; i < this->reg_handlers_vec.size(); ++i) {
            AccRegHandler *p_areg_handler = this->reg_handlers_vec[i];
            if (!this->IsEnabledByFilter(p_areg_handler->GetPReg()->GetName()))
                continue;

            INFO_PRINT("Build %s DB\n", p_areg_handler->GetPReg()->GetSectionName().c_str());
            rc2 = p_areg_handler->BuildDB(this->phy_errors);

            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(this->phy_errors,
                                           CHECK_NAME_RETRIEVE(p_areg_handler->
                                                               GetPReg()->
                                                               GetSectionName()),
                                           rc2,
                                           IBDIAG_ERR_CODE_FABRIC_ERROR,
                                           this->num_errors,
                                           this->num_warnings,
                                           true);
        }
    }
    //TODO chech if rc !+0 , if yes return

    if (this->to_get_pci_info) {

        AccRegHandler *p_areg_mpein = new AccRegSpecificHandler(new MPEINRegister(this), PCI_DPN_HEADER);

        AccRegHandler *p_areg_mpir =
                new AccRegSpecificHandler(new MPIRRegister(this, p_areg_mpein->data_map), PCI_DPN_HEADER);

        AccRegHandler *p_areg_slrg =
                new AccRegSpecificHandler(
                        new SLRGRegister(this, ACC_REG_PNAT_OOB_PORT,
                                ACC_REG_SLRG_PCI_INTERNAL_SECTION_NAME,
                                &p_areg_mpein->data_map,
                                &p_areg_mpir->data_map),
                        PCI_PORT_LANE_HEADER);


        AccRegHandler *p_areg_slrp =
                new AccRegSpecificHandler(
                        new SLRPRegister(this, ACC_REG_PNAT_OOB_PORT,
                                ACC_REG_SLRP_PCI_INTERNAL_SECTION_NAME,
                                &p_areg_mpein->data_map,
                                &p_areg_mpir->data_map),
                        PCI_PORT_LANE_HEADER);

        AccRegHandler *p_areg_sltp =
                new AccRegSpecificHandler(
                        new SLTPRegister(this, ACC_REG_PNAT_OOB_PORT,
                                ACC_REG_SLTP_PCI_INTERNAL_SECTION_NAME,
                                &p_areg_mpein->data_map,
                                &p_areg_mpir->data_map),
                        PCI_PORT_LANE_HEADER);

        AccRegHandler *p_areg_slrip =
                new AccRegSpecificHandler(
                        new SLRIPRegister(this, ACC_REG_PNAT_OOB_PORT,
                                ACC_REG_SLRIP_PCI_INTERNAL_SECTION_NAME,
                                &p_areg_mpein->data_map,
                                &p_areg_mpir->data_map),
                        PCI_PORT_LANE_HEADER);



        this->pci_reg_handlers_vec.insert(
                this->pci_reg_handlers_vec.end(),
                {
                    p_areg_mpein,
                    p_areg_mpir,
                    p_areg_slrg,
                    p_areg_slrp,
                    p_areg_sltp,
                    p_areg_slrip
                });

        if (this->to_get_mpcnt) {
            AccRegHandler *p_areg_mpcnt_cnt =
                    new AccRegSpecificHandler(new MPCNT_PCIe_Counters_Register(this, p_areg_mpein->data_map), PCI_DPN_HEADER);

            AccRegHandler *p_areg_mpcnt_lcnt =
                    new AccRegSpecificHandler(new MPCNT_PCIe_Lane_Counters_Register(this, p_areg_mpein->data_map), PCI_DPN_HEADER);

            AccRegHandler *p_areg_mpcnt_timers =
                    new AccRegSpecificHandler(new MPCNT_PCIe_Timers_Register(this, p_areg_mpein->data_map), PCI_DPN_HEADER);

            AccRegHandler *p_areg_mpcnt_phys =
                    new AccRegSpecificHandler(new MPCNT_PCIe_Phys_Layer_Register(this, p_areg_mpein->data_map), PCI_DPN_HEADER);


            this->pci_reg_handlers_vec.insert(
                    this->pci_reg_handlers_vec.end(),
                    {
                        p_areg_mpcnt_cnt,
                        p_areg_mpcnt_lcnt,
                        p_areg_mpcnt_timers,
                        p_areg_mpcnt_phys,
                    });
        }

        INFO_PRINT("Build PCI Registers DB\n");
        for (uint32_t i = 0; i < this->pci_reg_handlers_vec.size(); ++i) {
            AccRegHandler *p_areg_handler = this->pci_reg_handlers_vec[i];
            if (!this->IsEnabledByFilter(p_areg_handler->GetPReg()->GetName()))
                continue;

            INFO_PRINT("Build %s DB\n", p_areg_handler->GetPReg()->GetSectionName().c_str());
            rc2 = p_areg_handler->BuildDB(this->phy_errors);

            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(this->phy_errors,
                                           CHECK_NAME_RETRIEVE(p_areg_handler->
                                                               GetPReg()->
                                                               GetSectionName()),
                                           rc2,
                                           IBDIAG_ERR_CODE_FABRIC_ERROR,
                                           this->num_errors,
                                           this->num_warnings,
                                           true);
        }
        this->p_ibdiag->SetRailDataCollected(true);

        if (this->can_send_mads_by_lid && !this->to_get_mpcnt) {

            this->diagnostic_data_vec.insert(
                    this->diagnostic_data_vec.end(),
                    {
                        new DiagnosticDataPCIECntrs(),      // Diagnostic Data PCIe Counters
                        new DiagnosticDataPCIETimers(),     // Diagnostic Data PCIe Timers and States
                        new DiagnosticDataPCIELanes(),      // Diagnostic Data PCIe Lanes Counters
                    });

            string method_name = this->to_reset_pci_counters ? "Reset" : "Build";
            INFO_PRINT("%s PCI Counters DB\n", method_name.c_str());
            for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size(); ++dd_idx) {
                DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

                if (!this->IsEnabledByFilter(p_dd->GetName()))
                    continue;

                if (p_dd->GetDDType() != DD_PCI_TYPE)
                    continue;
                INFO_PRINT("%s %s DB\n", method_name.c_str(), this->to_reset_pci_counters ?
                                                                  SECTION_PCI_CNTRS :
                                                                  p_dd->GetSectionHeader().c_str());
                rc2 = this->BuildPCICountersDB(this->phy_errors,
                                               dd_idx,
                                               &p_areg_mpein->data_map);

                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(this->phy_errors,
                                               this->to_reset_pci_counters ?
                                                   CHECK_NAME_PHY_CNTRS_RESET((string)SECTION_PCI_CNTRS) :
                                                   CHECK_NAME_RETRIEVE(p_dd->GetSectionHeader()),
                                               rc2,
                                               IBDIAG_ERR_CODE_FABRIC_ERROR,
                                               this->num_errors,
                                               this->num_warnings,
                                               true);
                if (rc)
                    IBDIAGNET_RETURN(rc);
            }
        }
    }// end of get_p_info

    if (this->p_ibdiag->IsRailValidation()) {
        if (!this->to_get_pci_info) {
           AccRegHandler *p_areg_mpein = new AccRegSpecificHandler(new MPEINRegister(this), PCI_DPN_HEADER);
           this->pci_reg_handlers_vec.push_back(p_areg_mpein);

           AccRegHandler *p_areg_mpir = new AccRegSpecificHandler(new MPIRRegister(this, p_areg_mpein->data_map), PCI_DPN_HEADER);
           this->pci_reg_handlers_vec.push_back(p_areg_mpir);

           for (uint32_t i = 0; i < this->pci_reg_handlers_vec.size(); ++i) {
                AccRegHandler *p_areg_handler = this->pci_reg_handlers_vec[i];
                rc2 = p_areg_handler->BuildDB(this->phy_errors);

                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(this->phy_errors,
                                        CHECK_NAME_RETRIEVE(p_areg_handler->
                                                            GetPReg()->
                                                            GetSectionName()),
                                        rc2,
                                        IBDIAG_ERR_CODE_FABRIC_ERROR,
                                        this->num_errors,
                                        this->num_warnings,
                                        true);
            }

            this->p_ibdiag->SetRailDataCollected(true);
        }
    }

    IBDIAGNET_RETURN(rc);
}

bool PhyDiag::IsActive()
{
    IBDIAGNET_ENTER;

    if (this->IsFinished())
       return false;

    bool is_active = Stage::IsActive();
    if (is_active) {
       IBDIAGNET_RETURN(true);
    }

    if (p_ibdiag->IsRailValidation()) {
        this->MarkStageAsActive();
        IBDIAGNET_RETURN(true);
    }

    IBDIAGNET_RETURN(is_active);
}

//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit / 3 - some other error
int PhyDiag::HandleOption(string name, string value)
{
    IBDIAGNET_ENTER;

    bool bool_flag_val = true;

    if (value == OPTION_DEF_VAL_NULL)
        IBDIAGNET_RETURN(0);

    if (name == OPTION_PHY_INFO_GET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPhyInfoGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PHY_CABLE_DISCONNECTED) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetToGetCableDisconnectedPorts(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PPAMP_GET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPPAMPGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SHOW_CAP_REG) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetCapRegShow(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PHY_INFO_RESET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPhyInfoGet(bool_flag_val);
        this->SetPhyInfoClear(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PHY_CABLE_FULL_DATA) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPhyCableFullData(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PCI_INFO_GET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPciInfoGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_PCI_INFO_RESET) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetPciInfoClear(bool_flag_val);
        this->SetPciInfoGet(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BER_THRESHOLD_ERROR) {
        ERR_PRINT("Flag " OPTION_BER_THRESHOLD_ERROR " is deprecated\n");
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BER_THRESHOLD_WARNING) {
        ERR_PRINT("Flag " OPTION_BER_THRESHOLD_WARNING " is deprecated\n");
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_SHOW_BER_WARNINGS) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetShowBERWarnings(bool_flag_val);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_BER_THRESHOLD_TABLE) {
        this->SetBERThresholdTableFile(value);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_CREATE_BER_THRESHOLD_TABLE) {
        if (this->CreateBERThresholdTable(value))
            IBDIAGNET_RETURN(3);
        IBDIAGNET_RETURN(2); // Exit with no error
    } else if (name == OPTION_ENABLED_REGS) {
        if (this->ParseRegistersList(value, this->enabled_regs))
            IBDIAGNET_RETURN(3);
        this->CheckRegisterDependencies();
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_DISABLED_REGS) {
        if (this->ParseRegistersList(value, this->disabled_regs))
            IBDIAGNET_RETURN(3);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_SLRG_REG) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetSLRGGet(bool_flag_val);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_MPCNT_REG) {
        this->ParseBoolValue(value, bool_flag_val);
        this->SetMPCNTGet(bool_flag_val);
        IBDIAGNET_RETURN(0);

    } else if (name == OPTION_ACC_REG_PRIORITY) {

        string result;

        if (this->ParseAccRegPriorityValue(value, result)) {
            ERR_PRINT("wrong value for " OPTION_ACC_REG_PRIORITY " flag."
                      " options are: smp | gmp.\n");
            IBDIAGNET_RETURN(3);
        }

        this->SetAccRegPriority(result == "smp" ? VIA_SMP : VIA_GMP);

        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_UPHY) {
        if (this->ParseUPHYFileList(value))
            IBDIAGNET_RETURN(3);
        IBDIAGNET_RETURN(0);
    } else if (name == OPTION_UPHY_DUMPS) {
        if (this->ParseUPHYDumps(value)) {
            ERR_PRINT_ARGS("Illegal value for \'--%s\' option: %s\n",
                           OPTION_UPHY_DUMPS, value.c_str());
            IBDIAGNET_RETURN(3);
        }
        IBDIAGNET_RETURN(0);
    }

    //if we reached here than unknown option
    IBDIAGNET_RETURN(1);
}

AccRegVia_t PhyDiag::acc_reg_priority = NOT_SUP_ACC_REG;

int PhyDiag::ParseAccRegPriorityValue(string value, string &result)
{
    IBDIAGNET_ENTER;

    if (!(strncasecmp(value.c_str(), "smp", strlen("smp") + 1))) {
        result = "smp";
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    if (!(strncasecmp(value.c_str(), "gmp", strlen("gmp") + 1))) {
        result = "gmp";
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    IBDIAGNET_RETURN(1);
}

int PhyDiag::ParseRegistersList(const string &registers_list_str, set_string_t & registers_list)
{
    IBDIAGNET_ENTER;
    string::size_type index = 0;

    do {
        string::size_type x = registers_list_str.find_first_of(',', index);
        string register_str = registers_list_str.substr(index,
                (x == string::npos ? registers_list_str.size() : x) - index);

        if (register_str.empty()) {
            ERR_PRINT("Register name in list cannot be empty %s\n", registers_list_str.c_str() + index);
            IBDIAGNET_RETURN(3);
        }

        transform(register_str.begin(), register_str.end(), register_str.begin(), ::tolower);
        registers_list.insert(register_str);

        index = (x == string::npos ? string::npos : x + 1);
    } while (index != string::npos);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void PhyDiag::CheckRegisterDependencies()
{
    IBDIAGNET_ENTER;

    while (true) {
        set_string_t::size_type num_of_registers = this->enabled_regs.size();
        for (set_string_t::iterator it_reg = this->enabled_regs.begin();
                it_reg != this->enabled_regs.end(); ++it_reg) {
             map_string_2_set_string_t::iterator it_reg_dep =
                                                     this->register_dependencies.find(*it_reg);
             if (it_reg_dep != this->register_dependencies.end()) {
                 for (set_string_t::iterator it = it_reg_dep->second.begin();
                         it != it_reg_dep->second.end(); ++it) {
                      this->enabled_regs.insert(*it);
                 }
             }
        }

        if (num_of_registers == this->enabled_regs.size())
            break;
    }

    IBDIAGNET_RETURN_VOID;
}

int PhyDiag::LoadUPHYFile(const string &filename)
{
    INFO_PRINT("UPHY load JSON file '%s'\n", filename.c_str());

    UPHY::JsonLoader loader(filename);

    string errors = loader.errors().str();
    if (!errors.empty())
        WARN_PRINT("%s\n", errors.c_str());

    return loader.dataset() != NULL;
}

int PhyDiag::ParseUPHYFileList(const string &uphy_path)
{
    IBDIAGNET_ENTER;

    DIR *dir;
    dir = opendir(uphy_path.c_str());
    int errsv = errno;
    string input_type;
    if (dir) {
        input_type = "directory";
        struct dirent *file_entry;

        while ((file_entry = readdir(dir))) {
               string uphy_file = file_entry->d_name;
               string::size_type x = uphy_file.find_last_of('.');
               if (x == string::npos)
                   continue;

               if (uphy_file.substr(x + 1) != "json")
                   continue;

               LoadUPHYFile(uphy_path + SEPERATOR + uphy_file);
        }

        closedir(dir);
    } else {
        if ((errsv != ENOENT) && (errsv != ENOTDIR)) {
            ERR_PRINT("Failed to open UPHY directory: %s, error: %s\n", uphy_path.c_str(),
                      strerror(errsv));
            IBDIAGNET_RETURN(3);
        }

        input_type = "list";
        string::size_type index = 0;

        do {
            string::size_type x = uphy_path.find_first_of(',', index);
            string uphy_file = uphy_path.substr(index,
                    (x == string::npos ? uphy_path.size() : x) - index);

            if (uphy_file.empty()) {
                WARN_PRINT("UPHY file name cannot be empty %s\n", uphy_path.c_str() + index);
            } else {
                LoadUPHYFile(uphy_file);
            }

            index = (x == string::npos ? string::npos : x + 1);
        } while (index != string::npos);
    }

    if (UPHY::DB::instance().empty()) {
        ERR_PRINT("No UPHY json file in %s: %s\n", input_type.c_str(), uphy_path.c_str());
        IBDIAGNET_RETURN(3);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PhyDiag::ParseUPHYDumps(const string &uphy_dumps_args)
{
    IBDIAGNET_ENTER;

    string::size_type   index = 0;

    do {
        string::size_type x = uphy_dumps_args.find_first_of(',', index);
        string            v = uphy_dumps_args.substr(
                                index, (x == string::npos ? uphy_dumps_args.size() : x) - index );

        if(!v.empty()) {
            std::transform(v.begin(), v.end(), v.begin(), ::tolower);
            uphy_dumps.push_back(v);
        }
        else
        {
            ERR_PRINT("Type cannot be empty %s\n", uphy_dumps_args.c_str() + index);
            IBDIAGNET_RETURN(3);
        }

        index = (x == string::npos ? string::npos : x + 1);
    } while (index != string::npos);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

bool PhyDiag::IsEnabledByFilter(const string &register_name)
{
    IBDIAGNET_ENTER;
    if (register_name == "slrg")
        return this->to_get_slrg;

    if (this->enabled_regs.count(register_name) || this->enabled_regs.empty())
        IBDIAGNET_RETURN(!this->disabled_regs.count(register_name))

    IBDIAGNET_RETURN(false);
}

int PhyDiag::ParseBERThresholdTable()
{
    IBDIAGNET_ENTER;

    this->to_load_ber_thresholds_table = false; // Load table only once in case of long run iterations

    int rc = IBDIAG_SUCCESS_CODE;

    IBDiag::PrintFileTimestamp(ber_thresholds_table_file, "BER Threshold Table");

    ifstream fs(this->ber_thresholds_table_file.c_str());
    if (fs.fail()) {
        ERR_PRINT("Failed to open configuration file: %s\n",
                  this->ber_thresholds_table_file.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    int lineNum = 0;

    while (fs.good()) {
        char line[1024] = {'\0'};
        fs.getline(line, 1024);
        lineNum++;

        // skip empty lines and comments
        if (!line[0] || line[0] == '#')
            continue;

        char techn_str[128]= {0};
        char speed_str[128] = {0};
        char fec_op_str[128] = {0};
        char media_type_str[128] = {0};
        char ber_type_str[128] = {0};
        char err_str[128] = {0};
        char warn_str[128] = {0};
        char normal_str[128] = {0};
        char tail_str[128] = {0};

        if (sscanf(line, "%[^:] : %[^:] : %[^:] : %[^:] : %[^=]"
                " = %[-0-9.(default)] , %[-0-9.(default)], %[-0-9.(default)] %s",
                techn_str, speed_str, fec_op_str, media_type_str,
                ber_type_str, err_str, warn_str, normal_str, tail_str) != 8) {
            ERR_PRINT("Failed to parse line %d, invalid format: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        SerDesVersion technology = char2serdesversion(techn_str);
        if (technology == SerDes_NA) {
            ERR_PRINT("Failed to parse line %d, invalid tecnology: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        IBLinkSpeed speed = char_name2speed(speed_str);
        if (speed == IB_UNKNOWN_LINK_SPEED) {
            ERR_PRINT("Failed to parse line %d, invalid speed: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        IBFECMode fec_op = char2fec(fec_op_str);
        if (fec_op == IB_FEC_NA) {
            ERR_PRINT("Failed to parse line %d, invalid FEC option: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        MediaType media_type = char2media_type(media_type_str);
        if (media_type == MT_UNKNOWN) {
            ERR_PRINT("Failed to parse line %d, invalid Media type: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        IBBERType ber_type = short_char2ber_type(ber_type_str);
        if (ber_type >= NUM_OF_BER_TYPES) {
            ERR_PRINT("Failed to parse line %d, invalid BER type: %s\n", lineNum, line);
            rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            continue;
        }

        map_ber_thresholds_t::iterator itr =
                this->ber_thresholds_table.find(BER_thresholds_key(technology,
                                                                speed,
                                                                fec_op,
                                                                media_type));
        if (itr == ber_thresholds_table.end()) {
            BER_thresholds_key key = BER_thresholds_key(technology,
                                                      speed,
                                                      fec_op,
                                                      media_type);
            // insert the new key with a default vector
            vec_ber_thresholds_val_t value(NUM_OF_BER_TYPES);
            std::fill(value.begin(), value.end(), BER_thresholds_warning_error(1, 1, 1));
            auto insert_ret = this->ber_thresholds_table.insert(
                std::pair<BER_thresholds_key, vec_ber_thresholds_val_t>(
                    key, value));

            itr = insert_ret.first;
        }

        if (string(err_str) != BER_THRESHOLD_TABLE_DEFAULT_STR) {
            double error = 0;
            CLEAR_STRUCT(tail_str);
            if (sscanf(err_str ,"%lf %s", &error, tail_str) != 1) {
                ERR_PRINT("Invalid error value in line %d: %s, "
                          "continuing with default error value: %e\n",
                          lineNum, line, itr->second[ber_type].error);
                rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            } else
                itr->second[ber_type].error = error;
        }

        if (string(warn_str) != BER_THRESHOLD_TABLE_DEFAULT_STR) {
            double warning = 0;
            CLEAR_STRUCT(tail_str);
            if (sscanf(warn_str ,"%lf %s", &warning, tail_str) != 1) {
                ERR_PRINT("Invalid warning value in line %d: %s, "
                          "continuing with default warning value: %e\n",
                          lineNum, line, itr->second[ber_type].warning);
                rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            } else
                itr->second[ber_type].warning = warning;
        }

        if (string(normal_str) != BER_THRESHOLD_TABLE_DEFAULT_STR) {
            double normal = 0;
            CLEAR_STRUCT(tail_str);
            if (sscanf(normal_str ,"%lf %s", &normal, tail_str) != 1) {
                ERR_PRINT("Invalid normal value in line %d: %s, "
                          "continuing with default normal value: %e\n",
                          lineNum, line, itr->second[ber_type].normal);
                rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
            } else
                itr->second[ber_type].normal = normal;
        }
    } //while

    IBDIAGNET_RETURN(rc);
}

int PhyDiag::CreateBERThresholdTable(const string &file_path)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    INFO_PRINT("Create template BER Threshold Table file: %s\n\n", file_path.c_str());

    ofstream fs(file_path.c_str());
    if (fs.fail()) {
        ERR_PRINT("Failed to create/open BER Threshold Table file: %s\n", file_path.c_str());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    fs << "# " << IBDIAGNET_VERSION << endl;
    fs << "#" << endl;
    fs << "#" << endl;

    fs << "# Technology Mapping:" << endl;
    fs << "#" << endl;
    fs << "# " << serdesversion2char(SerDes_16nm) << " = "
       << serdesversion2char(SerDes_16nm) << "nm" << endl
       << "# " << serdesversion2char(SerDes_7nm) << " = "
       << serdesversion2char(SerDes_7nm) << "nm" << endl;
    fs << "#" << endl;

    fs << "# Speed Values:" << endl;
    fs << "#" << endl;
    fs << "# " << speed2char_name(IB_LINK_SPEED_25) << ", "
       << speed2char_name(IB_LINK_SPEED_50) << ", "
       << speed2char_name(IB_LINK_SPEED_100) << endl;
    fs << "#" << endl;

    fs << "# FEC mode Mapping:" << endl;
    fs << "#" << endl;
    fs << "# " << fec2char(IB_FEC_NO_FEC) << " = "
               << DEC(IB_FEC_NO_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_FIRECODE_FEC) << " = "
               << DEC(IB_FEC_FIRECODE_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_RS_FEC) << " = "
               << DEC(IB_FEC_RS_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_LL_RS_FEC) << " = "
               << DEC(IB_FEC_LL_RS_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_RS_FEC_544_514) << " = "
               << DEC(IB_FEC_RS_FEC_544_514) << endl;
    fs << "# " << fec2char(IB_FEC_MLNX_STRONG_RS_FEC) << " = "
               << DEC(IB_FEC_MLNX_STRONG_RS_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_MLNX_LL_RS_FEC) << " = "
               << DEC(IB_FEC_MLNX_LL_RS_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_MLNX_ADAPTIVE_RS_FEC) << " = "
               << DEC(IB_FEC_MLNX_ADAPTIVE_RS_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_MLNX_COD_FEC) << " = "
               << DEC(IB_FEC_MLNX_COD_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_MLNX_ZL_FEC) << " = "
               << DEC(IB_FEC_MLNX_ZL_FEC) << endl;
    fs << "# " << fec2char(IB_FEC_MLNX_RS_544_514_PLR) << " = "
               << DEC(IB_FEC_MLNX_RS_544_514_PLR) << endl;
    fs << "# " << fec2char(IB_FEC_MLNX_RS_271_257_PLR) << " = "
               << DEC(IB_FEC_MLNX_RS_271_257_PLR) << endl;
    fs << "#" << endl;

    fs << "# Media type Values:" << endl;
    fs << "#" << endl;
    fs << "# ";
    for (u_int8_t media_type = 1;  media_type < MT_NUM_OF_MEDIA_TYPES; media_type++ )
        fs << media_type2char((MediaType)media_type) << (media_type + 1 == MT_NUM_OF_MEDIA_TYPES ? "" : ", ");
    fs << endl << "#" << endl;

    fs << "# BER type Mapping:" << endl;
    fs << "#" << endl;
    for (u_int8_t ber_type = 0; ber_type < NUM_OF_BER_TYPES; ber_type++ )
         fs << "# " << ber_type2char((IBBERType)ber_type) << " = "
            << ber_type2short_char((IBBERType)ber_type) << endl;
    fs << "#" << endl;

    fs << "# " << "<Technology>:<Speed>:<FEC_Mode_ID>:<Media_Type>:<BER_type> = "
               << "<error_value>,<warning_value>,<normal_value>" << endl;
    fs << "#" << endl;
    fs << endl;

    for (const auto &entry : this->ber_thresholds_table) {
        const BER_thresholds_key &key = entry.first;
        const vec_ber_thresholds_val_t &vec_ber_thresholds_values = entry.second;
        for (size_t ber_type = 0; ber_type < vec_ber_thresholds_values.size(); ber_type++) {
            fs << serdesversion2char(key.technology) << ':'
               << speed2char_name(key.speed) << ':'
               << DEC(key.fec_mode) << ':'
               << media_type2char(key.media_type) << ':'
               << ber_type2short_char(static_cast<IBBERType>(ber_type)) << '='
               << FLOAT(vec_ber_thresholds_values[ber_type].error, 1, true)   << ','
               << FLOAT(vec_ber_thresholds_values[ber_type].warning, 1, true) << ','
               << FLOAT(vec_ber_thresholds_values[ber_type].normal, 1, true)
               << endl;
        }
    }

    fs.close();
    INFO_PRINT("BER Threshold Table file was created: %s\n", file_path.c_str());

    IBDIAGNET_RETURN(rc);
}

int PhyDiag::HandleSpecialPorts(IBNode *p_curr_node,
                                IBPort *p_curr_port,
                                u_int32_t i)
{
    IBDIAGNET_ENTER;

    struct SMP_MlnxExtPortInfo *p_curr_mepi = this->p_ibdm_extended_info->
            getSMPMlnxExtPortInfo(p_curr_port->createIndex);
    if (!p_curr_mepi) {
        IBDIAG_LOG(TT_LOG_LEVEL_ERROR,
                "Error DB - Can't find MlnxExtPortInfo for " \
                "node=%s, port=%u\n",
                p_curr_node->name.c_str(), i);
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    // check if this special port support Diagnostic Data
    if (!(p_curr_mepi->SpecialPortCapabilityMask & (1 << EnSPCapIsDiagnosticDataSupported)))
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

/*
 * private methods
 */
//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS /
//IBDIAGNET_ERR_CODE_DB_ERR / IBDIAGNET_ERR_CODE_NO_MEM /
//IBDIAGNET_ERR_CODE_FABRIC_ERROR
int PhyDiag::BuildPhyCountersDB(list_p_fabric_general_err &phy_errors,
                                u_int32_t dd_idx)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = &forwardClbck<PhyDiag,
                    &PhyDiag::PhyCountersGetClbck>;
    clbck_data.m_p_obj = this;


    DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

    clbck_data.m_data1 = (void *)(uintptr_t)dd_idx;
    clbck_data.m_data2 = p_dd;

    ProgressBarPorts p_progress_bar;
    clbck_data.m_p_progress_bar = &p_progress_bar;

    for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
            nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s", (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_dd->IsDDPageSupportNodeType(p_curr_node))
            continue;

        if (p_dd->GetPageId() != DIAGNOSTIC_DATA_PAGE_IDENTIFICATION_PAGE) {
            struct VS_DiagnosticData *p_vs_diagnostic_data =
                             this->getPhysLayerNodeCounters(p_curr_node->createIndex,
                                                            PAGE_IDENTIFICATION_DD_IDX);
            if (p_vs_diagnostic_data) {
                struct DDPageIdentification page_identification;
                DDPageIdentification_unpack(&page_identification,
                                            (uint8_t *)&(p_vs_diagnostic_data->data_set));
                if (!p_dd->IsDDPageSupportedInNode(&page_identification))
                    continue;
            }
        }

        // skip special nodes
        if (p_curr_node->isSpecialNode())
            continue;

        if (!IsSupportDiagnosticDataCountersFailed(p_curr_node->appData1.val,
                                                  p_dd->GetNotSupportedBit())) {
            if (!this->p_capability_module->IsSupportedGMPCapability(
                    p_curr_node, EnGMPCapIsDiagnosticDataSupported)) {
                p_curr_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;

                phy_errors.push_back(new FabricErrPhyNodeNotSupportCap(p_curr_node,
                        "This device does not support diagnostic data MAD capability"));

                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                continue;
            }
        } else
            continue;

        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)                       //port not exist
                continue;

            //Skip ports that are not in sub fabric.
            if (!p_curr_port->getInSubFabric())
                continue;
            if (p_curr_port->get_internal_state() <= IB_PORT_STATE_DOWN &&
                    (!p_dd->EnableDisconnectedPorts() ||
                        p_curr_node->type == IB_CA_NODE))
                continue;

            // special ports
            if (p_curr_port->isSpecialPort()) {
                int special_rc = HandleSpecialPorts(p_curr_node, p_curr_port, i);

                switch (special_rc) {
                    case IBDIAG_ERR_CODE_DB_ERR:
                        IBDIAGNET_RETURN(special_rc);
                    case IBDIAG_ERR_CODE_FABRIC_ERROR:
                        continue;
                }
            }

            clbck_data.m_data3 = p_curr_port;
            struct VS_DiagnosticData phys_cntrs;

            p_progress_bar.push(p_curr_port);
            // set/get phys' layer counters
            if(this->to_reset_counters)
                this->p_ibis_obj->VSDiagnosticDataPageClear(p_curr_port->base_lid,
                                                            p_dd->IsPerNode() ? 0 : p_curr_port->num,
                                                            p_dd->GetPageId(),
                                                            &phys_cntrs,
                                                            &clbck_data);
            else {
                //get phys' layer counters

                if (!this->to_get_cable_full_data &&
                    ((p_dd->GetPageId() == DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_PAGE ||
                      p_dd->GetPageId() == DIAGNOSTIC_DATA_MODULE_INFO_PAGE) &&
                     p_curr_port->p_combined_cable)) {
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Combined cable info has already been collected "
                          "for the port= %s GUID=" U64H_FMT "\n",
                         p_curr_port->getName().c_str(), p_curr_port->guid_get());
                    continue;
                }

                this->p_ibis_obj->VSDiagnosticDataGet(p_curr_port->base_lid,
                                                      p_dd->IsPerNode() ? 0 : p_curr_port->num,
                                                      p_dd->GetPageId(),
                                                      &phys_cntrs,
                                                      &clbck_data);

            }

            if (clbck_error_state)
                goto exit;
            if (p_dd->IsPerNode())
                break;
        }
    }

exit:
    this->p_ibis_obj->MadRecAll();

    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}


//Returns: IBDIAGNET_SUCCESS_CODE / IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS /
//IBDIAGNET_ERR_CODE_DB_ERR / IBDIAGNET_ERR_CODE_NO_MEM /
//IBDIAGNET_ERR_CODE_FABRIC_ERROR
int PhyDiag::BuildPCICountersDB(list_p_fabric_general_err &phy_errors,
                                u_int32_t dd_idx,
                                map_akey_areg *data_map)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_ibdiag->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = &forwardClbck<PhyDiag,
                &PhyDiag::PCICountersGetClbck>;
    clbck_data.m_p_obj = this;

    DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

    clbck_data.m_data1 = (void *)(uintptr_t)dd_idx;

    ProgressBarPorts progress_bar;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_akey_areg::iterator iter = data_map->begin();
            iter != data_map->end(); ++iter) {
         AccRegKeyDPN *p_acc_reg_key_dpn = (AccRegKeyDPN *)iter->first;
         if (!p_acc_reg_key_dpn)
             continue;

         IBNode *p_curr_node = this->p_discovered_fabric->
                               getNodeByGuid(p_acc_reg_key_dpn->node_guid);
         if (!p_curr_node)
             continue;

         if (this->to_get_phy_info) {
             struct VS_DiagnosticData *p_vs_diagnostic_data =
                                    this->getPhysLayerNodeCounters(p_curr_node->createIndex,
                                                                   PAGE_IDENTIFICATION_DD_IDX);
             if (p_vs_diagnostic_data) {
                 struct DDPageIdentification page_identification;
                 DDPageIdentification_unpack(&page_identification,
                                             (uint8_t *)&(p_vs_diagnostic_data->data_set));
                 if (!p_dd->IsDDPageSupportedInNode(&page_identification))
                     continue;
             }
        }

         if (!IsSupportDiagnosticDataCountersFailed(p_curr_node->appData1.val,
                                                   p_dd->GetNotSupportedBit())) {
             if (!this->p_capability_module->IsSupportedGMPCapability(
                     p_curr_node, EnGMPCapIsDiagnosticDataSupported)) {
                 p_curr_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;

                 phy_errors.push_back(new FabricErrNodeNotSupportCap(p_curr_node,
                         "This device does not support diagnostic data MAD capability"));

                 rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                 continue;
             }
         } else
             continue;

        IBPort *p_curr_port = NULL;

        //get the first valid active port of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
            p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)                       //port not exist
                continue;
            //Skip ports that are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port->is_data_worthy())
                continue;

            break;
         }

         if (!p_curr_port)
             continue;

         clbck_data.m_data2 = p_acc_reg_key_dpn;
         clbck_data.m_data3 = p_curr_port;
         struct VS_DiagnosticData pci_cntrs;
         u_int8_t page_number = (u_int8_t)p_dd->GetPageId();
         u_int32_t attribute_modifier = 0;
         attribute_modifier = ((p_acc_reg_key_dpn->depth & 0x3f) << 24) | (page_number << 16)
                               | (p_acc_reg_key_dpn->pci_node << 8) | p_acc_reg_key_dpn->pci_idx;

         progress_bar.push(p_curr_port);
         //get phys' layer counters
         if (this->to_reset_pci_counters)
             this->p_ibis_obj->VSDiagnosticDataPageClear_AM(p_curr_port->base_lid,
                                                            attribute_modifier,
                                                            &pci_cntrs,
                                                            &clbck_data);
         else
             this->p_ibis_obj->VSDiagnosticDataGet_AM(p_curr_port->base_lid,
                                                      attribute_modifier,
                                                      &pci_cntrs,
                                                      &clbck_data);

         if (clbck_error_state)
             goto exit;
    }

exit:
    this->p_ibis_obj->MadRecAll();

    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void PhyDiag::PhyCountersGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data3);

    if (clbck_error_state || !p_ibdiag)
        return;

    rec_status &= 0x00ff;   //bits [8-15] are class specific - don't need them here
    DiagnosticDataInfo *p_dd = (DiagnosticDataInfo *)clbck_data.m_data2;

    // By Mellanox Vendor Specific MAD Specification (Rev 1.5)
    // In the case of access to unsupported page number, the MAD will return with PageCurrentRevision = 0.
    // Note that revision numbers start from 1. Implementation note: some HCAs do not conform to this
    // but rather return bad MAD status (0x1c - invalid value in one or more fields in attribute or attribute modifier).
    //
    // We check this condition for not marking HCA Node as not support any DiagnosticData mad
    if (rec_status == IBIS_MAD_STATUS_INVALID_FIELD && p_port->p_node->type == IB_CA_NODE) {
        stringstream error_desc;
        error_desc << "The firmware of this device does not support VSDiagnosticData"
                   << p_dd->GetSectionHeader() << "Get, Page ID: " << DEC(p_dd->GetPageId());

        this->phy_errors.push_back(new FabricErrPhyNodeNotSupportCap(p_port->p_node, error_desc.str()));

        IBDIAGNET_RETURN_VOID;
    }

    if (rec_status &&
        IsSupportDiagnosticDataCountersFailed(p_port->p_node->appData1.val,
                                              p_dd->GetNotSupportedBit()))
        IBDIAGNET_RETURN_VOID;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;

        this->phy_errors.push_back(
                new FabricErrPhyNodeNotSupportCap(p_port->p_node,
                        "The firmware of this device does not support VSDiagnosticData"));

        IBDIAGNET_RETURN_VOID;
    } else if (rec_status) {           //failed to get response
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;

        this->phy_errors.push_back(new FabricErrPhyPortNotRespond(p_port, "VSDiagnosticDataGet"));

        IBDIAGNET_RETURN_VOID;
    }

    struct VS_DiagnosticData *p_dd_struct =
        (struct VS_DiagnosticData *)p_attribute_data;

    //check if the requested page in DiagnosticDataInfo is unsupported
    if (p_dd_struct->CurrentRevision == 0 ||
            p_dd_struct->BackwardRevision > p_dd->GetSupportedVersion() ||
            p_dd->GetSupportedVersion() > p_dd_struct->CurrentRevision) {

        p_port->p_node->appData1.val |= p_dd->GetNotSupportedBit();
        stringstream error_desc;

        error_desc << "The firmware of this device does not support VSDiagnosticData"
                   << p_dd->GetSectionHeader() << "Get, Page ID: " << DEC(p_dd->GetPageId())
                   << ", Current Revision: " << +p_dd_struct->CurrentRevision
                   << ", Backward Revision: " << +p_dd_struct->BackwardRevision
                   << ", Supported Version: " << DEC(p_dd->GetSupportedVersion());

        this->phy_errors.push_back(
                new FabricErrPhyNodeNotSupportCap(p_port->p_node, error_desc.str()));

        IBDIAGNET_RETURN_VOID;
    }

    u_int32_t dd_idx = (u_int32_t)(intptr_t)clbck_data.m_data1;

    int rc;
    if (p_dd->IsPerNode())
        rc = this->addPhysLayerNodeCounters(p_port->p_node, *p_dd_struct, dd_idx);
    else
        rc = this->addPhysLayerPortCounters(p_port, *p_dd_struct, dd_idx);

    if (rc)
        //already set last error
        clbck_error_state = rc;

    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::PCICountersGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data)
{
    IBPort *p_port = ProgressBar::complete<IBPort>(clbck_data.m_p_progress_bar, clbck_data.m_data3);
    if (clbck_error_state || !p_ibdiag)
        return;

    rec_status &= 0x00ff;   //bits [8-15] are class specific - don't need them here
    u_int32_t dd_idx = (u_int32_t)(intptr_t)clbck_data.m_data1;
    DiagnosticDataInfo *p_dd = this->diagnostic_data_vec[dd_idx];
    AccRegKey *p_acc_reg_key = (AccRegKey *)clbck_data.m_data2;

    if (rec_status &&
        IsSupportDiagnosticDataCountersFailed(p_port->p_node->appData1.val,
                                              p_dd->GetNotSupportedBit()))
        IBDIAGNET_RETURN_VOID;

    if (rec_status == IBIS_MAD_STATUS_UNSUP_METHOD_ATTR) {         //not supported for this device
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;

        this->phy_errors.push_back(new FabricErrPhyNodeNotSupportCap(p_port->p_node,
                "The firmware of this device does not support VSDiagnosticData"));

        IBDIAGNET_RETURN_VOID;
    } else if (rec_status) {           //failed to get response
        p_port->p_node->appData1.val |= NOT_SUPPORT_DIAGNOSTIC_DATA;

        this->phy_errors.push_back(new FabricErrPhyPortNotRespond(p_port, "VSDiagnosticData"));

        IBDIAGNET_RETURN_VOID;
    }

    struct VS_DiagnosticData *p_dd_struct =
        (struct VS_DiagnosticData *)p_attribute_data;

    //check if the requested page in DiagnosticDataInfo is unsupported
    if (p_dd_struct->CurrentRevision == 0 ||
            p_dd_struct->BackwardRevision > p_dd->GetSupportedVersion() ||
            p_dd->GetSupportedVersion() > p_dd_struct->CurrentRevision) {

        p_port->p_node->appData1.val |= p_dd->GetNotSupportedBit();
        stringstream error_desc;

        error_desc << "The firmware of this device does not support VSDiagnosticData"
                   << p_dd->GetSectionHeader() << "Get, Page ID: " << DEC(p_dd->GetPageId())
                   << ", Current Revision: " << +p_dd_struct->CurrentRevision
                   << ", Backward Revision: " << +p_dd_struct->BackwardRevision
                   << ", Supported Version: " << DEC(p_dd->GetSupportedVersion());

        this->phy_errors.push_back(new FabricErrPhyNodeNotSupportCap(p_port->p_node, error_desc.str()));

        IBDIAGNET_RETURN_VOID;
    }

    int rc = this->addPhysLayerPCICounters(p_acc_reg_key, *p_dd_struct, dd_idx);
    if (rc)
        //already set last error
        clbck_error_state = rc;

    IBDIAGNET_RETURN_VOID;
}

IBNode* PhyDiag::getNodePtr(u_int32_t node_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec(this->nodes_vector, node_index)));
}

IBPort* PhyDiag::getPortPtr(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec(this->ports_vector, port_index)));
}

template <typename T>
void PhyDiag::addPtrToVec(vector<T*>& vector_obj, T *p_obj)
{
    IBDIAG_ENTER;
    if ((vector_obj.size() > p_obj->createIndex + 1) &&
            vector_obj[p_obj->createIndex]) //if exist already
        IBDIAG_RETURN_VOID;
    if (vector_obj.empty() || (vector_obj.size() < p_obj->createIndex + 1))
        for (int i = (int)vector_obj.size(); i < (int)p_obj->createIndex + 1; ++i)
            vector_obj.push_back(NULL);
    vector_obj[p_obj->createIndex] = p_obj;
    IBDIAG_RETURN_VOID;
}

template <typename T, typename V>
int PhyDiag::addDataToVec(vector<T*>& vector_obj,
        T *p_obj,
        vector<V*>& vector_data,
        V& data)
{
    IBDIAG_ENTER;
    if (!p_obj)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((vector_data.size() >= p_obj->createIndex + 1) &&
            vector_data[p_obj->createIndex])
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u)\n",
            typeid(data).name(), p_obj->getName().c_str(), p_obj->createIndex);
    if (vector_data.empty() || (vector_data.size() < p_obj->createIndex + 1))
        for (int i = (int)vector_data.size(); i < (int)p_obj->createIndex + 1; ++i)
            vector_data.push_back(NULL);

    V *p_curr_data = new V;
    *p_curr_data = data;

    vector_data[p_obj->createIndex] = p_curr_data;
    this->addPtrToVec(vector_obj, p_obj);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

template <typename T>
T * PhyDiag::getPtrFromVec(std::vector<T*> & vector_obj, u_int32_t idx)
{
    IBDIAG_ENTER;
    //not exists
    if (idx < 0 || (vector_obj.size() < idx + 1))
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(vector_obj[idx]);
}

template <typename T, typename V>
int PhyDiag::addDataToVecInVec(vector<T*>& vector_obj,
                               T *p_obj,
                               vector< vector<V*> >& vec_of_vectors,
                               u_int32_t data_idx,
                               V& data)
{
    IBDIAG_ENTER;
    if (!p_obj)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    //if exist already
    if ((vec_of_vectors.size() >= p_obj->createIndex + 1) &&
            (vec_of_vectors[p_obj->createIndex].size() >= data_idx + 1))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (%s, index=%u, vec_idx=%u)\n",
            typeid(data).name(), p_obj->getName().c_str(), p_obj->createIndex, data_idx);
    if (vec_of_vectors.empty() || (vec_of_vectors.size() < p_obj->createIndex + 1))
        vec_of_vectors.resize(p_obj->createIndex + 1);

    if (vec_of_vectors[p_obj->createIndex].empty() ||
            (vec_of_vectors[p_obj->createIndex].size() < data_idx + 1))
        for (int i = (int)vec_of_vectors[p_obj->createIndex].size(); i < (int)data_idx + 1; ++i)
            vec_of_vectors[p_obj->createIndex].push_back(NULL);

    V *p_curr_data = new V;
    *p_curr_data = data;

    vec_of_vectors[p_obj->createIndex][data_idx] = p_curr_data;
    this->addPtrToVec(vector_obj, p_obj);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

template <typename T>
T * PhyDiag::getPtrFromVecInVec(std::vector< std::vector<T*> > & vec_of_vectors,
        u_int32_t idx1,
        u_int32_t idx2)
{
    IBDIAG_ENTER;
    //not exists
    if (idx1 < 0 || (vec_of_vectors.size() < idx1 + 1))
        IBDIAG_RETURN(NULL);
    if (idx2 < 0 || (vec_of_vectors[idx1].size() < idx2 + 1))
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(vec_of_vectors[idx1][idx2]);
}

int PhyDiag::addDataToMapInVec(AccRegKey *p_acc_reg_key,
        vec_map_akey_diagnostic_data_t& vec_of_maps,
        u_int32_t data_idx,
        struct VS_DiagnosticData& data)
{
    IBDIAG_ENTER;

    stringstream sstream;
    sstream.str("");
    string key_data_str;

    if (!p_acc_reg_key)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    p_acc_reg_key->DumpKeyData(sstream);
    key_data_str = sstream.str();

    //add
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Adding %s for (Acc Reg Key: %s, vec_idx=%u)\n",
            typeid(data).name(), key_data_str.c_str(), data_idx);

    if (vec_of_maps.empty() ||
            (vec_of_maps.size() < data_idx + 1))
        for (int i = (int)vec_of_maps.size(); i < (int)data_idx + 1; ++i)
            vec_of_maps.push_back(NULL);

    if (!vec_of_maps[data_idx])
        vec_of_maps[data_idx] = new map_akey_diagnostic_data_t(keycomp);

    struct VS_DiagnosticData *p_curr_data = new struct VS_DiagnosticData;
    *p_curr_data = data;

    std::pair<std::map<AccRegKey *, struct VS_DiagnosticData *>::iterator,bool> ret;
    ret = vec_of_maps[data_idx]->insert(pair <AccRegKey *,
                                struct VS_DiagnosticData *>(p_acc_reg_key,
                                                            p_curr_data));

    if (!ret.second) {
        this->SetLastError("%s already exist in map for (Acc Reg Key: %s, vec_idx=%u)\n",
                           typeid(data).name(), key_data_str.c_str(), data_idx);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/**************/
//Returns: SUCCESS_CODE / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int PhyDiag::addPhysLayerPortCounters(
    IBPort *p_port,
    struct VS_DiagnosticData &physCounters,
    u_int32_t dd_idx)
{
    IBDIAG_RETURN(this->addDataToVecInVec(this->ports_vector,
                p_port,
                this->port_cidx_and_dd_idx_to_dd_page,
                dd_idx,
                physCounters));
}

struct VS_DiagnosticData *
PhyDiag::getPhysLayerPortCounters(u_int32_t port_index, u_int32_t dd_idx)
{

 IBDIAG_ENTER;
 IBDIAG_RETURN((this->getPtrFromVecInVec(this->port_cidx_and_dd_idx_to_dd_page, port_index, dd_idx)));
}

int PhyDiag::addPhysLayerNodeCounters(
    IBNode *p_node,
    struct VS_DiagnosticData &physCounters,
    u_int32_t dd_idx)
{
    IBDIAG_RETURN(this->addDataToVecInVec(this->nodes_vector,
                  p_node,
                  this->node_cidx_and_dd_idx_to_dd_page,
                  dd_idx,
                  physCounters));
}


struct VS_DiagnosticData * PhyDiag::getPhysLayerNodeCounters(u_int32_t node_index, u_int32_t dd_idx)
{

 IBDIAG_ENTER;
 IBDIAG_RETURN((this->getPtrFromVecInVec(this->node_cidx_and_dd_idx_to_dd_page, node_index, dd_idx)));
}

int PhyDiag::addPhysLayerPCICounters(
    AccRegKey *p_acc_reg_key,
    struct VS_DiagnosticData &pciCounters,
    u_int32_t dd_idx)
{
    IBDIAG_RETURN(this->addDataToMapInVec(p_acc_reg_key,
                this->dd_idx_and_akey_to_dd_page,
                dd_idx,
                pciCounters));
}

int PhyDiag::addBER(IBPort *p_port, long double ber)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->ports_vector, p_port,
                                     this->ber_vec,
                                     ber));
}

long double* PhyDiag::getBER(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec(this->ber_vec, port_index)));
}

int PhyDiag::addEffBER(IBPort *p_port, long double ber)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->addDataToVec(this->ports_vector, p_port,
                                     this->eff_ber_vec,
                                     ber));
}

long double* PhyDiag::getEffBER(u_int32_t port_index)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN((this->getPtrFromVec(this->eff_ber_vec, port_index)));
}

void PhyDiag::DumpCSVPhyCounters(CSVOut &csv_out, u_int32_t dd_type)
{
    IBDIAGNET_ENTER;

    struct VS_DiagnosticData *p_curr_data = NULL;
    char buffer[1024];
    stringstream sstream;

    for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size();
         ++dd_idx) {
        DiagnosticDataInfo * p_dd = this->diagnostic_data_vec[dd_idx];

        if (p_dd->GetDDType() != dd_type)
            continue;

        if (p_dd->DumpDiagnosticDataHeaderStart(csv_out))
            continue;

        for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
                nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
            IBNode *p_curr_node = (*nI).second;
            if (!p_curr_node->getInSubFabric())
                continue;

            // Diagnostic data pages per node
            if (p_dd->IsPerNode()) {
                p_curr_data = getPhysLayerNodeCounters(p_curr_node->createIndex, dd_idx);
                if (!p_curr_data)
                    continue;

                sstream.str("");
                sstream << PTR(p_curr_node->guid_get())
                        << ',' << DEC(p_curr_data->CurrentRevision);

                p_dd->DumpDiagnosticData(sstream, *p_curr_data);
                csv_out.WriteBuf(sstream.str());

                continue;
            }

            // go over all ports of this node:
            for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
                if (!p_curr_port || !p_curr_port->is_data_worthy())
                    continue;


                p_curr_data = getPhysLayerPortCounters(p_curr_port->createIndex, dd_idx);
                if (!p_curr_data)
                    continue;

                sstream.str("");
                snprintf(buffer, sizeof(buffer),
                         U64H_FMT "," U64H_FMT "," U32D_FMT "," U32D_FMT ",",
                         p_curr_port->p_node->guid_get(), p_curr_port->guid_get(),
                         p_curr_port->num, p_curr_data->CurrentRevision);
                sstream << buffer;

                p_dd->DumpDiagnosticData(sstream, *p_curr_data, p_curr_node);

                sstream << endl;
                csv_out.WriteBuf(sstream.str());
            }
        }

        p_dd->DumpDiagnosticDataHeaderEnd(csv_out);
    }
    IBDIAGNET_RETURN_VOID;
}

bool PhyDiag::CollectAccRegCableInfo(acc_reg_cable_info_map_t & cable_info)
{
    AccRegPortHandler *p_module = NULL;
    AccRegPortHandler *p_latched = NULL;

    for (const auto & x : reg_handlers_vec )
        if (x->GetPReg()->GetName() == ACC_REG_PDDR_MODULE_NAME)
            p_module = dynamic_cast<AccRegPortHandler*>(x);
        else if (x->GetPReg()->GetName() == ACC_REG_PDDR_LATCHED_NAME)
            p_latched = dynamic_cast<AccRegPortHandler*>(x);

    if (!p_module && !p_latched)
        IBDIAGNET_RETURN(false);

    if (p_module)
        for (auto & x : p_module->data_map) {
            ModuleInfoExt &rec = cable_info[x.first].first;

            rec.p_module_info = &x.second.regs.pddr_module;
            rec.module_info_ext = x.second.data.module_info_ext;
        }


    if (p_latched)
        for (auto & x : p_latched->data_map)
            cable_info[x.first].second = &x.second.regs.pddr_latched;

    IBDIAGNET_RETURN(!cable_info.empty());
}

void PhyDiag::DumpCSV_DDCableInfo(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;

    stringstream sstream;

    DiagnosticDataInfo *p_dd = NULL;
    u_int32_t dd_idx_module_info;
    for (dd_idx_module_info = 0; dd_idx_module_info < this->diagnostic_data_vec.size();
            ++dd_idx_module_info) {
         p_dd = this->diagnostic_data_vec[dd_idx_module_info];
         if (p_dd && p_dd->GetPageId() == DIAGNOSTIC_DATA_MODULE_INFO_PAGE)
             break;
    }
    DiagnosticDataModuleInfo *p_dd_module_info = (DiagnosticDataModuleInfo *)p_dd;

    p_dd = NULL;
    u_int32_t dd_idx_latched_flag_info;
    for (dd_idx_latched_flag_info = 0; dd_idx_latched_flag_info < this->diagnostic_data_vec.size();
            ++dd_idx_latched_flag_info) {
         p_dd = this->diagnostic_data_vec[dd_idx_latched_flag_info];
         if (p_dd && p_dd->GetPageId() == DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_PAGE)
             break;
    }
    DiagnosticDataLatchedFlagInfo *p_dd_latched_flag_info = (DiagnosticDataLatchedFlagInfo *)p_dd;

    if (!p_dd_module_info && !p_dd_latched_flag_info)
        IBDIAGNET_RETURN_VOID;

    int section_status = csv_out.DumpStart(SECTION_DD_CABLE_MODULE_INFO);

    if (!section_status) {
        sstream << "NodeGuid,PortGuid,PortNum,";
        DiagnosticDataModuleInfo::DumpModuleInfoHeader(sstream);
        sstream << ',';
        DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoHeader(sstream);
        sstream << endl;
        csv_out.WriteBuf(sstream.str());
    }

    for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
            nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
         IBNode *p_curr_node = (*nI).second;
         if (!p_curr_node->getInSubFabric())
            continue;

         // go over all ports of this node:
         for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
             IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
             if (!p_curr_port)
                 continue;

             //Skip ports that are not in sub fabric.
             if (!p_curr_port->getInSubFabric())
                 continue;
             if (p_curr_port->get_internal_state() <= IB_PORT_STATE_DOWN &&
                     !this->to_get_cable_disconnected_ports)
                 continue;

             struct VS_DiagnosticData *p_curr_data_module_info = NULL;
             if (p_dd_module_info)
                 p_curr_data_module_info = getPhysLayerPortCounters(p_curr_port->createIndex,
                                                                    dd_idx_module_info);

             struct VS_DiagnosticData *p_curr_data_latched_flag_info = NULL;
             if (p_dd_latched_flag_info)
                 p_curr_data_latched_flag_info = getPhysLayerPortCounters(p_curr_port->createIndex,
                                                                          dd_idx_latched_flag_info);

             if (!p_curr_data_module_info && !p_curr_data_latched_flag_info)
                 continue;

            if (!p_curr_port->p_combined_cable)
                this->ExportToIBPort(*p_curr_port, p_curr_data_module_info,
                                          p_curr_data_latched_flag_info);

            if (section_status)
                continue;

             sstream.str("");
             sstream << PTR(p_curr_port->p_node->guid_get()) << ","
                     << PTR(p_curr_port->guid_get()) << ","
                     << DEC(p_curr_port->num) << ",";

             DiagnosticDataModuleInfo::DumpModuleInfoData(sstream, p_curr_data_module_info);
             sstream << ",";
             DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoData(sstream,
                                                                    p_curr_data_latched_flag_info);

             sstream << endl;
             csv_out.WriteBuf(sstream.str());
         }
    }

    csv_out.DumpEnd(SECTION_DD_CABLE_MODULE_INFO);

    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::DumpCSV_AccRegCableInfo(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;

    stringstream             sstream;
    acc_reg_cable_info_map_t data(keycomp);

    int section_status = csv_out.DumpStart(SECTION_ACC_REG_CABLE_MODULE_INFO);

    if (!section_status) {
        sstream << "NodeGuid,PortGuid,PortNum,";
        DiagnosticDataModuleInfo::DumpModuleInfoHeader(sstream);
        sstream << ',';
        DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoHeader(sstream);
        sstream << endl;
        csv_out.WriteBuf(sstream.str());
    }

    if (!CollectAccRegCableInfo(data)) {
        csv_out.DumpEnd(SECTION_ACC_REG_CABLE_MODULE_INFO);
        IBDIAGNET_RETURN_VOID;
    }

    for (const auto & x : data)
    {
        const AccRegKeyPort * key = dynamic_cast<const AccRegKeyPort*>(x.first);

        if (!key)
            continue;

        IBPort * p_port = GetPort(key->node_guid, key->port_num);

        if (!p_port)
            continue;

        if (!p_port->p_combined_cable)
            this->ExportToIBPort(*p_port, &x.second.first, x.second.second);

        if (section_status)
            continue;

        sstream.str("");
        sstream << PTR(p_port->p_node->guid_get()) << ","
                << PTR(p_port->guid_get()) << ","
                << DEC(p_port->num) << ",";

        DiagnosticDataModuleInfo::DumpModuleInfoData(sstream, x.second.first.p_module_info,
                                                    x.second.first.module_info_ext);
        sstream << ",";
        DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoData(sstream, x.second.second);

        sstream << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_ACC_REG_CABLE_MODULE_INFO);

    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::ExportToIBPort(IBPort &port,
                    const VS_DiagnosticData *p_module_info,
                    const VS_DiagnosticData *p_latched_flag_info)
{
    IBDIAGNET_ENTER;

    //we are safe that at least one of input pointers is not NULL
    if (port.p_combined_cable) {
        if (this->to_get_cable_full_data || this->to_reset_counters)
            IBDIAGNET_RETURN_VOID;

        ERR_PRINT("Cable data has already been added to the port: %s\n", port.getName().c_str());
        ++this->num_errors;

        IBDIAGNET_RETURN_VOID;
    }

    port.p_combined_cable = new CombinedCableInfo("PHY_DD",
                                DiagnosticDataModuleInfo::CreateRecord(p_module_info),
                                DiagnosticDataLatchedFlagInfo::CreateRecord(p_latched_flag_info));

    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::ExportToIBPort(IBPort &port,
                    const struct ModuleInfoExt *p_ext,
                    const struct DDLatchedFlagInfo *p_latched_flag_info)
{
    IBDIAGNET_ENTER;

    //we are safe that at least one of input pointers is not NULL
    if (port.p_combined_cable) {
        if (this->to_get_cable_full_data || this->to_reset_counters)
            IBDIAGNET_RETURN_VOID;

        ERR_PRINT("Cable data has already been added to the port: %s\n", port.getName().c_str());
        ++this->num_errors;

        IBDIAGNET_RETURN_VOID;
    }

    port.p_combined_cable = new CombinedCableInfo("PHY_ACC_REG",
                                DiagnosticDataModuleInfo::CreateRecord(p_ext->p_module_info,
                                                            p_ext->module_info_ext),
                                DiagnosticDataLatchedFlagInfo::CreateRecord(p_latched_flag_info));

    IBDIAGNET_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_IO_ERR
int PhyDiag::DumpFile_DDCableInfo(const string &file_name)
{
    IBDIAGNET_ENTER;

    ofstream sout;

    int rc = this->p_ibdiag->OpenFile("PHY DD Cables Information", file_name, sout, false, true);

    if (rc) {
        this->SetLastError("Failed to open PHY DD Cables Information file for writing.");
        IBDIAGNET_RETURN(rc);
    }

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);
    this->DumpFile_DDCableInfo(sout);

    this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_IO_ERR
int PhyDiag::DumpFile_AccRegCableInfo(const string &file_name)
{
    IBDIAGNET_ENTER;

    ofstream sout;

    int rc = this->p_ibdiag->OpenFile("PHY AccReg Cables Information", file_name, sout, false, true);

    if (rc) {
        this->SetLastError("Failed to open PHY AccReg Cables Information file for writing.");
        IBDIAGNET_RETURN(rc);
    }

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);
    this->DumpFile_AccRegCableInfo(sout);

    this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void PhyDiag::DumpFile_DDCableInfo(ofstream &sout)
{
    IBDIAG_ENTER;

    DiagnosticDataInfo *p_dd = NULL;
    u_int32_t dd_idx_module_info;
    for (dd_idx_module_info = 0; dd_idx_module_info < this->diagnostic_data_vec.size();
            ++dd_idx_module_info) {
         p_dd = this->diagnostic_data_vec[dd_idx_module_info];
         if (p_dd && p_dd->GetPageId() == DIAGNOSTIC_DATA_MODULE_INFO_PAGE)
             break;
    }
    DiagnosticDataModuleInfo *p_dd_module_info = (DiagnosticDataModuleInfo *)p_dd;

    p_dd = NULL;
    u_int32_t dd_idx_latched_flag_info;
    for (dd_idx_latched_flag_info = 0; dd_idx_latched_flag_info < this->diagnostic_data_vec.size();
            ++dd_idx_latched_flag_info) {
         p_dd = this->diagnostic_data_vec[dd_idx_latched_flag_info];
         if (p_dd && p_dd->GetPageId() == DIAGNOSTIC_DATA_LATCHED_FLAG_INFO_PAGE)
             break;
    }
    DiagnosticDataLatchedFlagInfo *p_dd_latched_flag_info = (DiagnosticDataLatchedFlagInfo *)p_dd;

    if (!p_dd_module_info && !p_dd_latched_flag_info)
        IBDIAGNET_RETURN_VOID;

    for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
            nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
         IBNode *p_curr_node = (*nI).second;
         if (!p_curr_node->getInSubFabric())
            continue;

         // go over all ports of this node:
         for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
             IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
             if (!p_curr_port || !p_curr_port->is_data_worthy())
                 continue;

             struct VS_DiagnosticData *p_curr_data_module_info = NULL;
             if (p_dd_module_info)
                 p_curr_data_module_info = getPhysLayerPortCounters(p_curr_port->createIndex,
                                                                    dd_idx_module_info);

             struct VS_DiagnosticData *p_curr_data_latched_flag_info = NULL;
             if (p_dd_latched_flag_info)
                 p_curr_data_latched_flag_info = getPhysLayerPortCounters(p_curr_port->createIndex,
                                                                          dd_idx_latched_flag_info);

             if (!p_curr_data_module_info && !p_curr_data_latched_flag_info)
                 continue;

             sout << "-------------------------------------------------------" << endl
                  << "Port=" << +p_curr_port->num
                  << " Lid=" << PTR(p_curr_port->base_lid)
                  << " GUID=" << PTR(p_curr_port->guid_get())
                  << " Port Name=" << p_curr_port->getName() << endl
                  << "-------------------------------------------------------" << endl;
             DiagnosticDataModuleInfo::DumpModuleInfoData(sout, p_curr_data_module_info);
             sout << endl;
             DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoData(sout,
                                                                    p_curr_data_latched_flag_info);

             sout << endl << endl << endl;
         }
    }

    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::DumpFile_AccRegCableInfo(ofstream &sout)
{
    IBDIAGNET_ENTER;

    stringstream                sstream;
    acc_reg_cable_info_map_t    data(keycomp);

    if (!CollectAccRegCableInfo(data))
        IBDIAGNET_RETURN_VOID;

    for (const auto & x : data)
    {
        const AccRegKeyPort * key = dynamic_cast<const AccRegKeyPort*>(x.first);

        if (!key)
            continue;

        IBPort * p_port = GetPort(key->node_guid, key->port_num);

        if (!p_port)
            continue;

        sout << "-------------------------------------------------------" << endl
             << "Port=" << +p_port->num
             << " Lid=" << PTR(p_port->base_lid)
             << " GUID=" << PTR(p_port->guid_get())
             << " Port Name=" << p_port->getName() << endl
             << "-------------------------------------------------------" << endl;

        DiagnosticDataModuleInfo::DumpModuleInfoData(sout, x.second.first.p_module_info,
                                x.second.first.module_info_ext);
        sout << endl;
        DiagnosticDataLatchedFlagInfo::DumpLatchedFlagInfoData(sout, x.second.second);

        sout << endl << endl << endl;
    }

    IBDIAGNET_RETURN_VOID;
}


bool PhyDiag::isSupportFwBER(IBPort *p_port)
{
    IBDIAG_ENTER;

    DiagnosticDataInfo *p_dd;
    VS_DiagnosticData *p_curr_data = NULL;
    u_int32_t dd_idx = 0;

    for (dd_idx = 0; dd_idx < diagnostic_data_vec.size(); ++dd_idx) {
        p_dd = diagnostic_data_vec[dd_idx];
        if (p_dd->GetPageId() == DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE)
            break;
    }

    p_curr_data = getPhysLayerPortCounters(p_port->createIndex, dd_idx);

    if (p_curr_data)
        IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(false);
}


void PhyDiag::DumpCSVPCICounters(CSVOut &csv_out, u_int32_t dd_type)
{
    IBDIAGNET_ENTER;

    AccRegKey *p_curr_acc_reg_key = NULL;
    struct VS_DiagnosticData *p_curr_data = NULL;
    char buffer[1024];
    stringstream sstream;
    stringstream key_data_sstream;
    string key_data_str;

    for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size();
         ++dd_idx) {
         DiagnosticDataInfo *p_dd = this->diagnostic_data_vec[dd_idx];

         if (p_dd->GetDDType() != dd_type)
            continue;

         if (dd_idx_and_akey_to_dd_page.size() > dd_idx) {
             map_akey_diagnostic_data_t *p_map_akey_dd = this->dd_idx_and_akey_to_dd_page[dd_idx];

             // data not exist
             if (!p_map_akey_dd)
                 continue;

             if (p_dd->DumpDiagnosticDataHeaderStart(csv_out))
                 continue;

             for (map_akey_diagnostic_data_t::iterator iter = p_map_akey_dd->begin();
                  iter != p_map_akey_dd->end(); ++iter) {
                 p_curr_acc_reg_key = iter->first;
                 if (!p_curr_acc_reg_key)
                     continue;
                  p_curr_data = iter->second;
                  if (!p_curr_data)
                      continue;

                  sstream.str("");
                  key_data_sstream.str("");
                  p_curr_acc_reg_key->DumpKeyData(key_data_sstream);
                  key_data_str = key_data_sstream.str();
                  snprintf(buffer, sizeof(buffer), STR_FMT U32D_FMT ",", key_data_str.c_str(),
                           p_curr_data->CurrentRevision);
                  sstream << buffer;

                  p_dd->DumpDiagnosticData(sstream, *p_curr_data);

                  sstream << endl;
                  csv_out.WriteBuf(sstream.str());
             }

             p_dd->DumpDiagnosticDataHeaderEnd(csv_out);
         }
    }

    IBDIAGNET_RETURN_VOID;
}

int PhyDiag::DumpCSVSocketDirect()
{
    IBDIAGNET_ENTER;

    if (this->p_csv_out->DumpStart(SOCKET_DIRECT_TABLE_SECTION_NAME))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->p_csv_out->WriteBuf(SOCKET_DIRECT_TABLE_SECTION_HEADER "\n");

    AccRegHandler *p_mpir_handler = NULL;
    AccRegHandler *p_mpein_handler = NULL;

    for (uint32_t i = 0; i < this->pci_reg_handlers_vec.size(); ++i) {
        AccRegHandler *p_areg_handler = this->pci_reg_handlers_vec[i];

        if (!p_areg_handler || !p_areg_handler->GetPReg())
            continue;

        if (p_areg_handler->GetPReg()->GetName() == ACC_REG_MPIR_NAME)
            p_mpir_handler = p_areg_handler;

        if (p_areg_handler->GetPReg()->GetName() == ACC_REG_MPEIN_NAME)
            p_mpein_handler = p_areg_handler;
    }

    if (!p_mpir_handler || !p_mpein_handler) {
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    map_akey_areg &mpir_map = p_mpir_handler->data_map;
    map_akey_areg &mpein_map = p_mpein_handler->data_map;

    for (map_akey_areg::iterator mpir_it = mpir_map.begin(); mpir_it != mpir_map.end(); mpir_it++) {
        if (mpir_it->second.regs.mpir.sdm) { // Socket Direct node

            stringstream sstr;
            AccRegKeyDPN *cur_key = (AccRegKeyDPN*)mpir_it->first;
            map_akey_areg::iterator mpein_it = mpein_map.find(cur_key);

            // NodeGuid,PCIIndex,Depth,PCINode,device_status
            sstr << PTR(cur_key->node_guid)    << ','
                 << +cur_key->pci_idx          << ','
                 << +cur_key->depth            << ','
                 << +cur_key->pci_node         << ',';

            if (mpein_it == mpein_map.end())
                sstr << "NA";
            else
                sstr << PTR(mpein_it->second.regs.mpein.device_status);

            sstr << endl;

            this->p_csv_out->WriteBuf(sstr.str());
        }
    } // mpir loop

    this->p_csv_out->DumpEnd(SOCKET_DIRECT_TABLE_SECTION_NAME);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PhyDiag::DumpCSV_UPHY_Versions()
{
    IBDIAG_ENTER;

    if (this->p_csv_out->DumpStart(UPHY_VERSION_INFO_SECTION_NAME))
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->p_csv_out->WriteBuf(UPHY_VERSION_INFO_SECTION_HEADER "\n");

    for (map_str_pnode::iterator nodesI = p_discovered_fabric->NodeByName.begin();
            nodesI != p_discovered_fabric->NodeByName.end(); ++nodesI) {

        IBNode *p_node = nodesI->second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in NodeByName " \
                    "map for key = %s", (*nodesI).first.c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        PHYNodeData* phy_data = (PHYNodeData*)p_node->p_phy_data;

        // skip nodes that not support UPHY (peucg)
        if (!phy_data || !phy_data->uphy)
            continue;

        stringstream sstr;

        // NodeGuid, CLN_Version, CLN_JSON, DLN_Version, DLN_JSON
        sstr << PTR(p_node->guid_get())                                           << ", "
             << (phy_data->uphy->cln.version ?
                     phy_data->uphy->cln.version->to_string() : "N/A")            << ", "
             << (phy_data->uphy->cln.dataset ?
                     phy_data->uphy->cln.dataset->version().to_string() : "N/A")  << ", "
             << (phy_data->uphy->dln.version ?
                     phy_data->uphy->dln.version->to_string() : "N/A")            << ", "
             << (phy_data->uphy->dln.dataset ?
                     phy_data->uphy->dln.dataset->version().to_string() : "N/A")  << endl;

        this->p_csv_out->WriteBuf(sstr.str());
    }

    this->p_csv_out->DumpEnd(UPHY_VERSION_INFO_SECTION_NAME);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PhyDiag::DumpCSV_UPHY(string section_name)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    UPHY::DumpEngine dump_engine;
    AccRegHandler *p_areg_peucg = NULL;

    for (size_t i = 0; i < this->reg_handlers_vec.size(); ++i) {
        if (this->reg_handlers_vec[i]->GetPReg()->GetSectionName() == section_name) {
            p_areg_peucg = reg_handlers_vec[i];
            break;
        }
    }

    if (!p_areg_peucg)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    for (map_akey_areg::iterator iter = p_areg_peucg->data_map.begin();
            iter != p_areg_peucg->data_map.end(); iter++) {

        UPHY::DataKey *p_cur_key = (UPHY::DataKey*)iter->first;
        if (!p_cur_key)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

        AccRegKeyPortLane* p_plkey = (AccRegKeyPortLane*)(p_cur_key->base_key());
        if (!p_plkey)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

        IBNode *p_node = p_discovered_fabric->getNodeByGuid(p_plkey->node_guid);
        if (!p_node || !p_node->p_phy_data)
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

        PHYNodeData::UPHYData* p_uphy = ((PHYNodeData*)p_node->p_phy_data)->uphy;
        if (!p_uphy)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

        const UPHY::DataSet* p_dataset = (section_name == ACC_REG_PEUCG_DLN_INTERNAL_SECTION_NAME ?
                p_uphy->dln.dataset : p_uphy->cln.dataset);

        if (!dump_engine.add(p_dataset, p_cur_key, &iter->second.regs.peucg))
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    dump_engine.to_csv(*this->p_csv_out, this->uphy_dumps);

    IBDIAG_RETURN(rc);
}

int PhyDiag::DumpNetDumpExt()
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    u_int32_t dd_idx = 0;
    DiagnosticDataInfo *p_dd;
    ofstream sout;
    char port_line[1024] = {'\0'};

    rc = p_ibdiag->OpenFile("Network dump ext.", string(NET_DUMP_EXT_FILE), sout, false, true);

    if (rc){
        ERR_PRINT("Failed to open Network dump ext. file");
        IBDIAGNET_RETURN(rc);
    }

    IBFabric::GetSwitchLabelPortNumExplanation(sout);

    snprintf(port_line, 1024,
             "%-2s : %-16s : %-3s : %-18s : "
             "%-12s : %-4s : %-7s : %-7s : "
             "%-7s : %-24s : %-19s : %-6s : "
             "%-15s : %-15s : %-15s : %-10s : %-13s : %s",
             "Ty", "#",  "#IB", "GUID",
             "LID", "Sta", "PhysSta", "LWA",
             "LSA", "Conn LID (#)", "FEC mode", "RTR",
             "Raw BER", "Effective BER", "Symbol BER", "Symbol Err", "Effective Err", "Node Desc");
    sout << port_line << endl;

    for (; dd_idx < diagnostic_data_vec.size(); ++dd_idx) {
        p_dd = diagnostic_data_vec[dd_idx];
        if (p_dd->GetPageId() == DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE)
            break;
    }

    if (dd_idx == diagnostic_data_vec.size()) {
        ERR_PRINT("Can't find PHY_LAYER_STATISTICS_PAGE");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    map_str_pnode::iterator nodesI = p_discovered_fabric->NodeByName.begin();
    for ( ; nodesI != p_discovered_fabric->NodeByName.end(); ++nodesI) {

        IBNode *p_curr_node = (*nodesI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName " \
                               "map for key = %s", (*nodesI).first.c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        // go over all ports of this node, start from port 1, no need to dump port 0
        for (u_int32_t port_i = 1; port_i < (unsigned)p_curr_node->numPorts + 1; ++port_i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_i);
            if (!p_curr_port)
                continue;
            DumpNetDumpExtPort(sout, p_curr_port, p_curr_node, dd_idx, false);
        }
    }

    this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(rc);
}

int PhyDiag::DumpNetDumpExtAgg()
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    u_int32_t dd_idx = 0;
    DiagnosticDataInfo *p_dd;
    ofstream sout;
    char port_line[1024] = {'\0'};
    p_discovered_fabric->unvisitAllAPorts();

    rc = p_ibdiag->OpenFile("Aggregated network dump ext.", string(NET_DUMP_EXT_AGG_FILE), sout, false, true);

    if (rc){
        ERR_PRINT("Failed to open Network dump ext. aggregated file");
        IBDIAGNET_RETURN(rc);
    }

    IBFabric::GetSwitchLabelPortNumExplanation(sout);

    snprintf(port_line, 1024,
             "%-20s : %-2s : %-4s : %-3s : %-18s : "
             "%-12s : %-4s : %-7s : %-7s : "
             "%-7s : %-24s : %-19s : %-6s : "
             "%-15s : %-15s : %-15s : %-10s : %-13s : %s",
             "SystemGUID", "Ty", "#",  "#IB", "GUID",
             "LID", "Sta", "PhysSta", "LWA",
             "LSA", "Conn LID (#)", "FEC mode", "RTR",
             "Raw BER", "Effective BER", "Symbol BER", "Symbol Err", "Effective Err", "Node Desc");
    sout << port_line << endl;

    for (; dd_idx < diagnostic_data_vec.size(); ++dd_idx) {
        p_dd = diagnostic_data_vec[dd_idx];
        if (p_dd->GetPageId() == DIAGNOSTIC_DATA_PHY_LAYER_STATISTICS_PAGE)
            break;
    }

    if (dd_idx == diagnostic_data_vec.size()) {
        ERR_PRINT("Can't find PHY_LAYER_STATISTICS_PAGE");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    for (auto& name_system_pair : p_discovered_fabric->SystemByName) {
        IBSystem *p_curr_system = name_system_pair.second;
        if (!p_curr_system)
            continue;

        for (auto& name_node_pair : p_curr_system->NodeByName) {
            IBNode *p_curr_node = name_node_pair.second;
            if (!p_curr_node) {
                this->SetLastError("DB error - found null node in NodeByName " \
                                "map for key = %s", name_node_pair.first.c_str());
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            if (!p_curr_node->getInSubFabric())
                continue;

            // go over all ports of this node, start from port 1, no need to dump port 0
            for (u_int32_t port_i = 1; port_i < (unsigned)p_curr_node->numPorts + 1; ++port_i) {
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_i);
                if (!p_curr_port)
                    continue;
                APort *p_curr_aport = p_curr_port->p_aport;
                if (p_curr_aport) {
                    if (p_curr_aport->visited)
                        continue;
                    // visit all planes of this APort and mark it as visited
                    for (size_t plane = 1; plane < p_curr_aport->ports.size(); ++plane) {
                        IBPort *p_curr_plane = p_curr_aport->ports[plane];
                        if (!p_curr_plane || !p_curr_plane->p_node)
                            continue;
                        DumpNetDumpExtPort(sout, p_curr_plane, p_curr_plane->p_node, dd_idx, true);
                    }
                    p_curr_aport->visited = true;
                } else {
                    DumpNetDumpExtPort(sout, p_curr_port, p_curr_node, dd_idx, true);
                }
            }

        }
    }
    map_str_pnode::iterator nodesI = p_discovered_fabric->NodeByName.begin();
    for ( ; nodesI != p_discovered_fabric->NodeByName.end(); ++nodesI) {
    }

    this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(rc);
}
int PhyDiag::DumpNetDumpExtPort(ostream &sout, IBPort *p_curr_port, IBNode *p_curr_node,
                u_int32_t dd_idx, bool is_agg)
{
    IBDIAG_ENTER;
    VS_DiagnosticData *p_curr_data = NULL;
    DD_PhyStatisticalCounters phy_statistics;
    char port_line[1024] = {'\0'};
    int rc = IBDIAG_SUCCESS_CODE;

    if (!p_curr_port || !p_curr_port->getInSubFabric())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    if (!p_curr_port->isValid())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    SMP_PortInfo * p_curr_port_info = p_ibdiag->GetIBDMExtendedInfoPtr()->
            getSMPPortInfo(p_curr_port->createIndex);
    if (!p_curr_port_info)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

    SMP_MlnxExtPortInfo *p_mepi = p_ibdiag->GetIBDMExtendedInfoPtr()->
            getSMPMlnxExtPortInfo(p_curr_port->createIndex);

    char lid_port[128] = {'\0'};
    if (p_curr_port->p_remotePort) {
        snprintf(lid_port, 128,
                    "%-5d (%s)",
                    p_curr_port->p_remotePort->base_lid,
                    p_curr_port->p_remotePort->numAsString().c_str());
    }

    enum IBBERType ber_entry;
    double ber_report;
    double symbol_ber;
    string ber_description;
    list_p_fabric_general_err *phy_ber_errors;

    stringstream raw_ber_str;
    stringstream eff_ber_str;
    stringstream symbol_ber_str;
    stringstream effective_errors;
    stringstream symbol_errors;

    CLEAR_STRUCT(phy_statistics);
    p_curr_data = getPhysLayerPortCounters(p_curr_port->createIndex, dd_idx);
    if (p_curr_data) {
        DD_PhyStatisticalCounters_unpack(&phy_statistics,
                                            (uint8_t *)&(p_curr_data->data_set));

        double raw_ber = phy_statistics.raw_ber_coef *
                (1 / pow(10, phy_statistics.raw_ber_magnitude));
        double eff_ber = phy_statistics.effective_ber_coef *
                (1 / pow(10, phy_statistics.effective_ber_magnitude));

        raw_ber_str << scientific << raw_ber;
        eff_ber_str << scientific << eff_ber;
        symbol_errors << phy_statistics.phy_symbol_errors;

        // check if new counters in PPCNT phy_statistics are supported.
        if (this->p_capability_module->
                IsSupportedGMPCapability(p_curr_node, EnGMPCAPIsEffectiveCounterSupported)) {
            symbol_ber = phy_statistics.symbol_ber_coef *
                    (1 / pow(10, phy_statistics.symbol_ber_magnitude));
            symbol_ber_str << scientific << symbol_ber;
            ber_report = symbol_ber;
            ber_entry = IB_SYMBOL_BER;
            phy_ber_errors = &this->phy_symbol_ber_errors;
            effective_errors << phy_statistics.phy_effective_errors;
        } else {
            symbol_ber_str << "N/A";
            ber_report = eff_ber;
            ber_entry = IB_EFFECTIVE_BER;
            phy_ber_errors = &this->phy_effective_ber_errors;
            ber_description = "(Please upgrade FW to support Symbol BER)";
            effective_errors << "N/A";
        }

        if (p_curr_port->is_data_worthy() &&
                !p_curr_port->isSpecialPort() &&
                !p_curr_node->isSpecialNode() &&
                !is_agg) {
            ReportPortBERErrors(p_curr_port, ber_entry, ber_report, ber_description,
                            phy_ber_errors, p_curr_data);
        }
    }

    CLEAR_STRUCT(port_line);
    if (is_agg) {
        int plane = -1;
        if (p_curr_port->p_port_hierarchy_info)
            plane = p_curr_port->p_port_hierarchy_info->m_plane;

        snprintf(port_line, 1024,
                    "%-20p : %-2s : %-4d : %-3u : " U64H_FMT " : "
                    "%-4u (0x%-3x) : %-4s : %-7s : %-7s : "
                    "%-7s : %-24s : %-19s : %-6s : "
                    "%-15s : %-15s : %-15s : %-10s : %-13s : \"%s\"\n",

                    (void *)p_curr_node->system_guid_get(),
                    nodetype2char(p_curr_node->type),
                    plane,
                    p_curr_port->num,
                    p_curr_port->guid_get(),
                    p_curr_port->base_lid,
                    p_curr_port->base_lid,
                    portstate2char(p_curr_port->get_internal_state()),
                    portphysstate2char((IBPortPhysState)p_curr_port_info->PortPhyState),
                    p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                            NOT_AVAILABLE : width2char(p_curr_port->get_internal_width()),
                    p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                            NOT_AVAILABLE : speed2char(p_curr_port->get_internal_speed()),
                    lid_port,
                    fec2char(p_curr_port->get_fec_mode()),
                    !p_mepi ? RETRANS_MODE_NA :
                            retransmission2char((EnRetransmissionMode)p_mepi->RetransMode),
                    p_curr_data ? raw_ber_str.str().c_str() : "",
                    p_curr_data ? eff_ber_str.str().c_str() : "",
                    p_curr_data ? symbol_ber_str.str().c_str() : "",
                    p_curr_data ? symbol_errors.str().c_str() : "",
                    p_curr_data ? effective_errors.str().c_str() : "",
                    p_curr_node->description.c_str());
    } else {
        snprintf(port_line, 1024,
                    "%-2s : %-16s : %-3u : " U64H_FMT " : "
                    "%-4u (0x%-3x) : %-4s : %-7s : %-7s : "
                    "%-7s : %-24s : %-19s : %-6s : "
                    "%-15s : %-15s : %-15s : %-10s : %-13s : \"%s\"\n",

                    nodetype2char(p_curr_node->type),
                    p_curr_port->numAsString().c_str(),
                    p_curr_port->num,
                    p_curr_port->guid_get(),
                    p_curr_port->base_lid,
                    p_curr_port->base_lid,
                    portstate2char(p_curr_port->get_internal_state()),
                    portphysstate2char((IBPortPhysState)p_curr_port_info->PortPhyState),
                    p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                            NOT_AVAILABLE : width2char(p_curr_port->get_internal_width()),
                    p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN ?
                            NOT_AVAILABLE : speed2char(p_curr_port->get_internal_speed()),
                    lid_port,
                    fec2char(p_curr_port->get_fec_mode()),
                    !p_mepi ? RETRANS_MODE_NA :
                            retransmission2char((EnRetransmissionMode)p_mepi->RetransMode),
                    p_curr_data ? raw_ber_str.str().c_str() : "",
                    p_curr_data ? eff_ber_str.str().c_str() : "",
                    p_curr_data ? symbol_ber_str.str().c_str() : "",
                    p_curr_data ? symbol_errors.str().c_str() : "",
                    p_curr_data ? effective_errors.str().c_str() : "",
                    p_curr_node->description.c_str());

    }
    sout << port_line;

    IBDIAGNET_RETURN(rc);
}

void PhyDiag::ReportPortBERErrors(IBPort *p_curr_port, enum IBBERType ber_entry, double ber_report,
                            string ber_description,
                            list_p_fabric_general_err *phy_ber_errors,
                            VS_DiagnosticData *p_curr_data)
{
    const vec_ber_thresholds_val_t *ber_thresholds =
                                this->GetBerThresholdEntry(*p_curr_port);
    if (!ber_thresholds)
        phy_ber_errors->push_back(
                    new FabricErrBERThresholdNotFound(p_curr_port,
                                media_type2char(this->GetMediaType(*p_curr_port))));
    else if ((*ber_thresholds)[ber_entry].error < (*ber_thresholds)[ber_entry].warning)
            phy_ber_errors->push_back(new FabricErrBERThresholdValue(p_curr_port,
                                                media_type2char(
                                                    this->GetMediaType(*p_curr_port)),
                                                (*ber_thresholds)[ber_entry].error,
                                                (*ber_thresholds)[ber_entry].warning));

    // set Errors / Warnings for bad links
    if (p_curr_data && ber_thresholds) {
        if (ber_report > (*ber_thresholds)[ber_entry].error) {

            this->total_ber_errors++;

            phy_ber_errors->push_back(
                new FabricErrFwBERExceedThreshold(
                    p_curr_port,
                    (*ber_thresholds)[ber_entry].error,
                    ber_report,
                    ber_entry,
                    ber_description));

        } else if ((to_show_ber_warnings) &&
                (ber_report > (*ber_thresholds)[ber_entry].warning)) {

            FabricErrFwBERExceedThreshold *p_curr_fabric_ber_err =
                    new FabricErrFwBERExceedThreshold(p_curr_port,
                                                    (*ber_thresholds)[ber_entry].warning,
                                                    ber_report,
                                                    ber_entry,
                                                    ber_description);

            p_curr_fabric_ber_err->SetLevel(EN_FABRIC_ERR_WARNING);
            phy_ber_errors->push_back(p_curr_fabric_ber_err);
        }
    }

}
void PhyDiag::DumpCSVRawBER(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;

    char buffer[1024];
    stringstream sstream;

    if (csv_out.DumpStart(SECTION_RAW_BER))
        IBDIAG_RETURN_VOID;

    sstream << "NodeGuid,PortGuid,PortNum,field0" << endl;

#ifdef SHOW_PHY_HEADERS
    sstream  << "#NodeGuid,PortGuid,PortNum,BER"   << endl;
#endif

    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        if (isSupportFwBER(p_curr_port))
            continue;

        long double *p_ber = getBER(p_curr_port->createIndex);

        //info doesn't exist
        if (!p_ber)
            continue;

        //show exponent of the raw ber
        //in case of zero ber, the exponent is infinite, show 255 instead
        double ber_exponent = (*p_ber == 0) ?
                              ZERO_RAW_BER_EXPONENT :
                              -1.0 * log10(*p_ber);
        sstream.str("");
        //NodeGuid,PortGuid,PortNum,raw ber exponent
        snprintf(buffer, sizeof(buffer),
                 U64H_FMT "," U64H_FMT "," U32D_FMT ",%f",
                 p_curr_port->p_node->guid_get(), p_curr_port->guid_get(),
                 p_curr_port->num, ber_exponent);
        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_RAW_BER);
    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::DumpCSVEffectiveBER(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;
    SMP_MlnxExtPortInfo* p_ext_port_info = NULL;
    char buffer[1024];
    stringstream sstream;

    if (csv_out.DumpStart(SECTION_EFF_BER))
        IBDIAG_RETURN_VOID;

    sstream << "NodeGuid,PortGuid,PortNum,BER,FECMode,Retransmission" << endl;
    csv_out.WriteBuf(sstream.str());

    for (u_int32_t i = 0; i < this->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        if (isSupportFwBER(p_curr_port))
            continue;

        long double *p_reciprocal_ber = getEffBER(p_curr_port->createIndex);

        //info doesn't exist
        if (!p_reciprocal_ber)
            continue;
        p_ext_port_info =
            this->p_ibdm_extended_info->getSMPMlnxExtPortInfo(
                    p_curr_port->createIndex);
        if (!p_ext_port_info)
            continue;

        sstream.str("");
        //NodeGuid,PortGuid,PortNum, effective reciprocal_ber
        snprintf(buffer, sizeof(buffer),
                 U64H_FMT "," U64H_FMT "," U32D_FMT ",%Le," STR_FMT "," STR_FMT,
                 p_curr_port->p_node->guid_get(), p_curr_port->guid_get(),
                 p_curr_port->num,
                 (*p_reciprocal_ber) == 0 ? 0.0 : 1.0/(*p_reciprocal_ber),
                 fec2char(p_curr_port->get_fec_mode()),
                 retransmission2char((EnRetransmissionMode)p_ext_port_info->RetransMode));

        sstream << buffer << endl;
        csv_out.WriteBuf(sstream.str());
    }

    csv_out.DumpEnd(SECTION_EFF_BER);
    IBDIAGNET_RETURN_VOID;
}

void PhyDiag::DumpCSVSLRGExternalInfo(CSVOut &csv_out, AccRegHandler *p_areg_handler)
{
    IBDIAGNET_ENTER;
    bool any_data_per_port = false;
    char buff1[512] = {0}, buff2[512] = {0};
    char *p_in_line = NULL;
    u_int8_t gradeID = 0;
    u_int32_t grade = 0;
    int print_res = 0;
    stringstream sstream;

    if (csv_out.DumpStart(ACC_REG_SLRG_EXTERNAL_SECTION_NAME))
        IBDIAG_RETURN_VOID;

    //NodeGuid,PortGuid,PortNum,GradeID,Lane0Grade,Lane1Grade,Lane2Grade,Lane3Grade
    sstream << "NodeGuid,PortGuid,PortNum,GradeID";
    for (u_int8_t cnt = 0; cnt < LANE_NUM; cnt++)
        sstream << ",Lane" << (unsigned int)cnt <<  "Grade";
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

     for (map_akey_areg::iterator nI = p_areg_handler->data_map.begin();
         nI != p_areg_handler->data_map.end();) {

        any_data_per_port = false;
        gradeID = 0;
        p_in_line = buff2;
        print_res = 0;

        AccRegKey * p_aregkey = (*nI).first;
        if (!p_aregkey) {
            this->SetLastError("DB error - found null key in data_map");
            ++nI;
            continue;
        }

        if (((AccRegKeyPortLane *)p_aregkey)->lane != 0) {
            ++nI;
            continue;
        }

        for (u_int8_t ln = 0; ln < LANE_NUM; ln++) {
            p_aregkey = (*nI).first;
            if (nI == p_areg_handler->data_map.end() ||
                    ((AccRegKeyPortLane *)p_aregkey)->lane != ln) {
                any_data_per_port = false;
                break;
            }

            struct slrg_reg slrg = (slrg_reg&)(*nI).second;

            if (slrg.version == 0 || slrg.version == 1) { // 40nm, 28nm
                slrg_40nm_28nm slrg_28;
                slrg_40nm_28nm_unpack(&slrg_28, slrg.page_data.slrg_data_set.data);
                //GradeID=grade_version is the same for all lanes so take any available
                gradeID = slrg_28.grade_version;
                grade = slrg_28.grade;
            }

            if (slrg.version == 3) { // 16nm
                slrg_16nm slrg_16;
                slrg_16nm_unpack(&slrg_16, slrg.page_data.slrg_data_set.data);
                gradeID = slrg_16.grade_version;
                grade = slrg_16.grade;
            }

            any_data_per_port = true;

            //Lane0Grade,Lane1Grade,Lane2Grade,Lane3Grade
            print_res = sprintf(p_in_line, "," U32D_FMT, grade);
            if (print_res > 0)
                p_in_line += print_res;

            //increasing the map iterator
            ++nI;
        }
        if (any_data_per_port) {
            //NodeGuid,PortGuid,PortNum,GradeID,Lane0Grade,Lane1Grade,Lane2Grade,Lane3Grade
            snprintf(buff1, sizeof(buff1),
                    U64H_FMT "," U64H_FMT "," U32D_FMT "," UH_FMT,
                    ((AccRegKeyPortLane *)p_aregkey)->node_guid,
                    ((AccRegKeyPortLane *)p_aregkey)->port_guid,
                    ((AccRegKeyPortLane *)p_aregkey)->port_num,
                    gradeID);
            sstream.str("");
            sstream << buff1 << buff2 << endl;
            csv_out.WriteBuf(sstream.str());
        }
    }

    csv_out.DumpEnd(ACC_REG_SLRG_EXTERNAL_SECTION_NAME);
    IBDIAGNET_RETURN_VOID;
}


/*****************************************************************************/
static long double CalcBERFireCodeFEC(u_int64_t *p_counters)
{
    IBDIAGNET_ENTER;

    long double ber = 0.0;
    //(sum of FECCorrectableBlocks +
    // Kfcode*FECUncorrectableBlocks over all lanes)/time
    //FECCorrectableBlocks lane0 = fc_fec_corrected_blocks_lane0
    //FECUncorrectableBlocks lane0 = fc_fec_uncorrectable_blocks_lane0
    ber = p_counters[EN_FC_FEC_CORRECTED_BLOCKS_LANE0] +
          p_counters[EN_FC_FEC_CORRECTED_BLOCKS_LANE1] +
          p_counters[EN_FC_FEC_CORRECTED_BLOCKS_LANE2] +
          p_counters[EN_FC_FEC_CORRECTED_BLOCKS_LANE3] +
          K_FCODE *
          (p_counters[EN_FC_FEC_UNCORRECTABLE_BLOCKS_LANE0] +
           p_counters[EN_FC_FEC_UNCORRECTABLE_BLOCKS_LANE1] +
           p_counters[EN_FC_FEC_UNCORRECTABLE_BLOCKS_LANE2] +
           p_counters[EN_FC_FEC_UNCORRECTABLE_BLOCKS_LANE3]);
    IBDIAGNET_RETURN(ber);
}

static long double EDPLSum(u_int64_t *p_counters)
{
    IBDIAGNET_ENTER;
    //sum of edpls over lanes/time
    IBDIAGNET_RETURN(p_counters[EN_EDPL_LANE0] +
                     p_counters[EN_EDPL_LANE1] +
                     p_counters[EN_EDPL_LANE2] +
                     p_counters[EN_EDPL_LANE3]);
}

void PhyDiag::CalcRawBER()
{
    IBDIAGNET_ENTER;
    //calculate RAW BER
    long double ber = 0.0;
    struct VS_DiagnosticData *p_curr_data = NULL;
    u_int64_t                *p_curr_counters = NULL;
    long double k_rsfec = 0.0;

    for (u_int32_t i = 0; i < this->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        if (isSupportFwBER(p_curr_port))
            continue;

        p_curr_data = this->getPhysLayerPortCounters(p_curr_port->createIndex, 0);
        if (!p_curr_data)       //info not exist
            continue;
        p_curr_counters =
            ((struct DDPhysCounters*)&(p_curr_data->data_set.gen_set))->counter;

        IBFECMode port_fec = p_curr_port->get_fec_mode();
        ber = 0.0;

        //undefined fec port
        if (port_fec == IB_FEC_NA)
            ber = max(CalcBERFireCodeFEC(p_curr_counters),
                      EDPLSum(p_curr_counters));
        else if (port_fec == IB_FEC_NO_FEC) {
            //up to QDR (8/10)
            if (p_curr_port->get_common_speed() < IB_LINK_SPEED_14)
                //symbol error/time
                ber = p_curr_counters[EN_SYMBOL_ERRORS];
            else //64/66
                //sum of edpls over lanes/time
                ber = EDPLSum(p_curr_counters);
        } else if (port_fec == IB_FEC_FIRECODE_FEC)
            //(sum of FECCorrectableBlocks +
            // Kfcode*FECUncorrectableBlocks over all lanes)/time
            //FECCorrectableBlocks lane0 = fc_fec_corrected_blocks_lane0
            //FECUncorrectableBlocks lane0 = fc_fec_uncorrectable_blocks_lane0
            ber = CalcBERFireCodeFEC(p_curr_counters);
        else if (isRSFEC(port_fec)) {
            //PortFECCorrectedSymbolsCounter + Krsfec*PortFECUncorrectableBlocks
            //PortFECCorrectedSymbolsCounter = rs_fec_corrected_symbols_total
            //PortFECUncorrectableBlocks = rs_fec_uncorrectable_blocks
            SMP_MlnxExtPortInfo* p_ext_port_info =
                this->p_ibdm_extended_info->getSMPMlnxExtPortInfo(
                    p_curr_port->createIndex);
            if (!p_ext_port_info)
                continue;
            k_rsfec = p_ext_port_info->ActiveRSFECParity / 2 + 1;
            //calculate k_rsfec
            ber = p_curr_counters[EN_RS_FEC_CORRECTED_SYMBOLS_TOTAL] +
                  k_rsfec * p_curr_counters[EN_RS_FEC_UNCORRECTABLE_BLOCKS];
        }

        long double link_rate =
            (long double)(CalcLinkRate(p_curr_port->get_common_width(),
                                       p_curr_port->get_common_speed()));
        long double data_rate =
            //EN_TIME_SINCE_LAST_CLEAR is in msec
            (p_curr_counters[EN_TIME_SINCE_LAST_CLEAR] / 1000) * link_rate;

        if (data_rate) {
            ber /= data_rate;
            this->addBER(p_curr_port, ber);
        }
    }
    IBDIAGNET_RETURN_VOID;
}

int PhyDiag::CalcEffBER(u_int64_t ber_threshold_reciprocal_val,
                        list_p_fabric_general_err& ber_errors)
{
    IBDIAGNET_ENTER;
    int rc = 0;
    long double reciprocal_ber = 0.0;
    struct VS_DiagnosticData *p_curr_data = NULL;
    u_int64_t                *p_curr_counters = NULL;

    for (u_int32_t i = 0; i < this->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        if (isSupportFwBER(p_curr_port))
            continue;

        p_curr_data = this->getPhysLayerPortCounters(p_curr_port->createIndex, 0);
        if (!p_curr_data)       //info not exist
            continue;
        p_curr_counters =
            ((struct DDPhysCounters*)&(p_curr_data->data_set.gen_set))->counter;

        reciprocal_ber = 0.0;
        rc = this->p_ibdiag->CalcBER(p_curr_port,
                                     p_curr_counters[EN_TIME_SINCE_LAST_CLEAR]/1000,
                                     p_curr_counters[EN_SYMBOL_ERRORS],
                                     reciprocal_ber);

        if (rc == IBDIAG_ERR_CODE_DB_ERR)
            continue;

        if (reciprocal_ber == 0) {
            if (ber_threshold_reciprocal_val == OVERFLOW_VAL_64_BIT) {    //need to report anyway
                ber_errors.push_back(new FabricErrEffBERIsZero(p_curr_port));
            }
        } else if ((reciprocal_ber < (ber_threshold_reciprocal_val)) ||
                (ber_threshold_reciprocal_val == OVERFLOW_VAL_64_BIT) ) {
            ber_errors.push_back(
                    new FabricErrEffBERExceedThreshold(
                            p_curr_port, ber_threshold_reciprocal_val, reciprocal_ber));
        }

        this->addEffBER(p_curr_port, reciprocal_ber);
    }
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PhyDiag::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    int rc2 = 0;

    //Dump first external data
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_PHY_CNTRS_CHECK);
        PRINT("\n");
    } else if (this->to_get_phy_info) {
        rc2 = CalcEffBER(this->p_ibdiag->GetBERThreshold(), this->phy_errors);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(this->phy_errors,
                                       CHECK_NAME_EFF_BER_CALC,
                                       rc2,
                                       IBDIAG_ERR_CODE_CHECK_FAILED,
                                       this->num_errors,
                                       this->num_warnings);
        if (rc)
            IBDIAGNET_RETURN(rc);

        CalcRawBER();
        //dump phy to db file
        this->DumpCSVPhyCounters(*this->p_csv_out, DD_PHY_TYPE);
        this->DumpCSV_DDCableInfo(*this->p_csv_out);
        this->DumpCSV_AccRegCableInfo(*this->p_csv_out);

        if (this->DumpFile_DDCableInfo(PHY_DD_CABLE_FILE)) {
            ERR_PRINT("Writing cables info file failed\n");
            ++this->num_errors;
        }

        if (this->DumpFile_AccRegCableInfo(PHY_ACC_REG_CABLE_FILE)) {
            ERR_PRINT("Writing cables info file failed\n");
            ++this->num_errors;
        }

        this->DumpCSVRawBER(*this->p_csv_out);
        this->DumpCSVEffectiveBER(*this->p_csv_out);

        // Add CSV tables for UPHY
        if (!UPHY::DB::instance().empty()) {
            this->DumpCSV_UPHY_Versions();
            this->DumpCSV_UPHY(ACC_REG_PEUCG_DLN_INTERNAL_SECTION_NAME);
            this->DumpCSV_UPHY(ACC_REG_PEUCG_CLN_INTERNAL_SECTION_NAME);
        }

        rc2 = DumpNetDumpExt();
        DumpNetDumpExtAgg();
        SCREEN_PRINT("\n");
        int rc_effective_ber = this->AnalyzeCheckResults(this->phy_effective_ber_errors,
                                                         CHECK_NAME_EFF_BER,
                                                         rc2,
                                                         IBDIAG_ERR_CODE_CHECK_FAILED,
                                                         this->num_errors,
                                                         this->num_warnings);
        int rc_symbol_ber = this->AnalyzeCheckResults(this->phy_symbol_ber_errors,
                                                      CHECK_NAME_SYM_BER,
                                                      rc2,
                                                      IBDIAG_ERR_CODE_CHECK_FAILED,
                                                      this->num_errors,
                                                      this->num_warnings);
        if (rc_symbol_ber)
            IBDIAGNET_RETURN(rc_symbol_ber);
        if (rc_effective_ber)
            IBDIAGNET_RETURN(rc_effective_ber);
    }

    //first, dump the internal registers
    for (uint32_t i = 0; i < this->reg_handlers_vec.size(); ++i) {
        AccRegHandler *p_areg_handler = this->reg_handlers_vec[i];
        if (!p_areg_handler->GetDumpEnabled())
            if (!to_dump_cap_reg)
                continue;
        p_areg_handler->DumpCSV(*this->p_csv_out);
    }

    // second, dump the external registers
    // TODO: change to virtual functions empty DumpCSVExternalInfo in the Register Class.
    // SLRG will overrun this function
    for (uint32_t i = 0; i < this->reg_handlers_vec.size(); ++i) {
        AccRegHandler *p_areg_handler = this->reg_handlers_vec[i];
        if (p_areg_handler->GetPReg()->GetRegisterID() == ACCESS_REGISTER_ID_SLRG)
            this->DumpCSVSLRGExternalInfo(*this->p_csv_out, p_areg_handler);
    }

    if (this->to_get_pci_info) {

        if (!this->can_send_mads_by_lid) {
            INFO_PRINT("%s skipped\n", CHECK_NAME_PHY_CNTRS_CHECK);
            PRINT("\n");
        } else {
            this->DumpCSVPCICounters(*this->p_csv_out, DD_PCI_TYPE);
        }
    }

    if (this->to_get_pci_info || this->p_ibdiag->IsRailValidation()) {
        for (uint32_t i = 0; i < this->pci_reg_handlers_vec.size(); ++i) {
            AccRegHandler *p_areg_handler = this->pci_reg_handlers_vec[i];
            p_areg_handler->DumpCSV(*this->p_csv_out);
        }

        // Add CSV table for Socket-Direct nodes
        this->DumpCSVSocketDirect();
    }

    // Check PCI degradation
    // buid ignore list of management nodes
    list_p_sm_info_obj &sm_info_list =
            this->p_ibdiag->GetIBDMExtendedInfoPtr()->getSMPSMInfoListRef();

    set<const IBPort*> skipped;
    for (const auto& x : sm_info_list)
        skipped.insert(x->p_port);

    // Main loop
    for (const auto& reg : this->pci_reg_handlers_vec)
        if (reg->GetPReg() && reg->GetPReg()->GetRegisterID() == ACCESS_REGISTER_ID_MPEIN)
            for (const auto& data: reg->data_map) {
                const struct mpein_reg & mpein = data.second.regs.mpein;
                const AccRegKeyDPN* key = static_cast<const AccRegKeyDPN*>(data.first);

                // check skipped list
                if (skipped.count(key->p_port))
                    continue;

                //ignore if port state is 'Force Disable'
                if (mpein.port_state == 0xF0)
                    continue;
                    
                if (mpein.port_type == 0) {
                    if (mpein.link_width_enabled > mpein.link_width_active)
                        this->pci_errors.push_back(
                                new FabricPCIWidthDegradation(
                                        key->p_port,
                                        key->depth, key->pci_idx, key->pci_node,
                                        mpein.link_width_enabled, mpein.link_width_active));

                    #define PCI_MSB(v, x) do { x = 0; uint32_t z = v;  while (z) { x++; z >>= 1; } } while(0)

                    unsigned int enabled = 0, active = 0;
                    PCI_MSB(mpein.link_speed_enabled, enabled);
                    PCI_MSB(mpein.link_speed_active, active);

                    #undef PCI_MSB

                    if (enabled > active)
                        this->pci_errors.push_back(
                                new FabricPCISpeedDegradation(
                                        key->p_port,
                                        key->depth, key->pci_idx, key->pci_node,
                                        enabled, active));
                }
            }

    rc = this->AnalyzeCheckResults(this->pci_errors,
                                   CHECK_NAME_PCI_DEGRADATION,
                                   IBDIAG_SUCCESS_CODE,
                                   IBDIAG_ERR_CODE_CHECK_FAILED,
                                   this->num_errors,
                                   this->num_warnings);

    IBDIAGNET_RETURN(rc);
}

int PhyDiag::InitPhyDataOnNodes()
{
    IBDIAGNET_ENTER;

    for (map_str_pnode::iterator nodesI = p_discovered_fabric->NodeByName.begin();
            nodesI != p_discovered_fabric->NodeByName.end(); ++nodesI) {

        IBNode *p_node = nodesI->second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in NodeByName " \
                    "map for key = %s", (*nodesI).first.c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        p_node->p_phy_data = new PHYNodeData();

        for (phys_port_t i = 0; i <= p_node->numPorts; ++i) {
            IBPort *p_port = p_node->getPort(i);

            if (p_port)
                p_port->p_phy_data = new PHYPortData();
        }

    }

    IBDIAGNET_RETURN(0);
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 * IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int PhyDiag::SMPAccRegGetByDirect(IN direct_route_t *p_direct_route,
                               IN u_int8_t port_num,
                               OUT struct SMP_AccessRegister *p_smp_acc_reg,
                               const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    p_smp_acc_reg->type_op = EMAD_TLV_TYPE_OPERATION;
    p_smp_acc_reg->len_op = EMAD_TLV_TYPE_OPERATION_LEN;
    p_smp_acc_reg->class_op = EMAD_CLASS_ACC_REG;
    p_smp_acc_reg->type_reg = EMAD_TLV_TYPE_REG;
    p_smp_acc_reg->method = EMAD_OP_TLV_METHOD_QUERY;
    p_smp_acc_reg->dr = 0x1;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending ACC_REG MAD by direct = %s port = %u\n",
    this->p_ibis_obj->ConvertDirPathToStr(p_direct_route).c_str(),
    (int)port_num);
    /*
     * send the mad by direct route
     */
    ((ProgressBar *)p_clbck_data->m_p_progress_bar)->push(((IBNode *)p_clbck_data->m_data1));
    int rc = this->p_ibis_obj->SMPMadGetSetByDirect(
        p_direct_route,
        IBIS_IB_MAD_METHOD_SET, //access register has to be set
        IB_ATTR_SMP_ACCESS_REG,
        port_num,
        data_func_set_t(p_smp_acc_reg, IBIS_FUNC_LST(SMP_AccessRegister)),
        p_clbck_data);

    IBIS_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

bool PhyDiag::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    if (!this->to_get_phy_info || this->total_ber_errors == 0)
        IBDIAGNET_RETURN(false);

    summary_str << "High BER reported by "
                << this->total_ber_errors
                << " ports"
                << endl;

    IBDIAGNET_RETURN(true);
}

int PhyDiag::GMPAccRegGet(u_int16_t lid,
                          u_int32_t register_id,
                          OUT struct GMP_AccessRegister *p_gmp_acc_reg,
                          const clbck_data_t *p_clbck_data)
{
    IBDIAGNET_ENTER;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending ACC_REG MAD by lid = " U16H_FMT "\n", lid);

    ((ProgressBar *)p_clbck_data->m_p_progress_bar)->push(((IBNode *)p_clbck_data->m_data1));
    int rc = this->p_ibis_obj->VSMadGetSet(lid,
        IBIS_IB_MAD_METHOD_GET,
        IB_ATTR_GMP_ACCESS_REG,
        register_id,
        data_func_set_t(p_gmp_acc_reg, IBIS_FUNC_LST(GMP_AccessRegister)),
        p_clbck_data);

    IBDIAGNET_RETURN(rc & 0x00ff); //bits [8-15] are class specific - don't need them here
}

IBNode * PhyDiag::GetNode(uint64_t node_guid)
{
    IBDIAGNET_ENTER;

    map_guid_pnode::const_iterator it =
            this->GetFabric()->NodeByGuid.find(node_guid);

    IBDIAGNET_RETURN(it == this->GetFabric()->NodeByGuid.end() ? NULL : it->second);
}

PHYNodeData * PhyDiag::GetPHYNodeData(uint64_t node_guid)
{
    IBDIAGNET_ENTER;

    IBNode * node = GetNode(node_guid);
    IBDIAGNET_RETURN(node ? dynamic_cast<PHYNodeData*>(node->p_phy_data) : NULL )
}

IBPort * PhyDiag::GetPort(uint64_t node_guid, phys_port_t port_num)
{
    IBDIAGNET_ENTER;

    IBNode * node = GetNode(node_guid);

    IBDIAGNET_RETURN(node ? node->getPort(port_num) : NULL);
}

PHYPortData * PhyDiag::GetPHYPortData(uint64_t node_guid, phys_port_t port_num)
{
    IBDIAGNET_ENTER;

    IBPort * port = GetPort(node_guid, port_num);
    IBDIAGNET_RETURN(port ? dynamic_cast<PHYPortData*>(port->p_phy_data) : NULL )
}

int PhyDiag::InitExportAPI(list_p_fabric_general_err &export_data_errors)
{
    IBDIAGNET_ENTER;

    int rc;

    if (this->pf_export_data_phy_node && this->pf_export_data_phy_port)
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    if ((rc = this->p_ibdiag->LoadSymbol(
            this->p_ibdiag->GetExportLibHandle(), "export_data_phy_node",
            (void **)&this->pf_export_data_phy_node, export_data_errors)))
        goto exit;

    if ((rc = this->p_ibdiag->LoadSymbol(
            this->p_ibdiag->GetExportLibHandle(), "export_data_phy_port",
            (void **)&this->pf_export_data_phy_port, export_data_errors)))
        goto exit;

exit:
    if (rc) {
        this->pf_export_data_phy_node = NULL;
        this->pf_export_data_phy_port = NULL;
    }

    IBDIAGNET_RETURN(rc);
}

void ClearExportDataNode(export_data_phy_node_t *p_export_data_phy_node)
{
    delete p_export_data_phy_node->p_dd_page_identification;
    delete p_export_data_phy_node->p_dd_pci_perf_counters;
    delete p_export_data_phy_node->p_dd_pcie_timers_states;
    delete p_export_data_phy_node->p_dd_pcie_lanes_counters;

    for (size_t i = 0; i < 256; ++i)
        delete p_export_data_phy_node->p_ppll_reg[i];

    for (size_t i = 0; i < 256; ++i) {
        delete p_export_data_phy_node->p_slrg_reg[i];
        delete p_export_data_phy_node->p_slrp_reg[i];
        delete p_export_data_phy_node->p_sltp_reg[i];
    }

    CLEAR_STRUCT(*p_export_data_phy_node);
}

void ClearExportDataPort(export_data_phy_port_t *p_export_data_phy_port)
{
    delete p_export_data_phy_port->p_dd_phys_counters;
    delete p_export_data_phy_port->p_dd_trouble_shooting_info;
    delete p_export_data_phy_port->p_dd_operation_info;
    delete p_export_data_phy_port->p_dd_phy_info;
    delete p_export_data_phy_port->p_dd_module_info;
    delete p_export_data_phy_port->p_dd_link_down_info;
    delete p_export_data_phy_port->p_dd_link_up_info;
    delete p_export_data_phy_port->p_dd_latched_flag_info;
    delete p_export_data_phy_port->p_dd_rs_histograms;
    delete p_export_data_phy_port->p_dd_plr_counters;
    delete p_export_data_phy_port->p_dd_phy_statistical_counters;

    for (size_t i = 0; i < LANE_NUM; ++i) {
        delete p_export_data_phy_port->p_slrg_reg[i];
        delete p_export_data_phy_port->p_sllm_reg[i];
        delete p_export_data_phy_port->p_slrp_reg[i];
        delete p_export_data_phy_port->p_sltp_reg[i];
        for (size_t j = 0; j < SLRIP_IB_SEL_NUM; ++j)
            delete p_export_data_phy_port->p_slrip_reg[i][j];
    }

    CLEAR_STRUCT(*p_export_data_phy_port);
}

int PhyDiag::ExportData(export_session_handle_t session_handle,
                        list_p_fabric_general_err &export_data_errors)
{
    IBDIAGNET_ENTER;

    static export_data_phy_port_t export_data_phy_port;
    static export_data_phy_node_t export_data_phy_node;

    int rc, rc2;

    if ((rc = this->InitExportAPI(export_data_errors)))
        goto exit;

    CLEAR_STRUCT(export_data_phy_port);
    CLEAR_STRUCT(export_data_phy_node);

    // Nodes
    for (map_str_pnode::iterator nI = this->p_discovered_fabric->NodeByName.begin();
            nI != this->p_discovered_fabric->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = nI->second;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        export_data_phy_node.node_guid = p_curr_node->guid_get();

        // Diagnostic Data Pages
        for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size(); ++dd_idx) {
            DiagnosticDataInfo *p_dd = this->diagnostic_data_vec[dd_idx];

            if (p_dd->IsPerNode()) {
                if (p_dd->GetDDType() == DD_PCI_TYPE) {
                    if (dd_idx < this->dd_idx_and_akey_to_dd_page.size()) {
                        map_akey_diagnostic_data_t *p_map_akey_dd =
                            this->dd_idx_and_akey_to_dd_page[dd_idx];
                        if (!p_map_akey_dd)
                            continue;

                        for (map_akey_diagnostic_data_t::iterator mI = p_map_akey_dd->begin();
                                mI != p_map_akey_dd->end(); ++mI) {
                            AccRegKey *p_curr_acc_reg_key = mI->first;
                            if (!p_curr_acc_reg_key)
                                continue;
                            VS_DiagnosticData *p_curr_data = mI->second;
                            if (!p_curr_data)
                                continue;

                            p_dd->ExportDiagnosticData(NULL, &export_data_phy_node, *p_curr_data);
                        }
                    }
                } else {
                    VS_DiagnosticData *p_curr_data =
                        getPhysLayerNodeCounters(p_curr_node->createIndex, dd_idx);
                    if (!p_curr_data)
                        continue;

                    p_dd->ExportDiagnosticData(NULL, &export_data_phy_node, *p_curr_data);
                }
            }
        }

        // PHY Registers
        for (vec_pareg::iterator rI = this->reg_handlers_vec.begin();
                rI != this->reg_handlers_vec.end(); ++rI) {
            AccRegHandler *p_areg_handler = *rI;
            p_areg_handler->ExportData(NULL, &export_data_phy_node);
        }

        // PCI Registers
        for (vec_pareg::iterator rI = this->pci_reg_handlers_vec.begin();
                rI != this->pci_reg_handlers_vec.end(); ++rI) {
            AccRegHandler *p_areg_handler = *rI;
            p_areg_handler->ExportData(NULL, &export_data_phy_node);
        }

        // Export Data
        rc2 = pf_export_data_phy_node(session_handle, &export_data_phy_node);
        ClearExportDataNode(&export_data_phy_node);
        if (rc2) {
            ExportDataErr *p_curr_fabric_err =
                new ExportDataErr(p_curr_node, NULL, "Failed to export node data: %d", rc2);
            if (!p_curr_fabric_err) {
                this->SetLastError("Failed to allocate ExportDataErr");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto exit;
            }
            export_data_errors.push_back(p_curr_fabric_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
        }

        // Ports
        for (phys_port_t i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort(i);
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            export_data_phy_port.node_guid = p_curr_port->p_node->guid_get();
            export_data_phy_port.port_guid = p_curr_port->guid_get();
            export_data_phy_port.port_num = p_curr_port->num;

            // Diagnostic Data Pages
            for (u_int32_t dd_idx = 0; dd_idx < this->diagnostic_data_vec.size(); ++dd_idx) {
                DiagnosticDataInfo *p_dd = this->diagnostic_data_vec[dd_idx];

                if (!p_dd->IsPerNode()) {
                    if (p_dd->GetDDType() != DD_PCI_TYPE) {
                        VS_DiagnosticData *p_curr_data =
                            getPhysLayerPortCounters(p_curr_port->createIndex, dd_idx);
                        if (!p_curr_data)
                            continue;

                        p_dd->ExportDiagnosticData(&export_data_phy_port, NULL, *p_curr_data);
                    }
                }
            }

            // PHY Registers
            for (vec_pareg::iterator rI = this->reg_handlers_vec.begin();
                    rI != this->reg_handlers_vec.end(); ++rI) {
                AccRegHandler *p_areg_handler = *rI;
                p_areg_handler->ExportData(&export_data_phy_port, NULL);
            }

            // Export Data
            rc2 = pf_export_data_phy_port(session_handle, &export_data_phy_port);
            ClearExportDataPort(&export_data_phy_port);
            if (rc2) {
                ExportDataErr *p_curr_fabric_err =
                    new ExportDataErr(p_curr_node, p_curr_port, "Failed to export port data: %d", rc2);
                if (!p_curr_fabric_err) {
                    this->SetLastError("Failed to allocate ExportDataErr");
                    rc = IBDIAG_ERR_CODE_NO_MEM;
                    goto exit;
                }
                export_data_errors.push_back(p_curr_fabric_err);
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            }
        }
    }

exit:
    IBDIAG_RETURN(rc);
}

Plugin * create(IBDiag *p_ibdiag)
{
    return(new PhyDiag(p_ibdiag));
}

void destroy(Plugin *p)
{
    delete (PhyDiag *)p;
    return;
}
